require('engine/utility/tableUtil')
require('engine/json/json')

local json = require('json');

----------------------------------------------------------------------

local USE_TEST_DATA = false;

if not BUILD_DEBUG then
   USE_TEST_DATA = false;
end

local DEFAULT_NUM_ADS = 5;

----------------------------------------------------------------------

NativeAdRetriever = {
   _nativeAds = {},
   _adImages = {},
   _impressedAds = {},

   _enabled = false,
   _enabledForSpenders = false,
   _adSpacing = 4,
   _useExternalBrowser = false,
}

----------------------------------------------------------------------

---------------
-- INTERFACE --
---------------

function NativeAdRetriever:new(obj)

   obj = obj or {};
   setmetatable(obj, self);
   self.__index = self;

   return obj;
end


function NativeAdRetriever:getNativeAds()
   return self._nativeAds;
end


function NativeAdRetriever:getLocalAdImageFromAdUID(uid)
   return self._adImages[uid];
end


function NativeAdRetriever:getEnabled()

   return AbTestMgr:choose(self._enabled, AbTestMgr:getNativeAdsEnabled());
end


function NativeAdRetriever:getEnabledForSpenders()

   return AbTestMgr:choose(self._enabledForSpenders, AbTestMgr:getNativeAdsEnabledForSpenders());
end


function NativeAdRetriever:getDesiredAdSpacing()

   return AbTestMgr:choose(self._adSpacing, AbTestMgr:getNativeAdsSpacing());
end


function NativeAdRetriever:shouldUseExternalBrowser()

   return AbTestMgr:choose(self._useExternalBrowser, AbTestMgr:shouldAdsUseExternalBrowser());
end


function NativeAdRetriever:findAdWithUID(uid)

   for _, v in pairs(self._nativeAds) do
      if v.uid == uid then
         return v;
      end
   end

   return nil;
end

----------------------------------------------------------------------

function NativeAdRetriever:applyCloudData(data)

   Debug.assertType(data, {"table"}, "NativeAdRetriever:applyCloudData -> data");

   if data then

      if data.enable ~= nil then
         self._enabled = data.enable;
      end

      if data.enableForSpenders ~= nil then
         self._enabledForSpenders = data.enableForSpenders;
      end

      if data.adSpacing ~= nil then
         self._adSpacing = data.adSpacing;
      end

      if data.useExternalBrowser ~= nil then
         self._useExternalBrowser = data.useExternalBrowser;
      end

   end
end

----------------------------------------------------------------------

function NativeAdRetriever:attemptAdFetch(numAds)

   Debug.assertType(numAds, {"nil", "number"}, "NativeAdRetriever:attemptAdFetch -> numAds");

   numAds = numAds or DEFAULT_NUM_ADS;

   print("Attempting to fetch ads...");

   if self:shouldFetchAds() then

      print("Fetching ads!");
      self:_fetchNativeAds(numAds);

      return true;
   end

   return false;
end

----------------------------------------------------------------------

function NativeAdRetriever:shouldFetchAds()

   -- Set up default

   if not self:getEnabled() then
      print("Not fetching ads - disabled for all.");
      return false;
   end

   if App.data:isPayer() and not self:getEnabledForSpenders() then
      print("Not fetching ads - disabled for spenders and this account bought gems.");
      return false;
   end

   if self._nativeAds and not isTableEmpty(self._nativeAds) then
      -- We already have native ads.
      return false;
   end

   return true;

end

----------------------------------------------------------------------

function NativeAdRetriever:clearAdImpressions()
   self._impressedAds = {};
end

function NativeAdRetriever:handleAdImpression(ad)

   if ad ~= nil then

      Debug.assert(ad.imp_track, "Missing impression tracking url table!");

      if self._impressedAds[ad.uid] then
         return;
      end

      -- print("Handling impression for ad: ", ad.title);

      for _, impressionUrl in pairs(ad.imp_track) do

         local requestOptions = {

            priority = Request.PRIORITY_HIGH, -- time-sensitive
            successFunc = function() EventMgr:postEvent( { result = "success", title = ad.title, provider = ad.provider_type, url = impressionUrl }, "native-ad-impressed-finished" ); end,
            failureFunc = function() EventMgr:postEvent( { result = "failure", title = ad.title, provider = ad.provider_type, url = impressionUrl }, "native-ad-impressed-finished" ); end,
            url = impressionUrl,
            retryCount = 3,
            ignoreResponse = true,
         }

         NetworkMgr:sendWebRequest(WebRequest.newFromBase(requestOptions));

      end

      self._impressedAds[ad.uid] = true;

      EventMgr:postEvent( { title = ad.title, provider = ad.provider_type }, "native-ad-impressed" );

   else

      Debug.error("NativeAdRetriever:handleAdImpression->ad is nil");

   end

end

----------------------------------------------------------------------

function NativeAdRetriever:handleAdClick(ad)

   Debug.assert(ad.click_track, "Missing click tracking url table!");

   -- print("Handling click for ad: ", ad.title);

   for _, trackingUrl in pairs(ad.click_track) do

      local requestOptions = {

         priority = Request.PRIORITY_HIGH, -- time-sensitive
         successFunc = function() EventMgr:postEvent( { result = "success", title = ad.title, provider = ad.provider_type, url = trackingUrl }, "native-ad-clicked-finished" ); end,
         failureFunc = function() EventMgr:postEvent( { result = "failure", title = ad.title, provider = ad.provider_type, url = trackingUrl }, "native-ad-clicked-finished" ); end,
         url = trackingUrl,
         retryCount = 3,
         ignoreResponse = true,
      }

      NetworkMgr:sendWebRequest(WebRequest.newFromBase(requestOptions));

   end

   EventMgr:postEvent( { title = ad.title, provider = ad.provider_type }, "native-ad-clicked" );

end

function NativeAdRetriever:sendUserToURL(url)

   if self:shouldUseExternalBrowser() then
      App:gotoPage(url);
   else

      if MOAIWebViewIOS then

         StateMgr:push("state/state-view-native-ad.lua", { url = url } );

      elseif MOAIWebViewAndroid then

         print("Creating webview.");
         local webView = MOAIWebViewAndroid.new();

         print("Showing webview.");
         webView:show(true);

         print("Loading webview request.");
         webView:loadRequest(url);

      else
         Debug.error("Missing both webview classes!");
      end

   end

end

----------------------------------------------------------------------

--------------------
-- IMPLEMENTATION --
--------------------

function NativeAdRetriever:_fetchNativeAds(numAds)

   EventMgr:postEvent( { count = numAds }, "native-ad-request" );

   if not USE_TEST_DATA then

      local successFunc = function( content )

         print("Got native ad content from the Tango SDK!");

         print("Native ad content:", content);

         local adData = json.decode(content);

         self._nativeAds = self:_createAdsFromJSONResponse(adData);

         EventMgr:postEvent( { result = "success" }, "native-ad-response" );

         self:_downloadAdImages();

      end

      local failureFunc = function( )

         print("NativeAds: failed to get ads!");

         EventMgr:postEvent( { result = "failure" }, "native-ad-response" );

      end

      local adRequestData = {
         ad_info_request = {
            cid = "roadriot",
            adsize = numAds,
         }
      }

      local encodedJSON = json.encode(adRequestData);
      print("NativeAds: encodedJSON: " .. tostring(encodedJSON));

      local requestOptions = {
         requestClass = MOAITangoGetAdsRequest,
         requestData = { encodedJSON },
         successFunc = successFunc,
         failureFunc = failureFunc,
      }

      NetworkMgr:sendTangoRequest(TangoRequest.newFromBase(requestOptions));

   else

      local response = require("app/ad/native-ad-test-data")
      local adData = json.decode(response);

      self._nativeAds = self:_createAdsFromJSONResponse(adData);

      self:_downloadAdImages();

   end

end

----------------------------------------------------------------------

function NativeAdRetriever:_handleAdImageFetch(task, fileName)

   Debug.assert(MOAIEnvironment.tempDirectory);

   local dataBuffer = MOAIDataBuffer.new();
   dataBuffer:setString(task:getString());

   dataBuffer:save(MOAIEnvironment.tempDirectory .. "/" .. fileName);

   print("Downloaded image: " .. fileName);

end

----------------------------------------------------------------------

function NativeAdRetriever:_downloadAdImages()

   Debug.assert(MOAIEnvironment.tempDirectory);

   -- Download icon images for ads and store them in a temporary manner

   local validNativeAds = {};

   for _, v in ipairs(self._nativeAds) do

      if v.icon_xhdpi and v.icon_xhdpi.url then

         Debug.assertType(v.icon_xhdpi.url, "string", "NativeAdRetriever:_downloadAdImages -> v.icon_xhdpi.url");

         local fileName = string.match(v.icon_xhdpi.url, "^.*/(.*)$");

         if fileName ~= nil then

            local fullFilePath = MOAIEnvironment.tempDirectory .. "/" .. fileName;

            print("Image to download: " .. v.icon_xhdpi.url);

            if not MOAIFileSystem.checkFileExists(fullFilePath) then

               local requestOptions = {
                  priority = Request.PRIORITY_LOW,
                  callHandler = NativeAdRetriever,
                  successFunc = function(callHandler, task) callHandler:_handleAdImageFetch(task, fileName) end,
                  url = v.icon_xhdpi.url,
                  retryCount = 3,
               }

               local request = WebRequest.newFromBase(requestOptions);

               NetworkMgr:sendWebRequest(request);

            else
               print("Skipped image download; read from file instead: " .. fileName)
            end

            self._adImages[v.uid] = fullFilePath;

            table.insert(validNativeAds, v);
         else
            Debug.error("Missing image URL to download...");
         end
      end

   end

   self._nativeAds = validNativeAds;

end

----------------------------------------------------------------------

function NativeAdRetriever:_createAdsFromJSONResponse(adData)

   Debug.assert(adData.adinforesponse);

   local ads = {}

   if adData.adinforesponse.status == "OK" then
      Debug.assert(type(adData.adinforesponse.ads) == "table", "Ads entry was nil or not a table!");

      if adData.adinforesponse.ads then

         for _, adData in ipairs(adData.adinforesponse.ads) do

            if not isTableEmpty(adData) then

               local ad = {};

               -- Mandatory fields
               ad.uid = adData.uid;
               ad.provider_type = adData.provider_type;
               ad.title = adData.title;
               ad.subtitle = adData.subtitle;
               ad.click_url = adData.click_url;
               ad.icon_xhdpi = adData.icon_xhdpi;

               -- Optional fields
               ad.image_xhdpi = adData.image_xhdpi;
               ad.app_store_url = adData.app_store_url;
               ad.click_track = adData.click_track;
               ad.imp_track = adData.imp_track;
               ad.country = adData.country;

               if ad.uid and ad.provider_type and ad.title and ad.subtitle and ad.click_url and ad.icon_xhdpi then

                  table.insert(ads, ad);

               else

                  EventMgr:postEvent( { title = ad.title, provider = ad.provider_type, clickUrl = ad.click_url }, "native-ad-received-invalid-ad" );

               end
            end
         end
      end
   else
      Debug.error("Flurry native ad response had a non-OK status!");
   end

   return ads;
end

----------------------------------------------------------------------
