require('engine/utility/tableUtil')

----------------------------------------------------------------------

--NOTE: order of global index must match order of templateTbl
CONSUMABLE_HEADSTART_INDEX    =  1;
CONSUMABLE_POWERTIME_INDEX    =  2;
CONSUMABLE_DOUBLESCORE_INDEX  =  3;
CONSUMABLE_150PCT_INDEX       =  4;
CONSUMABLE_300PCT_INDEX       =  5;
CONSUMABLE_400PCT_INDEX       =  6;
CONSUMABLE_500PCT_INDEX       =  7;

----------------------------------------------------------------------

ConsumableMgr	=
{
   _consumableTpl = nil,
   
   --NOTE: order of dataKey and flurryCountID must match templateTbl
   _flurryCountID =
   {
      "headstart-count",
      "powertime-count",
      "doublescore-count",
      "150pctgold-count",
      "300pctgold-count",
      "400pctgold-count",
      "500pctgold-count",
   },
	_templateTbl   =
	{
		"app/template/consumable/head_start.lua",
      "app/template/consumable/powerup_duration.lua",
      "app/template/consumable/double_score.lua",
      "app/template/consumable/coinK_150pct.lua",
      "app/template/consumable/coinK_300pct.lua",
      "app/template/consumable/coinK_400pct.lua",
      "app/template/consumable/coinK_500pct.lua",
	},
};

----------------------------------------------------------------------
-- Core
----------------------------------------------------------------------


function ConsumableMgr:_createConsumableTemplates()

   local tpl = {};

   for k,v in ipairs(self._templateTbl) do
      table.insert(tpl, dofile(v));

      tpl[k].cost        = AbTestMgr:getConsumableCost(tpl[k].type) or tpl[k].cost;
      tpl[k].duration    = AbTestMgr:getConsumableDuration(tpl[k].type) or tpl[k].duration;
      tpl[k].tplPowerup  = AbTestMgr:getConsumablePowerTpl(tpl[k].type) or tpl[k].tplPowerup;
      tpl[k].quantity    = AbTestMgr:getConsumableQuantity(tpl[k].type) or tpl[k].quantity;
      tpl[k].onSale      = AbTestMgr:isConsumableOnSale(tpl[k].type) or tpl[k].onSale;
   end

   return tpl;

end

----------------------------------------------------------------------

function ConsumableMgr:_createConsumableId( index )

   return App.data:createConsumableId(index);

end

----------------------------------------------------------------------

function ConsumableMgr:_getTemplates()

   if self._consumableTpl == nil then
      self._consumableTpl = self:_createConsumableTemplates();
   end

   return self._consumableTpl;

end


----------------------------------------------------------------------
-- Utility
----------------------------------------------------------------------

function ConsumableMgr:canBuy( index )

   Debug.assertType(index, "number", "ConsumableMgr:canConsume -> index");

   local cost = self:getCost(index);

   if cost.gem and cost.gem > 0 then
      return App.data:getGemCount() >= cost.gem;
   elseif cost.gold and cost.gold > 0 then
      return App.data:getGoldCount() >= cost.gold;
   end

   return true;

end

----------------------------------------------------------------------

function ConsumableMgr:canConsume( index )

   Debug.assertType(index, "number", "ConsumableMgr:canConsume -> index");

   if self:isSingleUse(index) and self:getInventoryCount(index) > 0 then
      return true;
   end

   return false;

end

----------------------------------------------------------------------

function ConsumableMgr:getCost( index )

   Debug.assertType(index, "number", "ConsumableMgr:getCost -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:getCost -> invalid consumable index");

   return (templates[index] or {}).cost or {gold=0, gem=0};

end

----------------------------------------------------------------------

function ConsumableMgr:getDesc( index )

   Debug.assertType(index, "number", "ConsumableMgr:getDesc -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:getDesc -> invalid consumable index");

   return (templates[index] or {}).desc or "";

end

----------------------------------------------------------------------

function ConsumableMgr:getDuration( index )

   Debug.assertType(index, "number", "ConsumableMgr:getDuration -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:getDuration -> invalid consumable index");

   return math.max( (templates[index] or {}).duration or 0, 0);

end

----------------------------------------------------------------------

function ConsumableMgr:getHighestActiveCoinK()

   --NOTE:range must be based on the template table
   for i=CONSUMABLE_500PCT_INDEX, CONSUMABLE_150PCT_INDEX, -1 do

      if not self:isSingleUse(i) and not self:isDurationExpired(i) then

         --NOTE:returns index and time remaining
         return i, self:getTimeRemaining(i);
      end
   end

end

----------------------------------------------------------------------

function ConsumableMgr:getInventoryCount( index )

   Debug.assertType(index, "number", "ConsumableMgr:getInventoryCount -> index");

   if self:isSingleUse(index) then
      return App.data:getValue(self:_createConsumableId(index), 0);
   elseif not self:isDurationExpired(index) then
      return 1;
   end

   return 0;

end

----------------------------------------------------------------------

function ConsumableMgr:getName( index )

   Debug.assertType(index, "number", "ConsumableMgr:getName -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:getName -> invalid consumable index");

   return (templates[index] or {}).name or "";

end

----------------------------------------------------------------------

function ConsumableMgr:getQuantityPerSale( index )

   Debug.assertType(index, "number", "ConsumableMgr:getQuantityPerSale -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:getQuantityPerSale -> invalid consumable index");

   return math.max( (templates[index] or {}).quantity or 1, 1);

end


----------------------------------------------------------------------

function ConsumableMgr:getTemplate( index )

   Debug.assertType(index, "number", "ConsumableMgr:getTemplate -> index");

   local templates = self:_getTemplates();

   return templates[index] or {};

end


----------------------------------------------------------------------

function ConsumableMgr:getTemplateIndexByType( cType )

   Debug.assertType(cType, "string", "ConsumableMgr:getTemplateByType -> index");

   local templates = self:_getTemplates();

   for k,v in ipairs(templates) do

      if v.type and v.type == cType then
         return k;
      end
   end

end

----------------------------------------------------------------------

function ConsumableMgr:getTimeRemaining( index )

   Debug.assertType(index, "number", "ConsumableMgr:getQuantityPerSale -> index");

   if self:isSingleUse(index) or self:isDurationExpired(index) then
      return 0;
   end

   local timeRef     = TimeMgr:getTimeGMT();
   local templates   = self:_getTemplates();

   return math.max( (App.data:getValue(self:_createConsumableId(index), 0) + self:getDuration(index)) - timeRef, 0 );

end


----------------------------------------------------------------------

function ConsumableMgr:getTotalTypes()

   local templates = self:_getTemplates();

   return #templates or 0;

end

----------------------------------------------------------------------

function ConsumableMgr:isDurationExpired( index )

   Debug.assertType(index, "number", "ConsumableMgr:isConsumableActive -> index");

   local timeRef     = TimeMgr:getTimeGMT();
   local templates   = self:_getTemplates();

   if not self:isSingleUse(index) then
      return timeRef >= (App.data:getValue(self:_createConsumableId(index), 0) + self:getDuration(index));
   end

   return true;

end

----------------------------------------------------------------------

function ConsumableMgr:isOnSale( index )

   Debug.assertType(index, "number", "ConsumableMgr:isConsumableOnSale -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:isConsumableOnSale -> invalid consumable index");

   if templates[index].onSale ~= nil then
      return templates[index].onSale;
   end

   return false;

end

----------------------------------------------------------------------

function ConsumableMgr:isSingleUse( index )

   Debug.assertType(index, "number", "ConsumableMgr:isSingleUse -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:isSingleUse -> invalid consumable index");

   local duration = self:getDuration(index);

   if duration <= 0 then
      return true;
   end

   return false;

end


----------------------------------------------------------------------
-- Data Transactions
----------------------------------------------------------------------


function ConsumableMgr:buy( index, autoActivate, freeCost, freeUse )

   Debug.assertType(index, "number", "ConsumableMgr:getQuantityPerSale -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:getQuantityPerSale -> invalid consumable index");

   if not index or not templates[index] or not templates[index].cost then
      return;
   end

   local price = {currency = "gem", cost=0};

   if not freeCost then
      if templates[index].cost.gold and templates[index].cost.gold > 0 then
         price = {currency = "gold", cost=templates[index].cost.gold};
      elseif templates[index].cost.gem and templates[index].cost.gem > 0 then
         price = {currency = "gem", cost=templates[index].cost.gem};
      end
   end

   App.data:addConsumable(index, self._flurryCountID[index], price, autoActivate, freeUse);

end

----------------------------------------------------------------------

function ConsumableMgr:consume( index, freeUse )

   Debug.assertType(index, "number", "ConsumableMgr:consume -> index");

   local templates = self:_getTemplates();

   Debug.assert( templates[index], "ConsumableMgr:consume -> invalid consumable index");

   if not index or not templates[index] or not templates[index].cost then
      return;
   end

   App.data:consumeConsumable(index, freeUse);

end


----------------------------------------------------------------------




