require('app/contact/server/car-info-uploader')
require('app/contact/server/car-info-retriever')
require('app/contact/server/contact-retriever')
require('app/contact/server/picture-retriever')
require('app/contact/server/profile-retriever')
require('engine/file/fileMgr')

----------------------------------------------------------------------

local AUTO_FRIENDS_INTERVAL_SUCCESS       = 60*15;
local AUTO_FRIENDS_INTERVAL_RETRY         = 60*5;
local CAR_INFO_UPLOAD_FIRST               = 60*2;
local CAR_INFO_UPLOAD_DEFAULT_COOLDOWN    = 60*2;
local CAR_INFO_RETRIEVE_FIRST             = 60*2;
local CAR_INFO_RETRIEVE_DEFAULT_COOLDOWN  = 60*60;    --No need to update super frequently

local PROFILE_SAVE_FILENAME_LUA = "ROAD_RIOT_PROFILE.lua";
local PROFILE_SAVE_FILENAME_JSON = "p234r03f82i817l234e.jp5";

local FRIENDS_SAVE_FILENAME_LUA = "ROAD_RIOT_FRIENDS.lua";
local FRIENDS_SAVE_FILENAME_JSON = "3f32r83i99e1385n3d6.jp5";

local GHOSTS_SAVE_FILENAME_LUA = "ROAD_RIOT_GHOSTS.lua";
local GHOSTS_SAVE_FILENAME_JSON = "s83p810o4o89k132y8.jp5";

local GUESTS_SAVE_FILENAME_LUA = "38w7dht452.q89";
local GUESTS_SAVE_FILENAME_JSON = "11g8u23e4s77t.jp5";

local PICTURE_INFO_MAP_SAVE_FILENAME_LUA = "ROAD_RIOT_PICTURE_INFO_MAP.lua";
local PICTURE_INFO_MAP_SAVE_FILENAME_JSON = "p1i66c76t90u21r32e1s.jp5";

local GUEST_MODE_ID                       = "__GUEST__MODE__";
local CAR_INFO_CODE_SHIFT                 = 128;
----------------------------------------------------------------------

ContactMgr  =
{
   AUTO_FRIENDS_DISABLE      = 0;
   AUTO_FRIENDS_ENABLE       = 1;
   AUTO_FRIENDS_ONCE         = 2;
};

----------------------------------------------------------------------

function ContactMgr:init()

   self:_reset();

   EventMgr:addListener( self, self._onGhostInfoUpdated,    "ghost-info-updated" );
   EventMgr:addListener( self, self._onGuestInfoUpdated,    "guest-info-updated" );

end

----------------------------------------------------------------------

function ContactMgr:activate()

   if not self._active then

      self:_reset();

      self:_activateProfile();
      self:_activateFriends();
      self:_activateGhosts();
      self:_activateGuests();

      self._active = true;
      self._downloadPictureEnable = false;

      -- self:_deleteAllProfilePictures();

   end

end

----------------------------------------------------------------------

function ContactMgr:deactivate()
   self._active = false;
end

----------------------------------------------------------------------

function ContactMgr:isActive()
   return self._active;
end

----------------------------------------------------------------------

function ContactMgr:isReady()

   return      self._profile                 ~= nil
         and   self._profile:getAccountId()  ~= nil
         and   self._friends                 ~= nil;

end

----------------------------------------------------------------------

function ContactMgr:isProfileLoaded()
   return self._profile ~= nil;
end

----------------------------------------------------------------------

function ContactMgr:areContactLoaded()
   return self._friends ~= nil;
end

----------------------------------------------------------------------

function ContactMgr:isDownloading()

   return    self._friendsDownloader  ~= nil
         or  self._profileDownloader  ~= nil;

end

----------------------------------------------------------------------

function ContactMgr:isUploading()

   return self._carInfoUploader  ~= nil;

end

----------------------------------------------------------------------

function ContactMgr:isUpToDate()

   if LoginMgr:isTango() then

      return      self._profileStatus == STATUS_TANGO_DOWNLOADED
            and   self._friendsStatus == STATUS_TANGO_DOWNLOADED;
   else
      return true;
   end

end

----------------------------------------------------------------------

function ContactMgr:isBusy()

   return   (not self:isProfileLoaded() )
         or self._friendsDownloader ~= nil;

end

----------------------------------------------------------------------

function ContactMgr:getProfile()
   return self._profile;
end

----------------------------------------------------------------------

function ContactMgr:getContactCount()

   if self._friends then

      if self._profile then
         return #self._friends + #self._ghosts + #self._guests + 1;
      else
         return #self._friends + #self._ghosts + #self._guests;
      end

   end

   return 0;

end

----------------------------------------------------------------------

function ContactMgr:getContact( index )

   if self._friends and index then
      if index <= #self._friends then
         return self._friends[index];
      elseif index <= (#self._friends+#self._ghosts) then
         return self._ghosts[index-#self._friends];
      elseif index <= (#self._friends+#self._ghosts+#self._guests) then
         return self._guests[index-#self._friends-#self._ghosts];
      elseif index == (#self._friends+#self._ghosts+#self._guests + 1) then
         return self._profile;
      end
   end

end

----------------------------------------------------------------------

function ContactMgr:findContact( tangoId )

   if self._contactsMap and tangoId then
      return self._contactsMap[tangoId];
   end

end

----------------------------------------------------------------------

function ContactMgr:getFriendsCount()

   if self._friends then
      return #self._friends;
   end

   return 0;

end

----------------------------------------------------------------------

function ContactMgr:getFriend( index )

   if self._friends and index then
      return self._friends[index];
   end

end

----------------------------------------------------------------------

function ContactMgr:setAutoContactsMode( mode )
   self._friendsAutoMode = mode;
end

----------------------------------------------------------------------

function ContactMgr:setDownloadPictureEnable( e )
   self._downloadPictureEnable = e;
end
----------------------------------------------------------------------

function ContactMgr:isDownloadPictureEnable()
   return self._downloadPictureEnable;
end

----------------------------------------------------------------

function ContactMgr:getContactPicture( profile )

   Debug.assertType(self._pictureInfoMap, "table", "ContactMgr:getContactPicture -> self._pictureInfoMap");
   assert(self._pictureInfoMap);

   if profile.fake then
      return nil;
   end

   if profile then
      if profile:isGhost() then

         if BountyMgr:isBountyEntryCollectReady(profile.tangoId) or BountyMgr:isBountyEntryCollected(profile.tangoId) then
            return profile.defeated;
         else
            return profile.portrait;
         end
      else
         if self._pictureInfoMap then

            local info = self._pictureInfoMap[profile:getAccountId()];
            if info and info.state == "saved" then
               return info.filename;
            end
         end
      end
   end
end

function ContactMgr:_deleteAllProfilePictures()

   for accountID, entry in pairs(self._pictureInfoMap) do

      -- Switch working directory
      local workingDir = MOAIFileSystem.getWorkingDirectory();
      MOAIFileSystem.setWorkingDirectory( MOAIEnvironment.documentDirectory);

      local fileName = entry.filename;

      if MOAIFileSystem.deleteFile(fileName) then
         print("Successfully deleted picture:", fileName);
      else
         print("Failed to delete picture:", fileName);
      end

      -- Restore working directory
      MOAIFileSystem.setWorkingDirectory(workingDir);

   end

   self._pictureInfoMap = {};

end

----------------------------------------------------------------

function ContactMgr:update( dt )

   if self:isActive() then

      self:_updatePictureRetrieval( dt );

      self:_updateFriendRetrieval( dt );

      self:_updateProfileRetrieval( dt );

      self:_updateUploadCarInfo( dt );

      self:_updateRetrieveCarInfo( dt );

   end

end

----------------------------------------------------------------------

function ContactMgr:_reset()

   self._active                  = false;

   self._profile                 = nil;
   self._profileStatus           = STATUS_UNKNOWN;
   self._profileDownloadTimer    = nil;
   self._profileDownloader       = nil;

   self._friends                 = nil;
   self._friendsStatus           = STATUS_UNKNOWN;
   self._friendsDownloadTimer    = 0;
   self._friendsDownloader       = nil;
   self._friendsAutoMode         = self.AUTO_FRIENDS_ENABLE;

   self._ghosts                  = {};
   self._ghostsStatus            = STATUS_UNKNOWN;

	self._guests                  = {};
	self._guestsStatus            = STATUS_UNKNOWN;

   self._contactsMap             = {};

   self._pictureInfoMap          = loadFileMultiFormat(PICTURE_INFO_MAP_SAVE_FILENAME_LUA, PICTURE_INFO_MAP_SAVE_FILENAME_JSON) or {};
   self._pictureMap              = {};
   self._pictureRetriever        = nil;
   self._pictureToDownloadList   = {};

   --If we decide to upload, do it in a little while, it doesnt have to be super accurate
   self._carInfoUploadTimer      = CAR_INFO_UPLOAD_FIRST;
   self._carInfoUploader         = nil;

   self._carInfoRetrieveTimer    = CAR_INFO_RETRIEVE_FIRST;
   self._carInfoRetriever        = nil;


end

----------------------------------------------------------------

function ContactMgr:_saveProfile()

   if self._profile then

      local p =
      {
         firstName   = self._profile:getFirstName(),
         lastName    = self._profile:getLastName(),
         accountId   = self._profile:getAccountId(),
         carIndex    = self._profile:getCarIndex(),
         paintIndex  = self._profile:getPaintIndex(),
         armorUpgrade= self._profile:getArmorUpgrade(),
         flags       = self._profile:getFlags(),
      };

      saveFileMultiFormat(PROFILE_SAVE_FILENAME_LUA, PROFILE_SAVE_FILENAME_JSON, p);
   end

end

----------------------------------------------------------------

function ContactMgr:_loadProfile()

   local tbl = loadFileMultiFormat(PROFILE_SAVE_FILENAME_LUA, PROFILE_SAVE_FILENAME_JSON);

   if tbl then

      local  p  =  MOAITangoContact.new();

      p:setFirstName(tbl.firstName);
      p:setLastName(tbl.lastName);
      p:setAccountId(tbl.accountId);
      p:setCarIndex(tbl.carIndex);
      p:setPaintIndex(tbl.paintIndex);
      p:setArmorUpgrade(tbl.armorUpgrade);
      p:setFlags(tbl.flags);

      --It should be already saved but be 100% safe
      p:setAppInstalled(true);
      return p;

   end
end


----------------------------------------------------------------

function ContactMgr:_createProfile( first, last, accountId )

   local  p  =  MOAITangoContact.new();
   p:setFirstName(first);
   p:setLastName(last);
   p:setAccountId(accountId);
   p:setAppInstalled(true);

   return p;

end

----------------------------------------------------------------

function ContactMgr:_onProfileUpdated( profile, status )

   local backup         = self._profile;

   self._profile        = profile;
   self._profileStatus  = status;

   if self._profile then

      MOAIFlurry:setUserId(      self._profile:getAccountId()
                           ..    " "
                           ..    self._profile:getFirstName()
                           ..    " "
                           ..    self._profile:getLastName()      );

      self._profile.isGhost = function()
         return false;
      end

      self._profile.isGuestFriend = function()
         return false;
      end

      self._profile.isFriend = function()
         return true;
      end

      --TODO: tie this into the real league system
      self._profile.isLeague   = function() return false end

      if backup then
         self._profile:setCarIndex( backup:getCarIndex() );
         self._profile:setPaintIndex( backup:getPaintIndex() );
         self._profile:setArmorUpgrade( backup:getArmorUpgrade() );
      end

      self:_startPictureDownloadIfNeeded( self._profile );

      self:_rebuildContactList();

      self:_saveProfile();

   end

   EventMgr:postEvent(
      {
         status      = self._profileStatus,
         profile     = self._profile
      },
      "tango-profile-updated"        );


end

----------------------------------------------------------------

function ContactMgr:_activateProfile()

   if LoginMgr:isTango() then

      local profile  = self:_loadProfile();

      if profile and profile:getAccountId() ~= GUEST_MODE_ID then
         self:_onProfileUpdated( profile, STATUS_TANGO_LOCAL );
      else
         self:_onProfileUpdated( nil, STATUS_TANGO_NIL );
      end

   else

      local profile  = self:_loadProfile();

      if profile and profile:getAccountId() == GUEST_MODE_ID then
         self:_onProfileUpdated( profile, STATUS_GUEST );
      else
         self:_onProfileUpdated( self:_createProfile("You","",GUEST_MODE_ID), STATUS_GUEST);
      end

   end

end

----------------------------------------------------------------

function ContactMgr:_downloadProfile()

   if       self._profileDownloader == nil
      and   LoginMgr:isTango()
      and   OnlineMgr:isOnline()                   then

      self._profileDownloader = ProfileRetriever:new();
      self._profileDownloader:exec( self
         ,
         function( self, profile )
            assert(profile);

            self:_onProfileUpdated( profile, STATUS_TANGO_DOWNLOADED );

            self._profileDownloader = nil;
            self._profileDownloadTimer = nil;
         end
         ,
         function( self )
            self._profileDownloader    = nil;

            --If we have no profile == urgent... download again
            if    self._profileStatus  == STATUS_TANGO_NIL
               or self._profile        == nil               then
               self._profileDownloadTimer = 1;
            else
               --We already have a profile, give it some time to
               -- not jam it and make it worst
               self._profileDownloadTimer = 60;
            end
         end
         );
   end
end

----------------------------------------------------------------

function ContactMgr:_startPictureDownloadIfNeeded( contact )

   assert( contact );

   local accountId   = contact:getAccountId();
   local oldPicEntry = self._pictureInfoMap[accountId] or {};
   local oldPicUrl   = oldPicEntry.url;
   local newPicUrl   = nil;

   if contact:hasPhotoUrl() then
      newPicUrl = contact:getPhotoUrl();

   elseif contact:isGuestFriend() and contact.portrait ~= "" then

      newPicUrl = contact.portrait;

   end

   --If it failed while downloading.. it could have been caused by a crash, dont try again for a while
   if oldPicEntry.state == "downloading" and (os.time() - (oldPicEntry.timeStamp or 0) < (60*60*24*7)) then

      --No download for you!

   else

      --If the picture was not finish to be downloaded and saved
      if oldPicEntry.state ~= "saved" and oldPicEntry.state ~= "not-supported" then

         oldPicEntry = {};
         oldPicUrl  = nil;
      end

      if oldPicUrl ~= newPicUrl then

         local entry = nil;
         if newPicUrl then

            entry =
            {
               accountId   = accountId
            ,  url         = newPicUrl
            ,  filename    = "picture-"..accountId..".png"
            ,  state       = "waiting"
            };

            if contact:isAppInstalled() and #self._pictureToDownloadList > 0 then
               if (contact == self:getProfile()) or (#self._pictureToDownloadList < 2) then
                  table.insert(  self._pictureToDownloadList, 1, entry );
               else
                  table.insert(  self._pictureToDownloadList, 2, entry );
               end
            else
               table.insert(  self._pictureToDownloadList, entry );
            end

         end
      end
   end
end

----------------------------------------------------------------

function  ContactMgr:onEnterForeground()

   if self._friends and #self._friends == 0 then

      -- Make sure to check our contact asap
      self._friendsDownloadTimer = 0;
   end

end

----------------------------------------------------------------

function ContactMgr:_saveFriends()

   if self._friends then

      local tbl = {};

      for i,f in ipairs(self._friends) do

         local tbl2 =
         {
            firstName   = f:getFirstName(),
            lastName    = f:getLastName(),
            accountId   = f:getAccountId(),
            carIndex    = f:getCarIndex(),
            paintIndex  = f:getPaintIndex(),
            armorUpgrade= f:getArmorUpgrade(),
            flags       = f:getFlags(),
         };
         table.insert(tbl, tbl2);
      end

      saveFileMultiFormat(FRIENDS_SAVE_FILENAME_LUA, FRIENDS_SAVE_FILENAME_JSON, tbl);

   end
end

----------------------------------------------------------------

function ContactMgr:_loadFriends()

   local tbl = loadFileMultiFormat( FRIENDS_SAVE_FILENAME_LUA, FRIENDS_SAVE_FILENAME_JSON );

   if tbl then

      local friends  = {};

      for i,tbl2 in ipairs( tbl ) do

         local  p  =  MOAITangoContact.new();

         p:setFirstName(tbl2.firstName);
         p:setLastName(tbl2.lastName);
         p:setAccountId(tbl2.accountId);
         p:setCarIndex(tbl2.carIndex);
         p:setPaintIndex(tbl2.paintIndex);
         p:setArmorUpgrade(tbl2.armorUpgrade);

         p:setFlags(tbl2.flags);

         table.insert( friends, p );
      end

      return friends;

   end

end

----------------------------------------------------------------

function ContactMgr:_onFriendsUpdated( friends, status )

   local friendsBackupMap  = {};

   if self._friends then
      for i,friend in ipairs(self._friends) do
         friendsBackupMap[friend:getAccountId()] = friend;

      end
   end

   self._friends           = nil;
   self._friendsStatus     = status;

   if friends then

      local friendsMap = {};

      self._friends = {};

      for _, friend in ipairs(friends) do

         --Make sure we do not have duplicate...
         -- there has been bug coming from the server which duplicate some entries
         if friendsMap[friend:getAccountId()] == nil then

            local backup = friendsBackupMap[friend:getAccountId()];
            if backup then
               friend:setCarIndex( backup:getCarIndex() );
               friend:setPaintIndex( backup:getPaintIndex() );
               friend:setArmorUpgrade( backup:getArmorUpgrade() );
            end

            friendsMap[friend:getAccountId()]  = friend;

            table.insert(self._friends, friend);

            friend.isGhost = function()
               return false;
            end
            friend.isGuestFriend = function()
               return false;
            end
            friend.isFriend = function()
               return true;
            end

            self:_startPictureDownloadIfNeeded( friend );
         end
      end

      self:_rebuildContactList();

      self:_saveFriends();

   end

   EventMgr:postEvent(
      {
         status      = self._friendsStatus,
         friends     = self._friends
      },
      "tango-friends-updated"        );


  -- self._contactBookDisabled = nil;
  -- self._friendsDownloadTimer    = 0;

   self._friendsDownloader = nil;

end

----------------------------------------------------------------

function ContactMgr:_activateFriends()

   if LoginMgr:isTango() then

      local friends  = self:_loadFriends();

      if friends then

         self:_onFriendsUpdated( friends, STATUS_TANGO_LOCAL );

      else

         self:_onFriendsUpdated( nil, STATUS_TANGO_NIL );

      end

   else

      self:_onFriendsUpdated({}, STATUS_GUEST );

   end

end


----------------------------------------------------------------

function ContactMgr:_onFriendDownloadSuccess(friends)

   self:_onFriendsUpdated(friends, STATUS_TANGO_DOWNLOADED );

   if self._friendsAutoMode == self.AUTO_FRIENDS_ONCE then
      self._friendsAutoMode = self.AUTO_FRIENDS_DISABLE;
   end

   self._friendsDownloadTimer  = AUTO_FRIENDS_INTERVAL_SUCCESS;

   self._friendsDownloader = nil;

end

function ContactMgr:_onFriendDownloadFailure(code, message)

   print("Failed to download friends with code:", code, "message:", message);

   self._friendsDownloader = nil;

   self._friendsDownloadTimer = AUTO_FRIENDS_INTERVAL_RETRY;

end

function ContactMgr:_downloadFriends()

   if       self._friendsDownloader == nil
      and   LoginMgr:isTango()
      and   OnlineMgr:isOnline()                   then

         self._friendsDownloader = ContactRetriever:new();
         self._friendsDownloader:exec(self,
                                      function( self, friends )

                                         assert( friends );
                                         self:_onFriendDownloadSuccess(friends);
                                      end,

                                      function ( self, code, message )

                                         -- HACK
                                         -- This is a hack to fix the case where Tango reports a failure with "No contacts found"
                                         -- Example: <sdk_contacts.ERROR> Fetcher::handle_response: error: 18(No contacts found. Go to http://tango.me/nocontacts)
                                         -- This happens when authenticated to Tango but the user has no friends. :(

                                         if code == 18 then

                                            self:_onFriendDownloadSuccess({});

                                         else

                                            self:_onFriendDownloadFailure(code, message)

                                         end

                                      end
         );

   else

   end
end

----------------------------------------------------------------------

function ContactMgr:_updateFriendRetrieval( dt )

   --make sure we are not updating already
   if self._friendsDownloader == nil then

      self._friendsDownloadTimer   = self._friendsDownloadTimer - dt;

      if       self._friendsAutoMode ~= self.AUTO_FRIENDS_DISABLE
         and   self._friendsDownloadTimer < 0                            then

         self:_downloadFriends();

      end

   end
end

----------------------------------------------------------------

function ContactMgr:_saveGhosts()

   if self._ghosts then

      local tbl = {};

      for i,f in ipairs(self._ghosts) do

         local tbl2 =
         {
            accountId   = f:getAccountId(),
            firstName   = f:getFirstName(),
            lastName    = f:getLastName(),
            portrait    = f.portrait,
				defeated		= f.defeated;
				bountyText	= f.bountyText;
         };
         table.insert(tbl, tbl2);

      end

      saveFileMultiFormat(GHOSTS_SAVE_FILENAME_LUA, GHOSTS_SAVE_FILENAME_JSON, tbl);

   end
end

----------------------------------------------------------------

function ContactMgr:_loadGhosts()

   return loadFileMultiFormat( GHOSTS_SAVE_FILENAME_LUA, GHOSTS_SAVE_FILENAME_JSON );

end

----------------------------------------------------------------

function ContactMgr:_onGhostInfoUpdated( event )

   local ghosts = (event or {}).list;

   if LoginMgr:isTango() then
      self:_onGhostsUpdated(  (event or {}).list, STATUS_TANGO_LOCAL );
   else
      self:_onGhostsUpdated( (event or {}).list, STATUS_GUEST );
   end

end

----------------------------------------------------------------

function ContactMgr:_onGhostsUpdated( ghosts, status )

   self._ghosts        = {};
   self._ghostsStatus  = status;

   if ghosts then

      for i,ghost in ipairs(ghosts) do

         Debug.assert(ghost.firstName);
         Debug.assert(ghost.lastName);
         Debug.assert(ghost.accountId);
         Debug.assert(ghost.portrait);

         local c = MOAITangoContact.new();
         c:setFirstName(ghost.firstName);
         c:setLastName( ghost.lastName);
         c:setAccountId(ghost.accountId);
         c:setAppInstalled(true);
         c.portrait= ghost.portrait;
         c.defeated= ghost.defeated;
         c.bountyText= ghost.bountyText;
         c.tangoId = ghost.accountId;
         c.isGhost = function()
            return true;
         end
         c.isGuestFriend = function()
            return false;
         end
         c.isFriend = function()
            return false;
         end
         --TODO: tie this into the real league system
         c.isLeague = function()
            return false;
         end

         table.insert(self._ghosts, c);

      end


   end

   self:_rebuildContactList();
   self:_saveGhosts();

end

----------------------------------------------------------------

function ContactMgr:_activateGhosts()

   local ghosts  = self:_loadGhosts();

   if LoginMgr:isTango() then

      if ghosts then

         self:_onGhostsUpdated( ghosts, STATUS_TANGO_LOCAL );

      else

         self:_onGhostsUpdated( nil, STATUS_TANGO_NIL );

      end

   else

      self:_onGhostsUpdated(ghosts, STATUS_GUEST );

   end

end


----------------------------------------------------------------

function ContactMgr:clearGhosts( )

   self._ghosts = {};
   self:_rebuildContactList();

end

----------------------------------------------------------------

function ContactMgr:_saveGuests()

   if self._guests then

      local tbl = {};

      for i,f in ipairs(self._guests) do

         local tbl2 =
         {
            accountId   = f:getAccountId(),
            firstName   = f:getFirstName(),
            lastName    = f:getLastName(),
            portrait    = f.portrait,
         };
         table.insert(tbl, tbl2);

      end

      saveFileMultiFormat(GUESTS_SAVE_FILENAME_LUA, GUESTS_SAVE_FILENAME_JSON, tbl);
   end

end

----------------------------------------------------------------

function ContactMgr:_loadGuests()

   return loadFileMultiFormat( GUESTS_SAVE_FILENAME_LUA, GUESTS_SAVE_FILENAME_JSON );

end

----------------------------------------------------------------

function ContactMgr:_onGuestInfoUpdated( event )

   local guests = (event or {}).list;

   if LoginMgr:isTango() then
      self:_onGuestsUpdated( (event or {}).list, STATUS_TANGO_LOCAL );
   else
      self:_onGuestsUpdated( (event or {}).list, STATUS_GUEST );
   end

end

----------------------------------------------------------------

function ContactMgr:_onGuestsUpdated( guests, status )

   self._guests        = {};
   self._guestsStatus  = status;

   if guests then

      for i,guest in ipairs(guests) do

         Debug.assert(guest.firstName);
         Debug.assert(guest.lastName);
         Debug.assert(guest.accountId);
         Debug.assert(guest.portrait);

         local c = MOAITangoContact.new();
         c:setFirstName(guest.firstName);
         c:setLastName(guest.lastName);
         c:setAccountId(guest.accountId);
         c:setAppInstalled(true);
         c.portrait = guest.portrait;
         c.tangoId = guest.accountId;
         c.isGhost = function()
            return false;
         end
         c.isGuestFriend = function()
            return true;
         end

         if App.data:getValue("App.settings.download-guest-portraits", false) then
            self:_startPictureDownloadIfNeeded( c );
         end

         c.isFriend = function()
            return false;
         end

         --TODO: tie this into the real league system
         c.isLeague   = function() return true end

         table.insert(self._guests, c);

      end

   end

   self:_rebuildContactList();
   self:_saveGuests();

   EventMgr:postEvent(
      {
         status      = self._guestsStatus,
         friends     = self._guests
      },
      "guest-friends-updated"        );

end

----------------------------------------------------------------

function ContactMgr:_activateGuests()

   local guests  = self:_loadGuests();

   if LoginMgr:isTango() then
		self:_onGuestsUpdated(guests, STATUS_TANGO_LOCAL);
	else
      self:_onGuestsUpdated(guests, STATUS_GUEST );
   end

end

----------------------------------------------------------------

function ContactMgr:clearGuests( )

   self._guests = {};
   self:_rebuildContactList();

end

----------------------------------------------------------------

function ContactMgr:_rebuildContactList()

   self._contactsMap = {};

   if self._profile then
      self._contactsMap[self._profile:getAccountId()] = self._profile;
   end

   if self._ghosts then
      for i,ghost  in ipairs(self._ghosts) do
         self._contactsMap[ghost:getAccountId()] = ghost;
      end
   end

   if self._friends then
      for i,contact in ipairs(self._friends) do
         self._contactsMap[contact:getAccountId()] = contact;
      end
   end

	if self._guests then
      for i,guest  in ipairs(self._guests) do
         self._contactsMap[guest:getAccountId()] = guest;
      end
	end

end

----------------------------------------------------------------

function ContactMgr:_updatePictureRetrieval( dt )

   if self._pictureRetriever == nil and self:isDownloadPictureEnable() and (not App:isSavingForCriticalOnly()) then

      local lastDownload = math.floor( os.time() - (self._lastDownload or 0) );

      if #self._pictureToDownloadList > 0 and lastDownload > 2 then

         local todo = self._pictureToDownloadList[1];

         table.remove( self._pictureToDownloadList, 1);

         if todo then

            assert( todo.accountId );
            assert( todo.url );
            assert( todo.filename );

            self._lastDownload = os.time();

            todo.state         = "downloading";
            todo.timeStamp     = os.time();

            self._pictureInfoMap[todo.accountId] = todo;

            saveFileMultiFormat(PICTURE_INFO_MAP_SAVE_FILENAME_LUA, PICTURE_INFO_MAP_SAVE_FILENAME_JSON, self._pictureInfoMap);

            self._pictureRetriever = PictureRetriever:new( todo.accountId,
                                                           todo.url,
                                                           todo.filename );

            self._pictureRetriever:exec( self,
                                         function( )
                                            self._pictureRetriever = nil;
                                            todo.state  = "saved";
                                            self._pictureInfoMap[todo.accountId] = todo;

                                            saveFileMultiFormat(PICTURE_INFO_MAP_SAVE_FILENAME_LUA, PICTURE_INFO_MAP_SAVE_FILENAME_JSON, self._pictureInfoMap);

                                            EventMgr:postEvent( {acountId = todo.accountId}, "new-profile-picture");
                                         end,
                                         function(  self, errorEnum )
                                            self._pictureRetriever = nil;
                                            if errorEnum == ERROR_IMAGE_NOT_SUPPORTED then
                                               todo.state = "not-supported";
                                               self._pictureInfoMap[todo.accountId] = todo;

                                               saveFileMultiFormat(PICTURE_INFO_MAP_SAVE_FILENAME_LUA, PICTURE_INFO_MAP_SAVE_FILENAME_JSON, self._pictureInfoMap);

                                               print("Download failed with \"not-supported\"")
                                            else
                                               todo.failCount = (todo.failCount  or 0 ) + 1;
                                               if todo.failCount <= 2  then
                                                  todo.state  = "waiting";
                                                  table.insert( self._pictureToDownloadList, todo);
                                               end
                                            end
                                         end
            );

         end
      end
   end

end

----------------------------------------------------------------------

function ContactMgr:_updateProfileRetrieval( dt )

   assert(self:isActive());

   if    self._profileStatus == STATUS_TANGO_NIL
      or self._profileStatus == STATUS_TANGO_LOCAL then

      if self._profileDownloader == nil then
         self._profileDownloadTimer = (self._profileDownloadTimer or 0) - dt;
         if self._profileDownloadTimer < 0 then
            self:_downloadProfile();
         end
      end

   end

end

----------------------------------------------------------------------

function ContactMgr:_updateUploadCarInfo( dt )

   assert(self:isActive());

   if       self._profileStatus == STATUS_TANGO_DOWNLOADED
      and   self._carInfoUploader == nil                       then

      self._carInfoUploadTimer = (self._carInfoUploadTimer or 0) - dt;

      if self._carInfoUploadTimer < 0 then

         if self:_shouldUploadCarInfo() then

            self:_uploadCarInfo();

         else

            self._carInfoUploadTimer = CAR_INFO_UPLOAD_DEFAULT_COOLDOWN;

         end
      end
   end

end

----------------------------------------------------------------

function ContactMgr:_shouldUploadCarInfo()

   local p = self:getProfile();

   if p then

      return      p:getCarIndex()      ~= App.data:getCurrentPlayerVehicle()
            or    p:getPaintIndex()    ~= App.data:getCurrentPlayerVehiclePaintVariant()
            or    p:getArmorUpgrade()  ~= App.data:getCurrentArmorLevel();

   end

end

----------------------------------------------------------------

function ContactMgr:encodeCarInfo( carIndex, armorLevel, paintIndex )

   Debug.assert( carIndex     and carIndex >= 1    and carIndex   < CAR_INFO_CODE_SHIFT, "ContactMgr:encodeCarInfo -> carIndex     and carIndex   >= 1  and carIndex   < CAR_INFO_CODE_SHIFT");
   Debug.assert( paintIndex   and paintIndex >= 1  and paintIndex < CAR_INFO_CODE_SHIFT, "ContactMgr:encodeCarInfo -> paintIndex   and paintIndex >= 1  and paintIndex < CAR_INFO_CODE_SHIFT");
   Debug.assert( armorLevel   and armorLevel >= 1  and armorLevel < CAR_INFO_CODE_SHIFT, "ContactMgr:encodeCarInfo -> armorLevel   and armorLevel >= 1  and armorLevel < CAR_INFO_CODE_SHIFT");

   local code = (carIndex or 1);

   code = (code*CAR_INFO_CODE_SHIFT + (armorLevel or 1));
   code = (code*CAR_INFO_CODE_SHIFT + (paintIndex or 1));

   return code;

end

----------------------------------------------------------------

function ContactMgr:decodeCarInfo( code )

   Debug.assert( code and code >= 0, "ContactMgr:encodeCarInfo -> code and code >= 0");

   local code2      = math.floor( code );
   local code3      = math.floor( code/CAR_INFO_CODE_SHIFT );
   local code4      = math.floor( code/(CAR_INFO_CODE_SHIFT*CAR_INFO_CODE_SHIFT) );

   local paintIndex = code2  % CAR_INFO_CODE_SHIFT;
   local armorLevel = code3  % CAR_INFO_CODE_SHIFT;
   local carIndex   = code4 % CAR_INFO_CODE_SHIFT;

   return carIndex, armorLevel, paintIndex;

end

----------------------------------------------------------------

function ContactMgr:_uploadCarInfo()

   local p = self:getProfile();

   if       p
      and   self._carInfoUploader == nil
      and   LoginMgr:isTango()
      and   OnlineMgr:isOnline()          then

      local carIndex    = App.data:getCurrentPlayerVehicle();
      local armorLevel  = App.data:getCurrentArmorLevel();
      local paintIndex  = App.data:getCurrentPlayerVehiclePaintVariant();
      local code        = self:encodeCarInfo(carIndex, armorLevel,paintIndex);

      local carIndex2,
            armorLevel2,
            paintIndex2  = self:decodeCarInfo( code );

      Debug.assert( carIndex==carIndex2,      "ContactMgr:_uploadCarInfo", carIndex==carIndex2      );
      Debug.assert( armorLevel==armorLevel2,  "ContactMgr:_uploadCarInfo", armorLevel==armorLevel2  );
      Debug.assert( paintIndex==paintIndex2,  "ContactMgr:_uploadCarInfo", paintIndex==paintIndex2  );

      self._carInfoUploader = CarInfoUploader:new( self:getProfile():getAccountId(), code );

      self._carInfoUploader:exec(    self
                                    ,
                                       function( self )
                                          self._carInfoUploader      = nil;
                                          self._carInfoUploadTimer   = CAR_INFO_UPLOAD_DEFAULT_COOLDOWN;
                                          p:setCarIndex(carIndex);
                                          p:setPaintIndex(paintIndex);
                                          p:setArmorUpgrade(armorLevel);
                                          self:_saveProfile();
                                       end
                                    ,
                                       function( self, code, message )

                                          Debug.assert( nil, "ContactMgr:_uploadCarInfo -> Exec failed = " .. tostring(message) );

                                          self._carInfoUploader      = nil;
                                          self._carInfoUploadTimer   = CAR_INFO_UPLOAD_DEFAULT_COOLDOWN;
                                       end
                                    );
   end
end

----------------------------------------------------------------

function ContactMgr:_updateRetrieveCarInfo( dt )

   if       self._profileStatus == STATUS_TANGO_DOWNLOADED
      and   self._carInfoRetriever == nil                       then

      self._carInfoRetrieveTimer = (self._carInfoRetrieveTimer or 0) - dt;

      if self._carInfoRetrieveTimer < 0 then
            self:_retrieveCarInfo();
      end

   end
end


----------------------------------------------------------------

function ContactMgr:_retrieveCarInfo()

   local p = self:getProfile();

   self._carInfoRetriever = CarInfoRetriever:new( self );
   self._carInfoRetriever:exec(     self
                              ,
                                    function( self, infoTable )

                                       self._carInfoRetrieveTimer = CAR_INFO_RETRIEVE_DEFAULT_COOLDOWN;
                                       self._carInfoRetriever     = nil;

                                       for k,v in pairs(infoTable) do
                                          if self:getProfile() and k ~= self:getProfile():getAccountId() then
                                             local contact = self:findContact(k);
                                             if contact and v.code then

                                                local carIndex,
                                                      armorLevel,
                                                      paintIndex  = self:decodeCarInfo( v.code );

                                                local code2       = self:encodeCarInfo(carIndex, armorLevel,paintIndex);

                                                Debug.assert( v.code==code2,      "ContactMgr:_retrieveCarInfo", v.code==code2      );

                                                if carIndex then
                                                   contact:setCarIndex( carIndex );
                                                end

                                                if paintIndex and paintIndex >= 1 and paintIndex <= TemplateMgr:getPlayerVehiclePaintCount(paintIndex) then
                                                   contact:setPaintIndex( paintIndex );
                                                end

                                                if armorLevel and armorLevel >= ARMOR_LEVEL_MIN and armorLevel <= ARMOR_LEVEL_MAX then
                                                   contact:setArmorUpgrade( armorLevel );
                                                end
                                             end
                                          end
                                       end

                                       self:_saveFriends();

                                    end
                              ,
                                    function( self, code, message )

                                       Debug.assert( nil, "ContactMgr:_retrieveCarInfo -> Exec failed with code: " .. tostring(code) );

                                       self._carInfoRetrieveTimer = CAR_INFO_RETRIEVE_DEFAULT_COOLDOWN;
                                       self._carInfoRetriever     = nil;

                                    end);


end


----------------------------------------------------------------------
