require("engine/utility/functional");

local SAVE_TRANSACTIONS_COOLDOWN = 1.0;
local FILE_SAVE_MAX_ATTEMPT = 5;
local FILE_LOAD_MAX_ATTEMPT = 5;
local SECS_BETWEEN_UPLOADS = 5;

TransactionMgr = {

   _transactionQueue = {},
   _syncingTransactions = {},

   _saveTransactionFlag = false,
   _saveTransactionCooldown = 0,

   _forceDataMigration = false,
   _pendingMigration = false,

   _possessionTime = nil,
   _possessionOutDated = false,

   _uploadDelay = 1000,
   _juneUploadDelay = 1000,

   _dataBackup = nil,
}

---------------------------------------
-- PRIVATE --
---------------------------------------

function TransactionMgr:_reset()

   self._transactionQueue = {};
   self._syncingTransactions = {};

   self._saveTransactionFlag = false;
   self._saveTransactionCooldown = 0;

   self._forceDataMigration = false;
   self._pendingMigration = false;

   self._possessionTime = nil;
   self._possessionOutDated = false;

   self._uploadDelay = 1000;
   self._juneUploadDelay = 1000;

   self._dataBackup = nil;
end

function TransactionMgr:_getBackupFileName(appData)

   return (appData or App.data).fileName .. "._data_backup"
end

function TransactionMgr:_syncQueuedTransactions()

   -- We should never lose transactions.
   Debug.assert(#self._syncingTransactions == 0, "#self._syncingTransactions == 0");

   self._syncingTransactions = self._transactionQueue;
   self._transactionQueue = {};
end

function TransactionMgr:_requeueSyncingTransactions()

   -- If the transaction failed for connection reason rebuild our lists
   -- Moves all syncing transactions back into the queue

   local savedQueue = self._transactionQueue or {};

   self._transactionQueue = self._syncingTransactions;
   self._syncingTransactions = {};

   for _, t in ipairs(savedQueue) do
      -- Put the transactions that were syncing in the front of the queue
      table.insert( self._transactionQueue, 1, t );
   end

   self:_saveTransactionsSoon();

end

function TransactionMgr:_redoExecAllTransactions()

   Debug.assert( not self:isSyncing() );

   local sync = self._transactionQueue;

   self._transactionQueue = {};

   for _, t in ipairs(sync) do
      self:execTransaction(t);
   end

end

function TransactionMgr:_redoSetValuesAllTransactions()

   Debug.assert( not self:isSyncing() );

   local sync = self._transactionQueue;

   self._transactionQueue = {};

   for _, t in ipairs(sync) do

      local values = t:getValues();

      if values then
         App.data:setValues( values );
      end

   end

   self:_saveTransactionsSoon();

end

function TransactionMgr:_saveTransactionsSoon()

   self._saveTransactionFlag = true;

end

function TransactionMgr:_saveTransactions(appData)

   Debug.assertType(appData, "table", "TransactionMgr:_saveTransactions -> appData");

   local success1 = saveLocalFile( appData:_createTransactionSyncingFilename(), self._syncingTransactions, FILE_SAVE_MAX_ATTEMPT );
   local success2 = saveLocalFile( appData:_createTransactionToSyncFilename(),  self._transactionQueue,  FILE_SAVE_MAX_ATTEMPT );

   self._saveTransactionCooldown = 0;
   self._saveTransactionFlag = not (success1 and success2)

   if (not success1) or (not success2)  then
      EventMgr:postEvent( {critical = true}, "data-save-error"  );
   end

end

function TransactionMgr:_restoreTransactions(appData)

   Debug.assertType(appData, "table", "TransactionMgr:_restoreTransactions -> appData");

   self._syncingTransactions = loadLocalFile( appData:_createTransactionSyncingFilename(), FILE_LOAD_MAX_ATTEMPT ) or {};

   for _, t in ipairs(self._syncingTransactions) do
      DataTransaction.fixeMetatable( t );
   end

   self._transactionQueue = loadLocalFile( appData:_createTransactionToSyncFilename(), FILE_LOAD_MAX_ATTEMPT  ) or {};

   for _, t in ipairs(self._transactionQueue) do
      if t and type(t) == "table" then
         DataTransaction.fixeMetatable( t );
      end
   end

end

function TransactionMgr:_updateSave( dt )

   self._saveTransactionCooldown = self._saveTransactionCooldown + dt;

   if self._saveTransactionFlag and self._saveTransactionCooldown >= SAVE_TRANSACTIONS_COOLDOWN then
      self:save(App.data);
   end

end

-- If we have no games played or the default sports car is reportedly locked, it's safe to say that this is the default data that we're dealing with.
function TransactionMgr:_isSafeToOverwriteCloudData(cloudData)

   -- Somehow these fields were missing from the data table! This is most likely an error,
   -- so return false so other code does not make faulty assumptions about the data.
   if cloudData.gamePlayed == nil
   or cloudData.veh1mw == nil then
      return false;
   end

   return cloudData.gamePlayed == 0
      or cloudData.veh1mw == 0;

end

function TransactionMgr:_isSafeToOverwriteLocalData(localData)

   -- Somehow these fields were missing from the data table! This is most likely an error,
   -- so return false so other code does not make faulty assumptions about the data.
   if localData.gamePlayed == nil
   or localData.veh1mw == nil
   or localData.currentLevel == nil then
      return false;
   end

   return localData.gamePlayed <= 2
      or localData.veh1mw == 0
      or localData.currentLevel <= 1;

end

function TransactionMgr:_syncWithServer(dt)

--[[
    print("TransactionMgr:_syncWithServer - "
    .."\n OnlineMgr:isOnline() = "..tostring(OnlineMgr:isOnline())
    .."\n self:isSyncing() = "..tostring(self:isSyncing())
    .."\n LoginMgr:isTango() = "..tostring(LoginMgr:isTango())
    .."\n LoginMgr:isJune() = "..tostring(LoginMgr:isJune())
    .."\n self._juneUploadDelay = "..tostring(self._juneUploadDelay )
    .."\n self._juneUploadDelay > SECS_BETWEEN_UPLOADS = "..tostring(self._juneUploadDelay > SECS_BETWEEN_UPLOADS)
    .."\n self:hasQueuedTransactions() = "..tostring(self:hasQueuedTransactions())
    );
]]--
    -- TODO When we support anonymous Tango guest accounts, isTango will need to be changed.

    if OnlineMgr:isOnline() and not self:isSyncing() and not App.gamePlaying then
        if LoginMgr:isTango() then
            if PossessionMgr:getLastDownloadTime() and PossessionMgr:getPossessions() then

                --We are not in sync with with possession
                if PossessionMgr:getLastDownloadTime() ~= self._possessionTime then

                    if App.data:isGuestMode() or self._forceDataMigration then

                        print("TransactionMgr - DOING A GUEST -> TANGO DATA MIGRATION");

                        -- TODO Update migration when we support anonymous Tango accounts

                        -- We should push a state here.
                        -- TransactionMgr shouldn't have to do anything until the user picks their desired data.
                        -- If the user closes Road Riot, they should see the same screen when they reopen the game.

                        self._forceDataMigration = false;
                        self._pendingMigration = true;

                        local localPossessions = self:_createPossessionUploadTableForAllData();
                        local remotePossessions = PossessionMgr:getPossessions():getPossessionsTable();

                        if App.data:isPayer() or (remotePossessions["iap-verified"] or 0) > 0 then

                            EventMgr:postEvent({device=localPossessions, cloud=remotePossessions}, "need-data-resolution");

                        elseif self:_isSafeToOverwriteLocalData(localPossessions) then

                            -- We didn't have to go through a full resolution process.
                            self:chooseCloudData(true);

                        elseif self:_isSafeToOverwriteCloudData(remotePossessions) then

                            -- We didn't have to go through a full resolution process.
                            self:chooseLocalData(true);

                        else

                            EventMgr:postEvent({device=localPossessions, cloud=remotePossessions}, "need-data-resolution");

                        end

                    else

                        if self._possessionOutDated then

                            print("TransactionMgr - DATA OUTDATED. POSSIBLE MERGE.");
                            self._possessionTime = PossessionMgr:getLastDownloadTime();

                            if self:hasQueuedTransactions() then

                                local numAppliedTrans, numOriginalTrans = self:_replayTransactionsOntoCloudData();
                                local replayStatus = "error";

                                if numAppliedTrans == 0 then
                                    replayStatus = "none";
                                elseif numAppliedTrans == numOriginalTrans then
                                    replayStatus = "all";
                                elseif numAppliedTrans < numOriginalTrans then
                                    replayStatus = "some";
                                end

                                EventMgr:postEvent({strategy = "replay", replayStatus = replayStatus}, "data-merged");

                            elseif self:isSyncing() then

                                print("TransactionMgr - currently syncing. Retrying the sync.");
                                self:_requeueSyncingTransactions();
                                self:_replayTransactionsOntoCloudData();

                            else

                                print("TransactionMgr - No transactions. Just grab the cloud possessions.");
                                -- We don't have any transactions, but it looks like the possessions updated.
                                self:_applyPossessionDataOntoLocalData();
                                self._possessionOutDated = false;
                                EventMgr:postEvent({strategy="used-cloud"}, "data-merged");
                            end

                            App.data:ensureValidState();
                        else
                            self._possessionTime = PossessionMgr:getLastDownloadTime();
                        end

                    end

                else

                    Debug.assert(self._possessionTime);

                    --avoid sending data too frequently
                    self._uploadDelay = self._uploadDelay + dt;

                    if self._uploadDelay > SECS_BETWEEN_UPLOADS then

                        --There are transactions available
                        if not self:isSyncing() and self:hasQueuedTransactions() then
                            self._uploadDelay = 0;
                            self:_uploadTransactionsToServer();
                        end
                    end
                end
            else
                self._possessionTime = nil;
            end
        else
            self._possessionTime = nil;

            if LoginMgr:isJune() then
                --avoid sending data too frequently
                self._juneUploadDelay = self._juneUploadDelay + dt;

                if self._juneUploadDelay > SECS_BETWEEN_UPLOADS then

                    --There are transactions available
                    if not self:isSyncing() and self:hasQueuedTransactions() then
                        self._juneUploadDelay = 0;
                        self:_uploadTransactionsToJuneServer();
                    end
                end
            end
        end
    end
end


-- Returns the number of executed transactions.
function TransactionMgr:_attemptReplayTransactionQueue()

   Debug.assertType(App.data, {"table", "userdata"}, "TransactionMgr:_attemptReplayTransactionQueue -> App.data");

   local replayedValues = {};
   local numTransactionsApplied = 0;

   -- Phase 1: execute transactions until we hit a failure
   for _, transaction in ipairs(self._transactionQueue) do

      if self:_applyTransaction(transaction, replayedValues) then

         numTransactionsApplied = numTransactionsApplied + 1;
      else
         break;
      end

   end

   -- Phase 2: execute special transactions beyond the failure point
   if numTransactionsApplied < #self._transactionQueue then

      for i = numTransactionsApplied + 1, #self._transactionQueue do

         local transaction = self._transactionQueue[i];

         if transaction.alwaysExecute then

            self:_applyTransaction(transaction, replayedValues);
         end
      end
   end

   return replayedValues, numTransactionsApplied;
end


-- Returns true on success application, false otherwise.
function TransactionMgr:_applyTransaction(transaction, replayedValues)

   local values = transaction:buildValues(App.data);

   if type(values) == "table" then

      if App.data:setValues(values) > 0 then

         for k, v in pairs(values) do
            replayedValues[k] = v;
         end

         -- Call onSuccess after we change the real data
         transaction:onSuccess(App.data);
         transaction.applied = true;

         return true

      end

   end

   return false;

end


-- Set local data from cloud possessions
function TransactionMgr:_applyPossessionDataOntoLocalData(appData)

   -- print("TransactionMgr:_applyPossessionDataOntoLocalData")

   appData = appData or App.data;
   local possessions = PossessionMgr:getPossessions();

   for k, v in pairs(possessions:getPossessionsTable()) do
      appData:setValue(k, v);
   end

end

-- Get the possessions and apply our local transactions onto it until one fails.
-- If none fail, great. That's the ideal case.
-- If one fails, we need to stop trying to apply it and every transaction queued after that.
-- Then we set the local data to this newly resolved state.
function TransactionMgr:_replayTransactionsOntoCloudData()

   -- print("TransactionMgr:_replayTransactionsOntoCloudData")

   -- We actually need to set our local data from the cloud data because all the transactions depend on AppData...
   -- We can do this because we save a backup when we start queueing transactions.
   self:_applyPossessionDataOntoLocalData();

   -- Apply transactions until we fail
   local originalQueueSize = #self._transactionQueue;
   local replayedValues, numTransactionsApplied = self:_attemptReplayTransactionQueue();

   print("TransactionMgr: Applied " .. tostring(numTransactionsApplied) .. " out of " .. #self._transactionQueue .. " transactions.");

   -- Drop any transactions that were not applied in the previous step. They are destroyed.
   self._transactionQueue = filter(function(transaction) return transaction.applied; end, self._transactionQueue);

   self:_syncQueuedTransactions();

   -- Upload the data that we just merged.
   PossessionMgr:upload( self,
                         TransactionMgr._onUploadSuccess,
                         TransactionMgr._onUploadFailure,
                         replayedValues
   );

   return numTransactionsApplied, originalQueueSize;

end

function TransactionMgr:_uploadTransactionsToServer()

   print("TransactionMgr:_uploadTransactionsToServer");

   Debug.assert( LoginMgr:isTango() );
   Debug.assert( OnlineMgr:isOnline() );
   Debug.assert( self:hasQueuedTransactions() );
   Debug.assert( not self:isSyncing() );

   if LoginMgr:isTango()
      and OnlineMgr:isOnline()
      and self:hasQueuedTransactions()
      and not self:isSyncing() then

      local allValues = {};
      local count = 0;

      --Merge all values
      for _, t in ipairs(self._transactionQueue) do

         local values = t:getValues();

         if type(values) == "table" then

            for k, v in pairs(values) do

               allValues[k] = v;
               count        = count + 1;
            end
         end

      end

      self:_syncQueuedTransactions();

      self:_saveTransactionsSoon();

      if count > 0 then

         PossessionMgr:upload( self,
                               TransactionMgr._onUploadSuccess,
                               TransactionMgr._onUploadFailure,
                               allValues
         );

      else

         self._syncingTransactions = {};

      end

   end

end

function TransactionMgr:_uploadTransactionsToJuneServer()

    print("TransactionMgr:_uploadTransactionsToJuneServer");

    Debug.assert( LoginMgr:isJune() );
    Debug.assert( OnlineMgr:isOnline() );
    Debug.assert( self:hasQueuedTransactions() );
    Debug.assert( not self:isSyncing() );

    if LoginMgr:isJune()
        and OnlineMgr:isOnline()
        and self:hasQueuedTransactions()
        and not self:isSyncing() then

        local allValues = {};
        local count = 0;

        --Merge all values
        for _, t in ipairs(self._transactionQueue) do
            local values = t:getValues();

            if type(values) == "table" then

                for k, v in pairs(values) do

                    allValues[k] = v;
                    count        = count + 1;
                end
            end
        end

        self:_syncQueuedTransactions();
        self:_saveTransactionsSoon();
        print("TransactionMgr:_uploadTransactionsToJuneServer transationQ count post = "..tostring(#self._transactionQueue));
        if count > 0 then

            PossessionMgr:uploadToJune( self,
                                    TransactionMgr._onUploadSuccess,
                                    TransactionMgr._onUploadFailure,
                                    allValues
                                );
        else
            self._syncingTransactions = {};
        end
    end
end

function TransactionMgr:_uploadEverythingToServer( )

   print("TransactionMgr:_uploadEverythingToServer");

   Debug.assert( LoginMgr:isTango() );
   Debug.assert( OnlineMgr:isOnline() );
   Debug.assert( not self:isSyncing() );

   if LoginMgr:isTango()
   and OnlineMgr:isOnline()
   and self:isSyncing() then

      self._possessionTime = PossessionMgr:getLastDownloadTime();

      self:_syncQueuedTransactions();

      self:_saveTransactionsSoon();

      PossessionMgr:upload( self,
                            TransactionMgr._onUploadSuccess,
                            TransactionMgr._onUploadFailure,
                            self:_createPossessionUploadTableForAllData()
      );

   end
end


function TransactionMgr:uploadEverythingToJune( )
   print("TransactionMgr:_uploadEverythingToJune isJune: ", tostring(LoginMgr:isJune()), " isTango: ", tostring(LoginMgr:isTango()), " isOnline: ", tostring(OnlineMgr:isOnline()), " isSyncing: ", tostring(self:isSyncing()));
   if LoginMgr:isJune() and not LoginMgr:isTango()
   and OnlineMgr:isOnline()
   and not self:isSyncing() then
        PossessionMgr:uploadToJune( self,
                                    TransactionMgr._onUploadSuccess,
                                    TransactionMgr._onUploadFailure,
                                    self:_createPossessionUploadTableForAllData()
              );
   end
end

function TransactionMgr:_createPossessionUploadTableForAllData()

   local tblResult = {};

   local builtin = App.data:getBuiltinInfo();

   if not builtin then
      return tblResult;
   end

   for _, tbl in ipairs(builtin) do

      if tbl.server and tbl.name then
         tblResult[tbl.name] = App.data:getValue(tbl.name, tbl.default);
      end

   end

   return tblResult;

end

function TransactionMgr:_onUploadSuccess(params, possessions, ...)

   -- print("TransactionMgr:_onUploadSuccess");

   self._possessionOutDated = false;
   self._syncingTransactions = {};
   self:_saveTransactionsSoon();

end

function TransactionMgr:_onUploadFailure(params, outDated, ...)

   -- print("TransactionMgr:_onUploadFailure");

   --outDated means the server has "newer" values than we do
   if outDated then
      self._possessionTime = nil;
      self._possessionOutDated = true;
   end

   -- Something went wrong in the transmission... retry
   self:_requeueSyncingTransactions();

   self._possessionTime = nil;
   self:_saveTransactionsSoon();

end

function TransactionMgr:_onDownloadSuccess(event)

   print("TransactionMgr:_onDownloadSuccess");

   if self._pendingMigration then
      return;
   end

   -- This seems to cause loss of progress and causes trouble for a tiny benefit.
   -- This code does not seem to actually serve its intended purpose (update local data periodically when there's a change on the cloud)
   -- This callback is only fired when ther's a conflict, but that will be taken care of anyway.
   -- THis is just interfering with the actual transaction replay code.

   -- self:_applyPossessionDataOntoLocalData();

end

---------------------------------------
-- PUBLIC --
---------------------------------------

function TransactionMgr:init()

   EventMgr:addListener(self, TransactionMgr._onDownloadSuccess, "possessions-download-success");

end

function TransactionMgr:isSyncing()

   return #self._syncingTransactions > 0;

end

function TransactionMgr:hasQueuedTransactions()

   return #self._transactionQueue > 0;

end

function TransactionMgr:update(dt)

   self:_syncWithServer(dt);

   self:_updateSave(dt);

end

function TransactionMgr:execTransaction(t)

   -- print("TransactionMgr:execTransaction");

   if not Debug.assertType(t, "table", "TransactionMgr:execTransaction") then
      return nil;
   end

   Debug.assertType(App.data, "table", "TransactionMgr:execTransaction -> App.data");

   local values = t:buildValues(App.data);

   if values then

      -- Check the transaction queue. If it's currently empty, we need to update our back up.
      if not self:hasQueuedTransactions() then

         self._dataBackup = App.data:createBackup();
      end
      print("TransactionMgr - just built a transaction, about to queue it. Num already queued:", #self._transactionQueue);

      -- Apply the values to our local data and queue the transaction
      if App.data:setValues( values ) > 0 then

         --call on Success after we change the real data
         t:onSuccess(App.data);

         --If we are not authenticate, there is no need to sync up
         --with the server
         if LoginMgr:isTango() or LoginMgr:isJune() then
            table.insert( self._transactionQueue, t );
         end

         App.data:_saveDataSoon();
      end

      return true;

   end

end

function TransactionMgr:load(appData)

   self._dataBackup = loadLocalFile( self:_getBackupFileName(appData), FILE_LOAD_MAX_ATTEMPT ) or {};

   self:_restoreTransactions(appData);

   -- We just restored transactions, so any syncing transactions are no longer good.
   self:_requeueSyncingTransactions();

end

function TransactionMgr:save(appData)

   Debug.assertType(appData, "table", "TransactionMgr:save -> appData");

   local saveSuccess = saveLocalFile( self:_getBackupFileName(appData), self._dataBackup, FILE_SAVE_MAX_ATTEMPT );

   self:_saveTransactions(appData);

end

function TransactionMgr:flagForcedMigration()
   self._forceDataMigration = true;
end

-- Make the server take whatever we have.
function TransactionMgr:chooseLocalData(autoChose)

   print("TransactionMgr:chooseLocalData");

   if not LoginMgr:isTango() then
      Debug.error("TransactionMgr:chooseLocalData -> User is not connected to Tango!");
      return;
   end

   if not OnlineMgr:isOnline() then
      Debug.error("TransactionMgr:chooseLocalData -> User is not online!");
      return;
   end

   if self:isSyncing() then
      Debug.error("TransactionMgr:chooseLocalData -> TransactionMgr is already syncing!!");
      return;
   end

   local backup = App.data:createBackup();

   self._possessionTime = PossessionMgr:getLastDownloadTime();

   self:_syncQueuedTransactions();

   self:_saveTransactionsSoon();

   PossessionMgr:upload( self,

                         function( self, params, possessions, ... )

                            TransactionMgr:_onUploadSuccess(params, possessions, ...)

                            App.data:setGuestMode(nil);
                            TransactionMgr._pendingMigration = false;

                            if not autoChose then
                               EventMgr:postEvent({data = "local"}, "choose-data-success");
                            end

                         end,

                         function( self, params, outDated, ... )

                            -- Revert to previous backup
                            App.data:restoreFromBackup(backup);

                            self._possessionTime = nil;
                            self:_saveTransactionsSoon();

                            EventMgr:postEvent({}, "data-server-error");
                         end,

                         self:_createPossessionUploadTableForAllData()
   );

end

-- Take the server's values, apply them onto our local data, then upload.
function TransactionMgr:chooseCloudData(autoChose)

   print("TransactionMgr:chooseCloudData");

   if not LoginMgr:isTango() then
      Debug.error("TransactionMgr:chooseLocalData -> User is not connected to Tango!");
      return;
   end

   if not OnlineMgr:isOnline() then
      Debug.error("TransactionMgr:chooseLocalData -> User is not online!");
      return;
   end

   if self:isSyncing() then
      Debug.error("TransactionMgr:chooseLocalData -> TransactionMgr is already syncing!!");
      return;
   end

   local backup = App.data:createBackup();

   self._possessionTime = PossessionMgr:getLastDownloadTime();

   self:_syncQueuedTransactions();

   self:_applyPossessionDataOntoLocalData();

   self:_saveTransactionsSoon();

   PossessionMgr:upload( self,

                         function( self, params, possessions, ... )

                            TransactionMgr:_onUploadSuccess(params, possessions, ...)

                            App.data:setGuestMode(nil);
                            TransactionMgr._pendingMigration = false;

                            if not autoChose then
                               EventMgr:postEvent({data = "cloud"}, "choose-data-success");
                            end

                         end,

                         function( self, params, outDated, ... )

                            -- Revert to previous backup
                            App.data:restoreFromBackup(backup);

                            self._possessionTime = nil;
                            self:_saveTransactionsSoon();

                            EventMgr:postEvent({}, "data-server-error");
                         end,

                         self:_createPossessionUploadTableForAllData()
   );

end
