----------------------------------------------------------------------

local WEEKLY_WARNING             = "The race is almost over. The top 3 racers will earn a sweet prize!";
local WEEKLY_REMINDER            = "The rankings are in! Are you in the top 3?";

local DAILY_WARNING              = "Your Daily Challenge is about to expire. Make sure you complete it before it resets!";
local DAILY_REMINDER             = "Your Daily Challenge is ready. Complete it and get your reward!";

local KEY_REMINDER               = "Your keys are replenished. You are ready to go!";

local HOLIDAY_SALE               = "There's a Holiday Gem Sale going on!";
local HOLIDAY_REMINDER           = "The Holiday Gem Sale is almost over!";

local CAR_SALE                   = "There's a Car Sale going on!";
local CAR_SALE_REMINDER          = "The Car Sale is almost over!";

--Since our push noticiations are silent, I increase the safe range
local SAFE_HOURS_MIN             =  9;
local SAFE_HOURS_MAX             = 22;

local DAILY_WARNING_DELTA_TIME   = 60*60*1.5;   -- 1-1/2 hour
local DAILY_READY_DELTA_TIME     = 60*60*0.25;	-- 15 minutes

local WEEKLY_REMINDER_DELTA_TIME = 60*60*4;     --4 hours

----------------------------------------------------------------------

NotificationMgr  =
{
};

----------------------------------------------------------------------

function NotificationMgr:init()

   --add listeners
   EventMgr:addListener( self, self._dailyCollected,     "daily-reward" );
   EventMgr:addListener( self, self._weeklySeen,         "weekly-reset" );
   EventMgr:addListener( self, self._keyConsumed,        "consume-key-success" );
   EventMgr:addListener( self, self.onCloudMasterData,   "cloud:master-data-notification" );

   if not App.data:getValue( "notification_mgr_active" ) then
      EventMgr:addListener( self, self._masterUpdated,   "cloud:master-updated" );
   end

   self:_cancelAllLocalNotifications();

end

----------------------------------------------------------------

function NotificationMgr:onCloudMasterData( data )

   local content     = (data or {}).content;

   if content then
      self.infoTable = content;
   end
end

----------------------------------------------------------------------

function NotificationMgr:onEnterBackground()

   self:_cancelAllLocalNotifications();
   
   self:_setLocalNotifications();
   
   self.weeklySeen = false;

end

----------------------------------------------------------------------

function NotificationMgr:onEnterForeground()

   self:_cancelAllLocalNotifications();

end


----------------------------------------------------------------------

function NotificationMgr:_localNotificationInSeconds(time, message)

   if MOAINotificationsIOS then
      MOAINotificationsIOS.localNotificationInSeconds( time, message);
   end

   if MOAINotificationsAndroid then
      MOAINotificationsAndroid.localNotificationInSeconds( time, message );
   end

end

----------------------------------------------------------------------

function NotificationMgr:_cancelAllLocalNotifications()

   if MOAINotificationsIOS then
      MOAINotificationsIOS.cancelAllLocalNotifications();
   end

   if MOAINotificationsAndroid then
      MOAINotificationsAndroid.cancelAllLocalNotifications();
   end

end

----------------------------------------------------------------------

function NotificationMgr:_setLocalNotifications( )

   if App.data:getValue( "notification_mgr_active" ) then

      if App.data:getValue( "weekly_notification_active" ) then


         local ww = self:_calcWeeklyWarning( );
         local wr = self:_calcWeeklyReminder( );

         if ww > 0 then
            self:_localNotificationInSeconds(ww, WEEKLY_WARNING);
            EventMgr:postEvent({type="weekly-warning", timer=ww}, "post-notification");
         end

         if wr > 0 then
            self:_localNotificationInSeconds(wr, WEEKLY_REMINDER);
            EventMgr:postEvent({type="weekly-reminder", timer=wr}, "post-notification");
         end
         
         print("weekly warning:", ww, "weekly reminder:", wr);

      end

      if App.data:getValue( "daily_notification_active" ) then
      
         local dw = self:_calcDailyWarning();
         local dr = self:_calcDailyReminder();

         if dw > 0 then
            self:_localNotificationInSeconds(dw, DAILY_WARNING);
            EventMgr:postEvent({type="daily-warning", timer=dw},  "post-notification");
         end

         if dr > 0 then
            self:_localNotificationInSeconds(dr, DAILY_REMINDER);
            EventMgr:postEvent({type="daily-reminder", timer=dr}, "post-notification");
         end
         
         print("daily warning:", dw, "daily reminder:", dr);

      end

      if App.data:getValue( "key_notification_active" ) then
      
         local kr = self:_calcKeyReminder( );

         if kr > 0 then
            self:_localNotificationInSeconds(kr, KEY_REMINDER);
            EventMgr:postEvent({type="key-warning", timer=kr}, "post-notification");
         end

         self.shouldKeyRemind = false;

      end
      
      local hs = self:_calcSaleStart("holiday");
      local hr = self:_calcSaleReminder("holiday");

      if hs and hs > 0 then
         self:_localNotificationInSeconds(hs, HOLIDAY_SALE);
         EventMgr:postEvent({type="holiday-start", timer=hs},  "post-notification");
      end
      if hr and hr > 0 then
         self:_localNotificationInSeconds(hr, HOLIDAY_REMINDER);
         EventMgr:postEvent({type="holiday-reminder", timer=hr},  "post-notification");
      end
      
      local css = self:_calcSaleStart("vehicle");
      local csr = self:_calcSaleReminder("vehicle");

      if css and css > 0 then
         self:_localNotificationInSeconds(css, CAR_SALE);
         EventMgr:postEvent({type="vehicle-start", timer=css},  "post-notification");
      end
      if csr and csr > 0 then
         self:_localNotificationInSeconds(csr, CAR_SALE_REMINDER);
         EventMgr:postEvent({type="vehicle-reminder", timer=csr},  "post-notification");
      end

   end

end

----------------------------------------------------------------------

function NotificationMgr:_calcWeeklyWarning( )

   local date = os.date("*t", os.time())

   if date.wday > 1 and date.wday < 7 then

      -- not saturday timer
      local tD = (7-date.wday) * 24 * 60 * 60;
      local tH = (13-date.hour) * 60 * 60;
      
      return tD + tH - ((date.min*60) + date.sec);

   end
   
   return -1;

end

----------------------------------------------------------------------

function NotificationMgr:_calcWeeklyReminder( )

   local weeklyTimer     = RankingMgr:getTimeToReset()

   if weeklyTimer then

      local wNotifyTime    =  os.time() + weeklyTimer + WEEKLY_REMINDER_DELTA_TIME;
      local wNotifyDate    = os.date("*t", wNotifyTime  );

      if wNotifyDate then

         if wNotifyDate.hour  < SAFE_HOURS_MIN then
            wNotifyDate.hour = SAFE_HOURS_MIN;
            wNotifyDate.min  = 0;

         elseif wNotifyDate.hour  > SAFE_HOURS_MAX then

            -- We have to notify on the next day we cant do it before
            wNotifyDate.day    = wNotifyDate.day + 1;
            wNotifyDate.hour   = SAFE_HOURS_MIN;
            wNotifyDate.min    = 0;
         end

         local wNotifyTime2 = os.time(wNotifyDate);
         local wNotifyTimer = wNotifyTime2 - os.time();

         print("weekly timer:", weeklyTimer, "wNotifyTimer:", wNotifyTimer);

         return wNotifyTimer;

      end
   end


   return -1;

end

----------------------------------------------------------------------

function NotificationMgr:_calcDailyWarning( )

   if App.data:isDailyActive() and App.data:getDailyChallengeStreak() > 0 then

      local dailyTimer     = App.data:getDailyChallengeTimer();

      if dailyTimer then

         local dResetTime     = os.time() + dailyTimer;
         local dNotifyTime    = dResetTime - DAILY_WARNING_DELTA_TIME;
         local dNotifyDate    = os.date("*t", dNotifyTime  );

         if dNotifyDate then

            if dNotifyDate.hour  < SAFE_HOURS_MIN then
               dNotifyDate.hour = SAFE_HOURS_MIN;
               dNotifyDate.min  = 0;
            elseif dNotifyDate.hour  > SAFE_HOURS_MAX then
               dNotifyDate.hour   = SAFE_HOURS_MAX;
               dNotifyDate.min    = 0;
            end

            local dNotifyTime2 = os.time(dNotifyDate);
            if dNotifyTime2 then
               local dNotifyTimer = dNotifyTime2 - os.time();
               if dNotifyTimer then

                  --Because we could have change the time to adjust to safe hours
                  -- we could have a notify timer after the daily is reset
                  if dNotifyTimer > 0 and dNotifyTimer < dailyTimer then
                     return dNotifyTimer;
                  end
               end
            end

         end
      end
   end

   return -1;

end

----------------------------------------------------------------------

function NotificationMgr:_calcDailyReminder( )
   
   if (not App.data:isDailyActive()) and (App.data:getDailyChallengeStreak() > 0) then

      local dailyTimer     = App.data:getDailyChallengeTimer();

      if dailyTimer then

         local dResetTime     = os.time() + dailyTimer;
         local dNotifyTime    = dResetTime + DAILY_READY_DELTA_TIME;
         local dNotifyDate    = os.date("*t", dNotifyTime  );

         if dNotifyDate then

            if dNotifyDate.hour  < SAFE_HOURS_MIN then
               dNotifyDate.hour = SAFE_HOURS_MIN;
               dNotifyDate.min  = 0;
            elseif dNotifyDate.hour  > SAFE_HOURS_MAX then
               -- We have to notify on the next day we cant do it before
               dNotifyDate.day    = dNotifyDate.day + 1;
               dNotifyDate.hour   = SAFE_HOURS_MIN;
               dNotifyDate.min    = 0;
            end
            
            local dNotifyTime2 = os.time(dNotifyDate);
            
            if dNotifyTime2 then
               return dNotifyTime2 - os.time();
            end

         end
      end
   end
   
   return -1;

end

----------------------------------------------------------------------

function NotificationMgr:_calcKeyReminder( )

   local keyCount = App.data:getKeyCount();

   if keyCount <= 2 and (self.shouldKeyRemind or keyCount == 0)  then

     return App.data:calcFullKeyRegenTime() or -1;

   end

   return -1;

end

----------------------------------------------------------------------

function NotificationMgr:_calcSaleStart( type )

   local sale = nil;
   if self.infoTable and self.infoTable[type] and self.infoTable[type].daysToSeek then
      sale = PromoMgr:getFutureSales(type, self.infoTable[type].daysToSeek);
   end

   if sale and sale[1] and sale[1].from then
   
      local shiftTime = 0;
      if sale[1].notifyShift then
         shiftTime = sale[1].notifyShift;
      end

      local hNotifyDate    = os.date("*t", sale[1].from + shiftTime );
      local revisedDate    = nil;

      if hNotifyDate then
      
         --if the hour is greater than safe hours max, then move it forward one day, and use the revised date.
         --this avoids sending the notification when a sale isn't on.

         if hNotifyDate.hour  < SAFE_HOURS_MIN then
            hNotifyDate.hour = SAFE_HOURS_MIN;
            hNotifyDate.min  = 0;
         elseif hNotifyDate.hour  > SAFE_HOURS_MAX then
            --shift it forward one day
            revisedDate = os.date("*t", sale[1].from + shiftTime + 86400 );
            revisedDate.hour   = SAFE_HOURS_MAX;
            revisedDate.min    = 0;
         end
         
         if revisedDate then
            hNotifyDate = revisedDate;
         end

         local hNotifyTime2 = os.time(hNotifyDate);
         local hNotifyTimer = hNotifyTime2 - os.time();
         return hNotifyTimer;

      end

   end


   return -1;

end

----------------------------------------------------------------------

function NotificationMgr:_calcSaleReminder( type )

   local sale = nil;
   if self.infoTable and self.infoTable[type] and self.infoTable[type].daysToSeek then
      sale = PromoMgr:getFutureSales(type, self.infoTable[type].daysToSeek);
   end

   if sale and sale[1] and sale[1].to then
   
      local shiftTime = 0;
      if sale[1].reminderShift then
         shiftTime = sale[1].reminderShift;
      end

      local hNotifyDate    = os.date("*t", sale[1].to + shiftTime );

      if hNotifyDate then

         if hNotifyDate.hour < SAFE_HOURS_MIN then
            hNotifyDate.hour  = SAFE_HOURS_MIN;
            hNotifyDate.min   = 0;
         elseif hNotifyDate.hour > SAFE_HOURS_MAX then
            hNotifyDate.hour  = SAFE_HOURS_MAX;
            hNotifyDate.min   = 0;
         end

         local hNotifyTime2 = os.time(hNotifyDate);
         local hNotifyTimer = hNotifyTime2 - os.time();
         return hNotifyTimer;

      end

   end


   return -1;

end

----------------------------------------------------------------------

function NotificationMgr:_dailyCollected( )

 end
 
----------------------------------------------------------------------

function NotificationMgr:_weeklySeen( )

   self.weeklySeen = true;

end

----------------------------------------------------------------------

function NotificationMgr:_keyConsumed( )

   if App.data:getKeyCount() == 0 then
      --set key reminder available
      self.shouldKeyRemind = true;
   end

end

----------------------------------------------------------------------

function NotificationMgr:_masterUpdated( event )

   --this is a trick to avoid apple rejection
   if event.success then

      if self.infoTable and self.infoTable.info then

         local abTable = AbTestMgr:getNotificationInfo();
         if abTable then
            self.infoTable.info = abTable;
         end

      
         if not App.data:getValue("notification-setup-complete") then
         
            if self.infoTable.info.weekly_enabled then
               App.data:setValue( "weekly_notification_active", true );
            end
            if self.infoTable.info.daily_enabled then
               App.data:setValue( "daily_notification_active", true );
            end
            
            if self.infoTable.info.key_enabled then
               App.data:setValue( "key_notification_active", true );
            end
            
            if not self.infoTable.info.show_guide then
               App.data:setValue( "notification-guide-disabled", true );
            end
            
            App.data:setValue( "notification-setup-complete", true );

         end

      end


      App.data:setValue( "notification_mgr_active", true );
   end
   
end


----------------------------------------------------------------------