require('app/possession/task/possession-task-download')
require('app/possession/task/possession-task-upload')
require('app/server/tango')
----------------------------------------------------------------------

local POSSESSION_MGR_SAVE_FILENAME_LUA = "ROAD_RIOT_POSSESSION_MGR.lua";
local POSSESSION_MGR_SAVE_FILENAME_JSON = "p80os33s1i35o884n.jp5";

----------------------------------------------------------------------

local Base     = TangoServer;

PossessionMgr  = Base:new();

-------------------------------------------------------------------------------------

function PossessionMgr:init()

   assert(not self:isActive());

   self:_reset();
end

-------------------------------------------------------------------------------------

function PossessionMgr:activate()

   if ( not self:isActive() ) then

      self:_reset();

      self:_load();

      --If load failed (most likely, it is our first time)
      if self._possessions ~= nil and self._lastDownloadTime ~= nil then
         self._ready = true;
      else
         self._downloadTimer = .1;
      end

      self._active   = true;

   end
end

-------------------------------------------------------------------------------------

function PossessionMgr:deactivate()

   self:_failUploadAndClear();

   self:_reset();
end

-------------------------------------------------------------------------------------

function PossessionMgr:isActive()
   return self._active;
end

----------------------------------------------------------------------

function PossessionMgr:isReady()
   --As long as we try once to download data, we are ready
   return (not LoginMgr:isTango()) or self._ready;
end

----------------------------------------------------------------------

function PossessionMgr:isDownloading()
   return self._downloadTask ~= nil;
end

----------------------------------------------------------------------

function PossessionMgr:getLastDownloadTime()
   return self._lastDownloadTime;
end

----------------------------------------------------------------------

function PossessionMgr:getDownloadCount()
   return self._downloadCount;
end

----------------------------------------------------------------------

function PossessionMgr:isUploading()
   return self._uploadTask ~= nil;
end

----------------------------------------------------------------------

function PossessionMgr:hasUploadTaskWaiting()
   return #self._uploadTaskList > 0
end

----------------------------------------------------------------------

function PossessionMgr:clone( key  )

   if self._possessions then
      return self._possessions:clonePossession( key );
   end

end

----------------------------------------------------------------------

function PossessionMgr:getPossessions()
   return self._possessions;
end

-------------------------------------------------------------------------------------

function PossessionMgr:isBusy()

   return   (self:isDownloading())
         or (self:isUploading())
         or (self:hasUploadTaskWaiting());

end

----------------------------------------------------------------------

function PossessionMgr:update( dt )

   if not self:isActive() then

      --We are offline... failed transaction right away, no need to send them
      self:_failUploadAndClear();

   else

      if (not LoginMgr:isTango()) then

         self:_failUploadAndClear();

      elseif self:isDownloading() then

         self:_failUploadAndClear();

      elseif ( not self:isUploading() ) then

         if self._downloadTimer then

            self:_failUploadAndClear();

            self._downloadTimer = self._downloadTimer - dt;
            if self._downloadTimer < 0 then
               self:_download();
            end

         else

            self._uploadTimer = (self._uploadTimer or 0) - dt;

            if self:hasUploadTaskWaiting() then

               if not self._possessions then

                  self:_failUploadAndClear();

               else

                  if self._uploadTimer < 0 then
                     self:_nextUploadTask();
                  end

               end

            end
         end
      end
   end

end

----------------------------------------------------------------------

function PossessionMgr:_reset()

   self._active            = false;
   self._ready             = false;
   self._downloadTimer     = nil;
   self._lastDownloadTime  = nil;
   self._downloadCount     = 0;
   self._downloadTask      = nil;
   self._uploadTask        = nil;
   self._uploadTaskList    = {};
   self._uploadTimer       = nil;

end

----------------------------------------------------------------------

function PossessionMgr:_save()

   local tbl = {};
   if self._possessions then

      tbl.lastDownloadTime  = self._lastDownloadTime;
      tbl.possessions       = {};

      local ptable = self._possessions:getPossessionsTable ()
      for k,v in pairs(ptable) do

         local p = self._possessions:getPossession(k);
         if p then

            local entry =
            {
               value          = p:getValue(),
               lastModified   = p:getLastModified(),
               version        = p:getVersion(),
            }
            tbl.possessions[k] = entry;

         end
      end
   end

   saveFileMultiFormat(POSSESSION_MGR_SAVE_FILENAME_LUA, POSSESSION_MGR_SAVE_FILENAME_JSON, tbl);

end

----------------------------------------------------------------------

function PossessionMgr:_load()

   --Should be empty, but just in case
   self:_failUploadAndClear();

   local tbl = loadFileMultiFormat( POSSESSION_MGR_SAVE_FILENAME_LUA, POSSESSION_MGR_SAVE_FILENAME_JSON );

   if tbl then

      self._lastDownloadTime = tbl.lastDownloadTime;

      self._possessions      = MOAITangoPossessions.new();

      if tbl.possessions then

         for k, v in pairs(tbl.possessions) do

            local p = MOAITangoPossession.new();

            p:setValue(v.value);
            p:setLastModified(v.lastModified);
            p:setVersion(v.version);

            self._possessions:setPossession(k, p);

         end
      end

   else

      self._possessions = nil;

   end

end

----------------------------------------------------------------------

function PossessionMgr:_download()

   print("PossessionMgr:_download");

   --Should be empty, but just in case
   self:_failUploadAndClear();

   self._downloadTimer = nil;

   if LoginMgr:isTango() then
     -- if OnlineMgr:isOnline() and LoginMgr:isTango()  then
      self._downloadTask = PossessionTaskDownload:new()
      self._downloadTask:exec();
   else
      self:onDownloadFailure();
   end

end


-------------------------------------------------------------------------------------

function PossessionMgr:onDownloadSuccess( task, possessions )

   MOAIFlurry:event("PossessionMgr-Download", "status", "Download-Success", "online", OnlineMgr:getStatusString() );

   self._ready             = true;
   self._downloadTimer     = nil;
   self._downloadTask      = nil;
   self._possessions       = possessions;
   self._lastDownloadTime  = os.time();
   self._downloadCount     = self._downloadCount + 1;

   -- fail all other transaction because they would be based on old data
   -- we should have failed them before, so this list should be empty
   -- but it is better being safe than sorry
   self:_failUploadAndClear();

   self:_save();

   task:onSuccess( possessions );

   --self:_printoutAllPossessions();
   EventMgr:postEvent({}, "possessions-download-success");

end

-------------------------------------------------------------------------------------

function PossessionMgr:onDownloadFailure(  )

   MOAIFlurry:event("PossessionMgr-Download", "Status", "Download-Failed", "online", OnlineMgr:getStatusString() );

   --Try again in a little bit
   self._downloadTimer     = 30;
   self._downloadCount     = self._downloadCount + 1;
   self._lastDownloadTime  = nil;
   self._possessions       = nil;

   self:_failUploadAndClear();
end

----------------------------------------------------------------------

function PossessionMgr:upload( callObj, callSuccess, callFailure, params )

   print("PossessionMgr:upload");

   local task = PossessionTaskUpload:new( callObj, callSuccess, callFailure, params );

   self:_addTask( task );
end

----------------------------------------------------------------------

function PossessionMgr:uploadToJune( callObj, callSuccess, callFailure, params )
   print("PossessionMgr:uploadToJune");
   self:_failUploadAndClear();
   local task = PossessionTaskUpload:new( callObj, callSuccess, callFailure, params );
   task.uploadToJuneOnly = true;
    self._uploadTask = task;
      if self._uploadTask then
         self._uploadTask:exec();
      end
end

----------------------------------------------------------------------

function PossessionMgr:_nextUploadTask()

   assert( self._uploadTask == nil );

   if (#self._uploadTaskList > 0) then

      self._uploadTask = self._uploadTaskList[1];
      table.remove(self._uploadTaskList,1);

      if self._uploadTask then
         self._uploadTask:exec();
      end

      return true;

   end

end

-------------------------------------------------------------------------------------

function PossessionMgr:onUploadSuccess( task, possessions )

   MOAIFlurry:event("PossessionMgr-Upload", "Status", "Upload-Success", "online", OnlineMgr:getStatusString() );

   --Make sure to upload all possessions (it is mainly for the version)
   if possessions and self._possessions then
      local ptable = possessions:getPossessionsTable ()
      for k,v in pairs(ptable) do
         self._possessions:setPossession( k, possessions:getPossession(k) );
      end
   end

   --This will make us upload data in a little while (to avoid sending too much data)
   self._uploadTimer    = 10;
   self._uploadTask     = nil;

   self:_save();

   task:onSuccess(possessions);

end

-------------------------------------------------------------------------------------

function PossessionMgr:onUploadOutOfDate( task )

   MOAIFlurry:event("PossessionMgr-Upload", "Status", "Upload-OutOfDate", "online", OnlineMgr:getStatusString() );

   -- Our possessions are out dated... no need to keep them anymore
   -- force download
   self._downloadTimer   = 1;

   self:_failUploadAndClear(true);

end

-------------------------------------------------------------------------------------

function PossessionMgr:onUploadFailure( task, code, message )

   MOAIFlurry:event("PossessionMgr-Upload", "Status", "Upload-Failure", "online", OnlineMgr:getStatusString());

   self:_failUploadAndClear();

   -- Dont change our state... it failed but we can still
   -- try to upload... it is up to the user to do so
end

-------------------------------------------------------------------------------------

function PossessionMgr:_failUploadAndClear( outDated )

   self._uploadTimer = 2;

   if #self._uploadTaskList > 0 or self._uploadTask then

      --Backup old task to avoid failure callback adding to the list to get failed
      local oldTaskList    = self._uploadTaskList;
      local curTask        = self._uploadTask;

      self._uploadTaskList = {};
      self._uploadTask     = nil;

      if curTask then
         curTask:onFailure(outDated);
      end

      for i=1, #oldTaskList do
         oldTaskList[i]:onFailure(outDated);
      end

   end

end

-------------------------------------------------------------------------------------

function PossessionMgr:_addTask( task )
   table.insert( self._uploadTaskList, task );
end

----------------------------------------------------------------------

function PossessionMgr:_printoutAllPossessions()

   print("_printoutAllPossessions(possessions)", self._possessions);

   if self._possessions then
      local ptable = self._possessions:getPossessionsTable();
      for k,v in pairs(ptable) do
         print( 'Key=' .. tostring(k) .. ' Value=' .. tostring(v));
      end
   end
end

-------------------------------------------------------------------------------------
