
ProfilerMgr = {};

----------------------------------------------------------------------

function ProfilerMgr:init()
    self.lastMemoryPrintTime = 0;
    self.peakMemoryUsage = {};
    self.autoPrintEnabled = false;
    self.recordedTimestamps = {};
    self.recordedMemSnaps = {};
end

----------------------------------------------------------------------

function ProfilerMgr:update(dt)
    assert(dt >= 0);
    
    local curTime = MOAISim.getElapsedTime();

    if self.autoPrintEnabled and curTime >= self.lastMemoryPrintTime + 1 then

        self.lastMemoryPrintTime = curTime;

        self:_updatePeakMemoryUsage();
        self:printMemoryUsage();

    end
end

----------------------------------------------------------------------

function ProfilerMgr:_updatePeakMemoryUsage()
    local tbl = MOAISim.getMemoryUsage();

    for k, v in pairs(tbl) do
        if self.peakMemoryUsage[k] then
            if self.peakMemoryUsage[k] < v then
                self.peakMemoryUsage[k] = v;
            end
        else
            self.peakMemoryUsage[k] = v;
        end
    end
end

----------------------------------------------------------------------

-- Returns memory count in MB
function ProfilerMgr:_getPrettyMemCount(memcount)

    if not memcount then
        return -1;
    end

    return math.floor( (memcount / (1024*1024) ) * 100) / 100;
end

----------------------------------------------------------------------

function ProfilerMgr:printMemoryUsage(memTbl)
    memTbl = memTbl or MOAISim.getMemoryUsage();

    print("CURRENT: System:"  .. self:_getPrettyMemCount(memTbl["_sys_rss"]),
        "Total:"   .. self:_getPrettyMemCount(memTbl["total"]),
        "Heap:"    .. self:_getPrettyMemCount(memTbl["code"]),
        "Lua:"     .. self:_getPrettyMemCount(memTbl["lua"]),
        "Texture:" .. self:_getPrettyMemCount(memTbl["texture"]),
        "Font:"    .. self:_getPrettyMemCount(memTbl["font"]),
        "Audio:"   .. self:_getPrettyMemCount(memTbl["audio"]));
        
    print("PEAKS:   System:"  .. self:_getPrettyMemCount(self.peakMemoryUsage["_sys_rss"]),
        "Total:"   .. self:_getPrettyMemCount(self.peakMemoryUsage["total"]),
        "Heap:" .. self:_getPrettyMemCount(self.peakMemoryUsage["code"]),
        "Lua:" .. self:_getPrettyMemCount(self.peakMemoryUsage["lua"]),
        "Texture:" .. self:_getPrettyMemCount(self.peakMemoryUsage["texture"]),
        "Font:"    .. self:_getPrettyMemCount(self.peakMemoryUsage["font"]),
        "Audio:" .. self:_getPrettyMemCount(self.peakMemoryUsage["audio"]));
end

----------------------------------------------------------------------

function ProfilerMgr:getTotalMemoryUsed(memTbl)
   memTbl = memTbl or MOAISim.getMemoryUsage();
   
   return self:_getPrettyMemCount(memTbl["total"])
end

----------------------------------------------------------------------

function ProfilerMgr:beginTimestamp(name)
   self.recordedTimestamps[name] = MOAISim.getDeviceTime();
end

----------------------------------------------------------------------

-- Returns the time taken in seconds, or -1 if an error occurred.
function ProfilerMgr:endTimestamp(name, noPrint)
   if self.recordedTimestamps[name] then
      local timeDiff = MOAISim.getDeviceTime() - self.recordedTimestamps[name];

      if not noPrint then
         print(name .. " FINISHED. Time taken: " .. timeDiff .. " seconds");
      end

      self.recordedTimestamps[name] = nil;

      return timeDiff;
   else
      print("WARNING: Called ProfilerMgr:endTimestamp(" .. name .. "), but matching beginTimestamp was never called.");
      return -1;
   end
end

----------------------------------------------------------------------

function ProfilerMgr:beginMemSnapshot(name)
   local memTbl = MOAISim.getMemoryUsage();
   self.recordedMemSnaps[name] = self:_getPrettyMemCount(memTbl["total"]);
end

----------------------------------------------------------------------

function ProfilerMgr:endMemSnapshot(name, noPrint)
   if self.recordedMemSnaps[name] then
      local memTbl = MOAISim.getMemoryUsage();
      local memUsage = self:_getPrettyMemCount(memTbl["total"]);
      local memDiff = memUsage - self.recordedMemSnaps[name];

      if not noPrint then
	 print(name .. " FINISHED. Memory usage change: " .. memDiff .. " MB");
      end

      self.recordedMemSnaps[name] = nil;

      return memDiff;
   else
      print("WARNING: Called ProfilerMgr:endMemSnapshot(" .. name .. "), but matching beginMemSnapshot was never called.");
      return 0;
   end
end

----------------------------------------------------------------------
