----------------------------------------------------------------------
require('app/time/time-manager')
require('engine/object/object')
require('app/store/server/purchase-validator')
require('engine/json/json')

----------------------------------------------------------------------

local Base                                               = Object;

StoreController                                          = Base:new();

StoreController.REQUEST_FAILURE_NONE                      =  0; -- Unused.
StoreController.REQUEST_FAILURE_ALREADY_PURCHASE          =  1; -- Used when attempting to purchase an item the player already purchased. (Android)
StoreController.REQUEST_FAILURE_PURCHASE_PRODUCT_FAILED   =  2; -- Used when a purchase fails. (Android)
StoreController.REQUEST_FAILURE_IAP_NOT_SUPPORTED         =  3; -- Unused.
StoreController.REQUEST_FAILURE_BILLING_FAILED            =  4;
StoreController.REQUEST_FAILURE_CONSUME_FAILED            =  5;
StoreController.REQUEST_FAILURE_REFUND                    =  6;
StoreController.REQUEST_FAILURE_CANCELED                  =  7;
StoreController.REQUEST_FAILURE_FAILED                    =  8;
StoreController.REQUEST_FAILURE_INVALID_INFO              =  9;
StoreController.REQUEST_FAILURE_RECEIPT_DECLINED          = 10;
StoreController.REQUEST_FAILURE_RECEIPT_CHECK_FAILED      = 11;

----------------------------------------------------------------------

function StoreController:new( storeMgr )

   local ctrl        = Base.new(self);

   ctrl._storeMgr    = storeMgr;
   ctrl._requests    = {}
   ctrl._decoder     = require('json');

   return ctrl;

end

----------------------------------------------------------------------

function StoreController:init()

   self:_loadRequests();

end

----------------------------------------------------------------------

function StoreController:update( dt )


end

----------------------------------------------------------------

function StoreController:buy( id, needReceipt, callObj, callSuccess, callFailure )

   local success = false;

   if not Debug.assertType(id, "string", "StoreController:buy -> id") then
      return success;
   end

   if not Debug.assertType(callSuccess, "function", "StoreController:buy -> callSuccess") then
      return success;
   end

   if not Debug.assertType(callFailure, "function", "StoreController:buy -> callFailure") then
      return success;
   end

   if (not self._currRequest) and self:canMakePayments() then

      print("StoreController:Buy - (not self._currRequest) and self:canMakePayments()");
      local request = self:_createRequest( id, needReceipt, callObj, callSuccess, callFailure );

      if request then
      print("StoreController:Buy - request ok. attempting request");

         success = self:_requestAttempt(request);

         if success then
      print("StoreController:Buy - request attempt ok. attempting makePurchase");

            --Delay a little to let the UI show the spinning wheel first
            delay( 0.2,
                   function()
                     self:_makePurchase( request );
                  end                                    );

         end

      else

         Debug.assert( request, "StoreController:buy -> self._currRequest");

      end

   else

      Debug.assert( (not self._currRequest),    "StoreController:buy -> (not self._currRequest)"   );

   end

   return success;

end

----------------------------------------------------------------

function StoreController:_loadRequests()

   self._requests = App.data:getValue("StoreController.requestTable",  {} );

end

----------------------------------------------------------------

function StoreController:_saveRequests()

   local request = {};

   if self._requests then

      for k,v in pairs(self._requests) do
         --Do it manually since we dont want to copy thing like callObj and callSuccess and callFailure
         request[k] =
         {
            requestId      = v.requestId, --aka payload
            id             = v.id,
            productId      = v.productId,
            status         = v.status,
            needReceipt    = v.needReceipt,
            purchased      = v.purchased,
            consumed       = v.consumed,
            verified       = v.verified,
            rewarded       = v.rewarded,
            done           = v.done,
            failCount      = v.failCount,
         };

         --Because we do not want to carried too much data,
         -- only save purchase info if needed to
         if (v.purchased) and (not v.rewarded) then
            request[k].purchaseData   = v.purchaseData;
            request[k].receipt        = v.receipt;
            request[k].signature      = v.signature;
         end

      end

   end

   App.data:setValue("StoreController.requestTable",  request);

end

----------------------------------------------------------------

function StoreController:_createRequest( id, needReceipt, callObj, callSuccess, callFailure )

   if not Debug.assertType(id, "string", "StoreController:_createRequest -> id") then
      return;
   end

   if self._storeMgr then

      local requestId
      repeat
         requestId = MOAIEnvironment.generateGUID ()
      until self._requests [ requestId ] == nil

      local request =   {
                           requestId      = requestId;
                           id             = id,
                           productId      = self._storeMgr:getStoreProductId(id),
                           needReceipt    = needReceipt,
                           status         = "created",
                           timeStamp      = TimeMgr:getTimeGMT(),
                           callObj        = callObj,
                           callSuccess    = callSuccess,
                           callFailure    = callFailure,
                        };

      self._requests [ requestId ] = request;

      self:_saveRequests();

      return request;

   else

      Debug.assert( self._storeMgr, "StoreController:_createRequest -> self._storeMgr" );

   end

end

----------------------------------------------------------------

function StoreController:_changeRequestId( request, requestId )

   if request and requestId and requestId ~= request.requestId then

      local oldRequest = self:_getRequestById( requestId );

      Debug.assert( oldRequest == nil, "StoreController:_changeRequestId -> oldRequest == nil" );

      if request.requestId then
         self._requests [ requestId ]  = nil;
      end

      request.requestId             = requestId;
      self._requests [ requestId ]  = request;

      self:_saveRequests();

   else

      Debug.assert( request,                    "StoreController:_changeRequestId -> request" );
      Debug.assert( requestId,                  "StoreController:_changeRequestId -> requestId" );

   end

end

----------------------------------------------------------------

function StoreController:_setRequestValue( request, key, value, save )

   if request and key then

      request[key] = value;

      if save then
         self:_saveRequests();
      end

   else

      Debug.assert( request,  "StoreController:_setRequestValue -> request"  );
      Debug.assert( key,      "StoreController:_setRequestValue -> key"     );

   end

end

----------------------------------------------------------------

function StoreController:_getRequestById( requestId )

   if requestId then

      return self._requests [ requestId ];

   else

      Debug.assert( requestId, "StoreController:_getRequestById -> requestId" );

   end
end

----------------------------------------------------------------

function StoreController:_validatePurchase( request, receipt, signature )

   if request then

      self:_setRequestValue(request, "receipt",    receipt           );
      self:_setRequestValue(request, "signature",  signature,  true  );

      if ( not request.needReceipt ) then

         self:_requestSucceed( request );

      else

         if request.productId and request.receipt and request.signature then

            self:_setRequestValue( request, "status", "validating", true );

            request._validateAttempt = 0;
            request._purchaseValidator = ValidatePurchasedItem:new(  request.productId,
                                                                     request.id,
                                                                     request.receipt,
                                                                     request.signature,
                                                                     nil,
                                                                  function( _ , response )
                                                                     if response then
                                                                        if response["Valid"] == true then
                                                                           self:_setRequestValue(request,  "verified", true, true );
                                                                           self:_requestSucceed( request );
                                                                        else
                                                                           self:_requestFailed( request, StoreController.REQUEST_FAILURE_RECEIPT_DECLINED, "receipt verification declined" );
                                                                        end
                                                                     else
                                                                        Debug.assert( response, "StoreController:_validatePurchase -> response" );
                                                                        self:_requestFailed( request, StoreController.REQUEST_FAILURE_RECEIPT_CHECK_FAILED, "receipt verification - no response"  );
                                                                     end
                                                                  end,
                                                                  function( _ , code, message )
                                                                     request._validateAttempt = request._validateAttempt + 1;

                                                                     if request._validateAttempt <= 5 then
                                                                        request._purchaseValidator:exec()
                                                                     else
                                                                        self:_requestFailed( request, StoreController.REQUEST_FAILURE_RECEIPT_CHECK_FAILED, "receipt verification failed", true );
                                                                     end
                                                                  end
                                                                  );

            request._purchaseValidator:exec();

         else

            self:_requestFailed( request, StoreController.REQUEST_FAILURE_RECEIPT_CHECK_FAILED, "receipt verification - invalid parameters"  );

            Debug.assert(request.productId,  "StoreController:_validatePurchase -> request.productId" );
            Debug.assert(request.receipt,    "StoreController:_validatePurchase -> request.receipt"   );
            Debug.assert(request.signature,  "StoreController:_validatePurchase -> request.signature" );

         end

      end

   else

      Debug.assert( requestId, "StoreController:_validatePurchase -> request" );

   end

end


----------------------------------------------------------------

function StoreController:_requestAttempt(request)

   if request and request.id and self._storeMgr and self._currRequest == nil then

      local amount = self._storeMgr:getProductAmount(request.id);
      local cost   = self._storeMgr:getProductCost(request.id);

      self:_setRequestValue( request, "status", "purchasing", true );

      EventMgr:postEvent( {
                           id             = request.id,
                           requestId      = request.requestId,
                           productId      = request.productId,
                           type           = "gems",
                           cost           = (cost or {}).dollar,
                           amount         = amount,
                           },
                           "IAP.attempt"
                        );

      self._currRequest = request;

      return true;

   else

      Debug.assert(self._storeMgr,        "StoreController:_requestAttempt -> self._storeMgr"      );
      Debug.assert(request,               "StoreController:_requestAttempt -> request"             );
      Debug.assert((request or {}).id,    "StoreController:_requestAttempt -> (request or {}).id"  );

   end

end

----------------------------------------------------------------

function StoreController:_requestSucceed( request, recover )

   if request and request.id and self._storeMgr then

      local gemBefore   = App.data:getGemCount();
      local amountTable = self._storeMgr:getProductAmount(request.id);
      local cost        = self._storeMgr:getProductCost(request.id);
      local recovered   = request.callSuccess == nil or recover;

      if amountTable.gem and amountTable.gem > 0 then
         App.data:addGem(  amountTable.gem, "store" );
      end

      if amountTable.gold and amountTable.gold > 0 then
         App.data:addGold(  amountTable.gold, "store" );
      end

      -- Add new currencies/items here in the future
      -- TODO Move the currency giving code elsewhere

      self:_setRequestValue( request, "status",    "success"         );
      self:_setRequestValue( request, "rewarded",  true              );
      self:_setRequestValue( request, "done",      true,       true  );

      EventMgr:postEvent( {
                              type           = "gems",
                              id             = request.id,
                              requestId      = request.requestId,
                              productId      = request.productId,
                              verified       = request.verified,
                              cost           = (cost or {}).dollar,
                              gemAmount      = amountTable.gem,
                              numberPurchased= App.data:deltaValue("number-purchase",1),
                              gemBefore      = gemBefore,
                              recovered      = recovered,
                           },
                           "IAP.success" );


      if request.verified  then
         EventMgr:postEvent(  {
                                 id          = request.id,
                                 requestId   = request.requestId,
                                 productId   = request.productId,
                              },
                              "iap-verification-accepted");
      end

      if request == self._currRequest then
         self._currRequest = nil;
      end

      if request.callSuccess then

         request.callSuccess( request.callObj );
         request.callSuccess = nil;

      end

   else

      Debug.assert(self._storeMgr,        "StoreController:_requestSucceed -> self._storeMgr"      );
      Debug.assert(request,               "StoreController:_requestSucceed -> request"             );
      Debug.assert((request or {}).id,    "StoreController:_requestSucceed -> (request or {}).id"  );


   end

end

----------------------------------------------------------------

function StoreController:_requestFailed( request, code, message, again )

   if request then

      local amountTable = self._storeMgr:getProductAmount(request.id);
      local cost        = self._storeMgr:getProductCost(request.id);

      request.failCount = (request.failCount or 0) + 1;

      self:_setRequestValue( request, "status", "failed");
      self:_setRequestValue( request, "done",   true,    true);

      EventMgr:postEvent(
                           {
                              id             = request.id,
                              requestId      = request.requestId,
                              productId      = request.productId,
                              gemAmount      = amountTable.gem,
                              cost           = (cost or {}).dollar,
                              code           = code,
                              reason         = message,
                              failureReason  = request.failureReason,
                           },
                           "IAP.failure"
                        );


      if code == StoreController.REQUEST_FAILURE_RECEIPT_DECLINED then
         EventMgr:postEvent({}, "iap-verification-declined");
      elseif code == StoreController.REQUEST_FAILURE_RECEIPT_CHECK_FAILED  then
         EventMgr:postEvent({}, "iap-verification-failure");
      end

      if request.callFailure then
         request.callFailure(request.callObj, code, message);
         request.callFailure = nil;
      end

      if request == self._currRequest then
         self._currRequest = nil;
      end

   else

      Debug.assert(request, "StoreController:_requestFailed -> request" );

   end

end

----------------------------------------------------------------

function StoreController:_decode( data )

   if data and self._decoder then

      return self._decoder.decode(data);

   else

      Debug.assert( data,           "StoreController:_decode -> data"         );
      Debug.assert( self._decoder,  "StoreController:_decode -> self._decoder");

   end

end

----------------------------------------------------------------

function StoreController:_encode( data )

   if data and self._decoder then

      return self._decoder.encode(data);

   else

      Debug.assert( data,           "StoreController:_encode -> data"         );
      Debug.assert( self._decoder,  "StoreController:_encode -> self._decoder");

   end

end

----------------------------------------------------------------

function StoreController:onEnterBackground()

end

----------------------------------------------------------------

function StoreController:onEnterForeground()

end

----------------------------------------------------------------

function StoreController:isBusy()

   return self._currRequest;

end

----------------------------------------------------------------
