require('engine/audio/audio-layer');
require('engine/audio/audio-music');
require('engine/audio/audio-sound');

----------------------------------------------------------------

local HACK_VOLUME_SCALE = 0.5;

----------------------------------------------------------------

AudioMgr =
{
   muted          = false,
   globalLevel    = 1,
   soundLevel     = 1,
   musicLevel     = 1,
   soundList      = {},
   musicList      = {},
   cacheList      = {},
   updateList     = {},
   layerMap       = {},
   layerList      = {},
};

----------------------------------------------------------------

function AudioMgr:init( global, globalSoundLevel, globalMusicLevel )

   MOAIUntzSystem.initialize();

   self.globalLevel =  global or 1;

   self:setGlobalSoundVolume( globalSoundLevel or 1 );

   self:setGlobalMusicVolume( globalMusicLevel or 1 );
   
end

----------------------------------------------------------------

function AudioMgr:update(dt)

   --Decrease in case we remove to the list
   for i=#self.updateList,1,-1 do
   
      self.updateList[i]:update(dt);
      
   end
   
end

----------------------------------------------------------------

function AudioMgr:findInList( object, list )

   for i=1,#list do
   
      local obj = list[i];
      
      if obj == object then
         return i;
      end
      
   end
   
   return -1;
   
end

----------------------------------------------------------------

function AudioMgr:clearCache(  )

   self.cacheList = {};

   AudioObject:clearCache(  )

end

----------------------------------------------------------------

function AudioMgr:cacheSound( filename, memory, loop, priority )

   if self.cacheList[filename] == nil then
      self.cacheList[filename] = AudioSound:new( filename, memory, loop, priority );
   end

end

----------------------------------------------------------------

function AudioMgr:addToUpdate( object )

   if self:findInList( object, self.updateList ) <= 0 then
   
      table.insert( self.updateList, object );

   end
   
end

----------------------------------------------------------------

function AudioMgr:rmvFromUpdate( object )

   local idx = self:findInList( object, self.updateList );
   
   if idx > 0 then
   
      table.remove( self.updateList, idx );

   end
end

----------------------------------------------------------------

function AudioMgr:mute()

   if not self.muted then

      self.muted = true;
      
      self:_UpdateSoundLevel();
      self:_UpdateMusicLevel();
   
   end
end

----------------------------------------------------------------

function AudioMgr:unMute()

   if self.muted then

      self.muted = false;
      
      self:_UpdateSoundLevel();
      self:_UpdateMusicLevel();
   
   end
end

----------------------------------------------------------------

function AudioMgr:setGlobalVolume( v )

   self.globalLevel = v;

   self:_UpdateSoundLevel();
   self:_UpdateMusicLevel();

end

----------------------------------------------------------------

function AudioMgr:setGlobalMusicVolume( v )

   self.musicLevel = v;

   self:_UpdateMusicLevel();

end

----------------------------------------------------------------

function AudioMgr:setGlobalSoundVolume( v )

   self.soundLevel = v;

   self:_UpdateSoundLevel();

end

----------------------------------------------------------------

function AudioMgr:_UpdateSoundLevel()

   local v = self:_calcSoundVolume();
   
   for i=1,#self.soundList do
   
      local sound = self.soundList[i];
      sound:setParentVolume( v );

   end
   
   for i=1,#self.layerList do
   
      local layer = self.layerList[i];
      layer:setSoundVolume( v );

   end
end

----------------------------------------------------------------

function AudioMgr:_UpdateMusicLevel()

   local v = self:_calcMusicVolume();
   
   for i=1,#self.musicList do
   
      local music = self.musicList[i];
      music:setParentVolume( v );

   end
   
   for i=1,#self.layerList do
   
      local layer = self.layerList[i];
      layer:setMusicVolume( v );

   end
end

----------------------------------------------------------------

function AudioMgr:getLayer( key )

   local layer = self.layerMap[key];

   if layer == nil then
   
      layer  = AudioLayer:new();
      
      layer:setMusicVolume( self:_calcMusicVolume() );
      layer:setSoundVolume( self:_calcSoundVolume() );

      self.layerMap[key] = layer;

      table.insert( self.layerList, layer);
      
   end
   
   return layer;

end

----------------------------------------------------------------

function AudioMgr:addSound( filename, memory, loop, priority )

   local sound = AudioSound:new( filename, memory, loop, priority );
   
   sound:setParentVolume( self:_calcSoundVolume() );

   table.insert( self.soundList, sound );

   return sound;
   
end

----------------------------------------------------------------

function AudioMgr:addMusic( filename, memory, loop )

   local music = AudioMusic:new( filename, memory, loop );
   
   music:setParentVolume( self:_calcMusicVolume() );

   table.insert( self.musicList, music );

   return music;
   
end

----------------------------------------------------------------

function AudioMgr:_calcSoundVolume() 

   if self.muted then
      return 0;
   else
      return self.soundLevel * self.globalLevel * HACK_VOLUME_SCALE;
   end
end

----------------------------------------------------------------

function AudioMgr:_calcMusicVolume() 

   if self.muted then
      return 0;
   else
      return self.musicLevel * self.globalLevel * HACK_VOLUME_SCALE;
   end
end

----------------------------------------------------------------
