require('engine/object/object')

----------------------------------------------------------------

AUDIO_STATE_PLAYING  = 1;
AUDIO_STATE_PAUSING  = 2;
AUDIO_STATE_PAUSED   = 3;
AUDIO_STATE_STOPING  = 4;
AUDIO_STATE_STOPPED  = 5;
AUDIO_STATE_DONE     = 6;

AUDIO_TYPE_SOUND     = 1;
AUDIO_TYPE_MUSIC     = 2;
AUDIO_TYPE_LAYER     = 3;

----------------------------------------------------------------

local UNTZ_ENABLE    = true;

----------------------------------------------------------------

local Base  = Object;
AudioObject = Base:new( );

AudioObject.untzPool = {};

----------------------------------------------------------------

function AudioObject:new( filename, type, memory, loop, priority )

   local s = Base.new( self );

   s.filename        = filename;
   --because of lua logic, if we do memory or true, it will always be true
   s.memory          = memory or false;
   s.type            = type;
   s.priority        = priority;
   s.state           = AUDIO_STATE_STOPPED;
   s.looping         = loop or false;
   s.myVolume        = 0;
   s.seekData        = nil;
   s.parentVolume    = 1;
   s.parent          = nil;
   s.childList       = {};
   s.untz            = nil;

   --Not every object will be assosiate with a "real" sound (i.e. layer)
   if s.filename then
      s.untz = self:loadUntz( s.filename, s.memory );
   end

   s:setLooping( loop or false );
   s:setMyVolume( s.myVolume );

   return s;

end


----------------------------------------------------------------

function AudioObject:clearCache()

   self.untzPool = {};

end

----------------------------------------------------------------

function AudioObject:loadUntz( filename, memory  )

   local untz = nil;

   if UNTZ_ENABLE then

      local list = AudioObject.untzPool[filename];

      if list and #list > 0 then
         untz = list[1];
         table.remove(list, 1);
      else

         untz = MOAIUntzSound.new();

         if MOAIEnvironment.obbDirectory and MOAIFileSystem.checkFileExists(createFullFilename(MOAIEnvironment.obbDirectory, filename)) then

            untz:load( createFullFilename(MOAIEnvironment.obbDirectory, filename), memory );

         elseif MOAIFileSystem.checkFileExists(filename) then

            untz:load( filename, memory );

         else
            Debug.error("Attempted to load missing audio file: " .. tostring(filename));
            return nil;
         end

      end
   end

   return untz;

end

----------------------------------------------------------------

function AudioObject:freeUntz( untz, filename, memory  )

   if untz then
      list  = AudioObject.untzPool[filename];
      if list == nil then
         list = {};
         AudioObject.untzPool[filename] = list;
      end
      table.insert(list, untz);
   end

end

----------------------------------------------------------------

function AudioObject:setMyVolume( v )

   self.seekData       = nil;
   self:_setMyVolume( v );

end

----------------------------------------------------------------

function AudioObject:_setMyVolume( v )

   self.myVolume = v;
   self:_updateVolume();

end

----------------------------------------------------------------

function AudioObject:getMyVolume( )

   return self.myVolume;

end

----------------------------------------------------------------

function AudioObject:setParentVolume( v )

   self.parentVolume     = v;
   self:_updateVolume();

end

----------------------------------------------------------------

function AudioObject:getParentVolume( )

   return self.parentVolume;

end

----------------------------------------------------------------

function AudioObject:getFinalVolume( )

   return self.parentVolume*self.myVolume;

end

----------------------------------------------------------------

function AudioObject:_updateVolume()

   if self.untz then

      --it seems most of the "range" is at the bottom part
      local v = self:getFinalVolume();
      self.untz:setVolume( v*v );
   end

end

----------------------------------------------------------------

function AudioObject:setLooping( loop )

   self.looping = loop;

   if self.untz then
      self.untz:setLooping(loop);
   end

   for i=1,#self.childList do
      self.childList[i]:setLooping( loop );
   end

   --Monitor when the sound is done
   if self:isPlaying() and ( not loop ) then

      AudioMgr:addToUpdate( self );

   end
end

----------------------------------------------------------------

function AudioObject:isLooping( )

   return self.looping;

end

----------------------------------------------------------------

function AudioObject:play( fadeTime, volume )

   if self.state ~= AUDIO_STATE_PLAYING then

      --Not every object will be assosiate with a "real" sound (i.e. layer)
      if (self.untz == nil) and ( self.filename ~= nil ) then
         self.untz = self:loadUntz( self.filename, self.memory );
      end

      volume   = volume   or 1;
      fadeTime = fadeTime or 0;

      self:_setState( AUDIO_STATE_PLAYING );

      if fadeTime > 0 then
         self:seekMyVolume( volume, fadeTime );
      else
         self:setMyVolume( volume );

         --monitor being done
         if  (not self:isLooping()) then
            AudioMgr:addToUpdate( self );
         end
      end

      if self.untz then
         self.untz:play();
      elseif self.filename then
         self.untz = self:loadUntz( self.filename, self.memory );
         if self.untz then
            self.untz:play();
         end
      end

      for i=1,#self.childList do
         --no need to fade for our childList... they will use ours
         self.childList[i]:play();
      end

   end
end

----------------------------------------------------------------

function AudioObject:isPlaying( )

   return self.state == AUDIO_STATE_PLAYING;

end

----------------------------------------------------------------

function AudioObject:isDone()

   return self.state == AUDIO_STATE_DONE;

end

----------------------------------------------------------------

function AudioObject:_done()

   self:setMyVolume(0);

   self.state = AUDIO_STATE_DONE;

   AudioMgr:rmvFromUpdate( self );

   if self.parent then
      self.parent:onChildAudioDone( self );
   end

   self.untz = self:freeUntz(self.untz,self.filename, self.memory );
   --dont do anything about childList... we should have none anyway

end

----------------------------------------------------------------

function AudioObject:pause( fadeTime )

   if fadeTime and fadeTime > 0 and  self.state == AUDIO_STATE_PLAYING then

      self:_setState( AUDIO_STATE_PAUSING );

      self:seekMyVolume( 0, fadeTime );

   else

      self:_pause();

   end
end

----------------------------------------------------------------

function AudioObject:isPaused( )

   return self.state == AUDIO_STATE_PAUSED;

end

----------------------------------------------------------------

function AudioObject:isPausing( )

   return self.state == AUDIO_STATE_PAUSING;

end

----------------------------------------------------------------

function AudioObject:_pause()

   self:_setState( AUDIO_STATE_PAUSED );

   if self.untz then
      self.untz:pause();
   end

   self:setMyVolume( 0 );

   for i=1,#self.childList do
      --no need to fade in our childList... they will use ours
      self.childList[i]:_pause();
   end

   AudioMgr:rmvFromUpdate( self );

end

----------------------------------------------------------------

function AudioObject:stop( fadeTime )

   if self.state ~= AUDIO_STATE_STOPING then

      -- stop and fade even if we are done because of our children
      if     fadeTime and fadeTime > 0
         and self.state ~= AUDIO_STATE_STOPPED  then

         self:_setState( AUDIO_STATE_STOPING );

         self:seekMyVolume( 0, fadeTime );

      else

         self:_stop();

      end
   end
end

----------------------------------------------------------------

function AudioObject:isStoping( )

   return self.state == AUDIO_STATE_STOPING;

end

----------------------------------------------------------------

function AudioObject:_stop()

   self:_setState( AUDIO_STATE_STOPPED );

   self:setMyVolume( 0 );

   if self.untz then
      self.untz:stop();
   end


   for i=1,#self.childList do
      --no need to fade in our childList... they will use ours
      self.childList[i]:_stop();
   end

   AudioMgr:rmvFromUpdate( self );

   --we mark it done to free the memory, once the user stop us
   --if they play again the sound will be in cache anyway
   self:_done();

end

----------------------------------------------------------------

function AudioObject:isStopped( )

   return self.state == AUDIO_STATE_STOPPED;

end

----------------------------------------------------------------

function AudioObject:_setState( state )

   if self.state ~= state then
      self.state        = state;
   end

end

----------------------------------------------------------------

function AudioObject:update( dt )


   if self.state == AUDIO_STATE_PLAYING then

      self:_updatePlay(dt);

   elseif self.state == AUDIO_STATE_PAUSING then

      self:_updatePausing(dt);

   elseif self.state == AUDIO_STATE_STOPING then

      self:_updateStopping(dt);

   end
end

----------------------------------------------------------------

function AudioObject:seekMyVolume( target, duration, ease )

   local current = self.myVolume;

   if current ~= target then

      self.seekData    =
      {
         start    = current,
         target   = target,
         duration = duration,
         time     = 0,
         ease     = ease,
      };

      AudioMgr:addToUpdate( self );

   else

      self.seekData = nil;

   end
end

----------------------------------------------------------------

function AudioObject:_updateSeek(dt)

   local seek = self.seekData;

   if seek then

      seek.time  = seek.time + dt;

      if seek.time < seek.duration then

         local v = math.lerp( seek.start, seek.target, seek.time/seek.duration );

         --use internal to not stop the seeking process
         self:_setMyVolume( v );

         return false;
      else

         self.seekData = nil;
         self:_setMyVolume( seek.target );

         return true;

      end

   else

      return true;

   end
end

----------------------------------------------------------------

function AudioObject:_updatePlay(dt)

   if self:_updateSeek(dt) then

      if self:isLooping() or (self.type == AUDIO_TYPE_LAYER) then

         --We are looping... no need to monitor being done
         AudioMgr:rmvFromUpdate( self );

      else


         if self.untz and (not self.untz:isPlaying()) then
            self:_done();
         end
      end
   end
end

----------------------------------------------------------------

function AudioObject:_updatePausing(dt)

   if self:_updateSeek(dt) then

      self:_pause();

   end

end

----------------------------------------------------------------

function AudioObject:_updateStopping( dt )

   if self:_updateSeek(dt) then

      self:_stop();

   end

end

----------------------------------------------------------------
