
-- Parses CSV and returns a Lua table of rows, which are Lua tables of values.
-- This function should correctly handle embedded double quotes, commas, and new lines.
function parseCSV(stringData)

   local result = {}

   local COMMA_ID = string.byte(",")
   local DOUBLE_QUOTE_ID = string.byte("\"")
   local NEWLINE_ID = string.byte("\n")

   local valueBegin = nil
   local escaping = false
   local consecutiveQuotes = 0
   local row = {}
   local value = {}

   local function beginValue(i)

      valueBegin = i
   end

   local function finishValue(i, valueOverride)

      if not valueOverride and valueBegin then

         table.insert(value, string.sub(stringData, valueBegin, i - 1))

         -- Remove consecutive double quotes - these were escaped.
         local strValue = table.concat(value)
         strValue = string.gsub(strValue, "\"\"", "\"")

         table.insert(row, strValue)
      else
         table.insert(row, "")
      end

      value = {}
      valueBegin = nil
      escaping = false
   end

   local function finishRow()

      table.insert(result, row)

      valueBegin = nil
      escaping = false
      row = {}
   end

   -- Strip carriage returns
   stringData = string.gsub(stringData, "\r", "")

   -- Begin parsing
   for i = 1, #stringData do

      local charId = string.byte(stringData, i)
      local prevCharId = string.byte(stringData, i - 1)
      local nextCharId = string.byte(stringData, i + 1)

      if charId == DOUBLE_QUOTE_ID then

         -- Double quotes in values are escaped with double quotes.

         -- If we're just starting a value, this value is escaped with double quotes.
         if #value == 0 and not valueBegin then
            escaping = true
         end

         consecutiveQuotes = consecutiveQuotes + 1

      else

         if consecutiveQuotes % 2 == 1 then

            if valueBegin then
               finishValue(i - 1)
            else
               if consecutiveQuotes > 1 then
                  beginValue(i - consecutiveQuotes + 1)
               else
                  beginValue(i)
               end
            end

         else

            if charId == COMMA_ID then

               if prevCharId == COMMA_ID then
                  finishValue(i, "")
               elseif not escaping then
                  finishValue(i)
               end

            elseif charId == NEWLINE_ID then

               if not escaping then
                  finishValue(i)
                  finishRow()
               end

            else

               if not valueBegin then
                  beginValue(i)
               end
            end

         end

         consecutiveQuotes = 0

      end

   end

   if #row > 0 then
      finishRow()
   end

   return result

end
