require('engine/json/json')
require('app/debug')
require('engine/network/network-manager')

local SAVE_EXTENSION    = ".lua";
local EXTRACT_PATTERN   = "(.-)([^\\/]-%.?([^%.\\/]*))$";

local jsonModule        = require('json');

AppDirectory   = MOAIFileSystem.getWorkingDirectory ();

local CONTINUE_ON_LUA_SAVE_FAILURES  = true;
local CONTINUE_ON_JSON_SAVE_FAILURES = true;

local CONTINUE_ON_LUA_LOAD_FAILURES  = true;
local CONTINUE_ON_JSON_LOAD_FAILURES = true;

----------------------------------------------------------------

function splitFileName( url )

   return string.match(url, EXTRACT_PATTERN);

end

----------------------------------------------------------------

function extractPathName( url )

   local path, file, ext =  string.match(url, EXTRACT_PATTERN);
   return path;

end

----------------------------------------------------------------

function extractFileName( url )

   local path, file, ext =  string.match(url, EXTRACT_PATTERN);
   return file;

end

----------------------------------------------------------------

function extractExtensionName( url )

   local path, file, ext =  string.match( url, EXTRACT_PATTERN );
   return ext;

end

----------------------------------------------------------------

function createFullFilename( folder, filename )

   Debug.assertType(folder, "string", "folder");

   local full = "";

   if folder and #folder > 0 then

      local lastFolder = string.byte(folder, #folder);

      if       lastFolder == string.byte('/')
         or    lastFolder == string.byte('\\')   then
         full = folder..filename;
      else
         full = folder..'/'..filename;
      end
   else
      full = filename;
   end
   return full;
end

----------------------------------------------------------------

function loadJsonFileProtected( fileName, key, maxAttempt )

   if jsonModule and fileName and key then

      local retError    = nil;
      local data        = nil;

      local attempt     = 0;
      local maxAttempt2 = math.max( maxAttempt or 1, 1);

      while (data == nil) and (attempt < maxAttempt2) do

         attempt = attempt + 1;

         if MOAIFileSystem.checkFileExists ( createFullFilename(MOAIEnvironment.documentDirectory, fileName ) ) then

            local imageDataBuffer = MOAIDataBuffer.new();

            local workingDir = MOAIFileSystem.getWorkingDirectory ()

            MOAIFileSystem.setWorkingDirectory ( MOAIEnvironment.documentDirectory )

            imageDataBuffer:load(fileName)

            MOAIFileSystem.setWorkingDirectory ( workingDir );

            local dataEncrypted = imageDataBuffer:getString();

            if dataEncrypted then

               local jData, error = MOAIRoadRiot.decrypt(dataEncrypted, key);

               if jData then

                  local data = jsonModule.decode(jData);

                  return data;

               else

                  retError = error;

               end

            else

               data     = nil;
               retError = "loadJsonFileProtected -> dataEncrypted is nil";

            end

         else

            data     = nil;
            retError = "loadJsonFileProtected -> File Does not exist";

         end

      end

   else

      Debug.assert( jsonModule,  "loadLocalFileProtected->jsonModule"   );
      Debug.assert( fileName,    "loadLocalFileProtected->fileName"     );
      Debug.assert( key,         "loadLocalFileProtected->key"          );

   end

   return data, retError;

end

----------------------------------------------------------------

-- Saves the table to a file.
-- Returns encoded JSON before encryption, or nil if the save was unsuccessful.
function saveJsonFileProtected( fileName, data, key, maxAttempt )

   -- Check validity of parameters
   if not jsonModule or not fileName or not data or not key then

      Debug.assert( jsonModule,  "saveJsonFileProtected->jsonModule" );
      Debug.assert( fileName,    "saveJsonFileProtected->fileName"   );
      Debug.assert( data,        "saveJsonFileProtected->data"       );
      Debug.assert( key,         "saveJsonFileProtected->key"        );
   end

   -- Guarantee at least one attempt
   maxAttempt = math.max(1, maxAttempt or 3);

   local success = false;
   local attempt = 0;

   while not success and attempt < maxAttempt do

      attempt = attempt + 1;

      local jData = jsonModule.encode(data);

      local encryptedData = MOAIRoadRiot.encrypt(jData, key);

      local imageDataBuffer = MOAIDataBuffer.new();
      imageDataBuffer:setString(encryptedData)

      local workingDir = MOAIFileSystem.getWorkingDirectory()

      MOAIFileSystem.setWorkingDirectory(MOAIEnvironment.documentDirectory)

      success = imageDataBuffer:save(fileName);

      MOAIFileSystem.setWorkingDirectory(workingDir)

      if success then

         if attempt > 1 then

            MOAIFlurry:event( "Error-File-Save-Recovered",            "Filename", tostring(fileName), "Attempt", tostring(attempt), "MaxAttempt", tostring(maxAttempt2), "SaveFormat", "JSON" );
            MOAIFlurry:event( "Error-File-Save", "Type", "Recovered", "Filename", tostring(fileName), "Attempt", tostring(attempt), "MaxAttempt", tostring(maxAttempt2), "SaveFormat", "JSON" );
         end

         return jData;
      end

   end

   if not success then

      MOAIFlurry:event( "Error-File-Save", "Type", "Failed", "Filename", tostring(fileName), "Attempt", tostring(attempt), "MaxAttempt", tostring(maxAttempt2), "SaveFormat", "JSON" );
   end

end

----------------------------------------------------------------

function loadLocalFile( fileName, maxAttempt )

   Debug.assertType(fileName, "string", "fileName");

   local data = nil;
   local attempt     = 0;
   local maxAttempt2 = math.max( maxAttempt or 1, 1);
   local errorMsg = nil;

   while (data == nil) and (attempt < maxAttempt2) do

      attempt = attempt + 1;

      if MOAIFileSystem.checkFileExists ( createFullFilename(MOAIEnvironment.documentDirectory, fileName ) ) then

         local workingDir = MOAIFileSystem.getWorkingDirectory ()

         MOAIFileSystem.setWorkingDirectory ( MOAIEnvironment.documentDirectory )

         data, errorMsg = dofile_protected ( fileName );

         MOAIFileSystem.setWorkingDirectory ( workingDir );

      else

         data = nil;

      end

   end

   if data and attempt > 1 then
      MOAIFlurry:event( "Error-File-Load-Recovered", "Filename", fileName, "attempt", ""..attempt, "maxAttempt", ""..maxAttempt2 );
   end

   return data, errorMsg;

end

----------------------------------------------------------------

function saveLocalFile( fileName, data, maxAttempt )

   local success     = false;
   local attempt     = 0;
   local maxAttempt2 = math.max( maxAttempt or 3, 1);

   while (not success) and ( attempt < maxAttempt2) do

      attempt = attempt + 1;

      local serializer  = MOAISerializer.new ();
      serializer:serialize ( data );

      local gamestateStr = serializer:exportToString ();

      if gamestateStr then

         local workingDir = MOAIFileSystem.getWorkingDirectory ()
         MOAIFileSystem.setWorkingDirectory ( MOAIEnvironment.documentDirectory )

         success = save_lua_string_to_file_encrypted( fileName, gamestateStr )

         MOAIFileSystem.setWorkingDirectory ( workingDir )

         if not success then
            MOAIFlurry:event( "Error-File-Save", "Type", "Failed", "Filename", tostring(fileName), "Attempt", tostring(attempt), "MaxAttempt", tostring(maxAttempt2), "SaveFormat", "Lua" );
         else
            if attempt > 1 then
               MOAIFlurry:event( "Error-File-Save-Recovered",            "Filename", tostring(fileName), "Attempt", tostring(attempt), "MaxAttempt", tostring(maxAttempt2), "SaveFormat", "Lua" );
               MOAIFlurry:event( "Error-File-Save", "Type", "Recovered", "Filename", tostring(fileName), "Attempt", tostring(attempt), "MaxAttempt", tostring(maxAttempt2), "SaveFormat", "Lua" );
            end
         end

      else

         MOAIFlurry:event( "Error-File-Save", "Type", "Nil gamestateStr", "Filename", tostring(fileName), "Attempt", tostring(attempt), "MaxAttempt", tostring(maxAttempt2), "SaveFormat", "Lua" );

      end

   end

   -- MOAIFlurry:event( "Error-File-Save", "Type", "Failed", "Filename", tostring(fileName), "Attempt", tostring(attempt), "MaxAttempt", tostring(maxAttempt2), "SaveFormat", "JSON" );

   return success;

end

----------------------------------------------------------------

function saveFileMultiFormat( luaFileName, jsonFileName, data, key, maxAttempt )

   Debug.assertType(luaFileName, "string", "luaFileName");
   Debug.assertType(jsonFileName, "string", "jsonFileName");

   key = key or App:calcEncryptionKey();

   -- maxAttempt is optional; the save functions handle it

   local luaSuccess = saveLocalFile( luaFileName, data, maxAttempt );

   if not luaSuccess then

      -- saveLocalFile handles its own error reporting

      if not CONTINUE_ON_LUA_SAVE_FAILURES then
         return false;
      end
   end

   local jsonSuccess = saveJsonFileProtected( jsonFileName, data, key );

   if not jsonSuccess then

      -- saveJsonFileProtected handles its own error reporting

      if not CONTINUE_ON_JSON_SAVE_FAILURES then
         return false;
      end
   end

   print("Multiformat save successful.");

   return true;

end

----------------------------------------------------------------

function loadFileMultiFormat( luaFileName, jsonFileName, key, maxAttempt )

   Debug.assertType(luaFileName, "string", "luaFileName");
   Debug.assertType(jsonFileName, "string", "jsonFileName");

   key = key or App:calcEncryptionKey();

   local luaData, luaError = loadLocalFile( luaFileName, maxAttempt);

   if not luaData then

      if luaError then
         MOAIFlurry:event( "Data-Load-Failure", "Filename", tostring(luaFileName), "FileType", "Lua", "ErrorMsg", tostring(luaError) );
      end

      if not CONTINUE_ON_LUA_LOAD_FAILURES then
         return false;
      end
   end

   local jsonData, jsonError = loadJsonFileProtected( jsonFileName, key, maxAttempt );

   if not jsonData then

      if jsonError then
         MOAIFlurry:event( "Data-Load-Failure", "Filename", tostring(luaFileName), "FileType", "JSON", "ErrorMsg", tostring(jsonError) );
      end

      if not CONTINUE_ON_JSON_LOAD_FAILURES then
         return false;
      end
   end

   -- Check difference between the files
   if luaData and jsonData then

      local totalDiffs = countTableDifferences(luaData, jsonData);

      if totalDiffs > 0 then

         MOAIFlurry:event( "File-Save-Difference",
                           "differences",   tostring(totalDiffs),
                           "lua filename",  tostring(luaFileName),
                           "json filename", tostring(jsonFileName) );

         -- Only causes an abort in debug mode; release is unaffected
         Debug.error("There were differences between Lua & JSON files. Lua Filename: " .. tostring(luaFileName) .. " Diff count: " .. tostring(totalDiffs));

      end

   end

   -- Return one of the data tables - this can be tweaked later

   return luaData or jsonData;

end

----------------------------------------------------------------

function dofileToFile ( url, destination, callObj, callSuccess, callFailure )

   print("dofileToFile", url, destination, callObj, callSuccess, callFailure );

   local success = false;
   local tbl     = dofile(url);
   if tbl then
      success = saveLocalFile( destination,  tbl );
   end

   if success then
      if  callSuccess then
         callSuccess(callObj);
      end
   else
      if  callFailure then
         callFailure(callObj);
      end
   end
end

----------------------------------------------------------------

function downloadToFile ( url, destination, callHandler, downloadSuccessFunc, downloadFailureFunc, priority )

   print("downloadToFile - url:", url);

   local startStamp = os.time();

   EventMgr:postEvent( {state="started", url=url, startTime=startStamp} , "file-download");

   -- Inject callbacks from parameters into another callback that retrieves the data and then saves it to a file.
   -- This means that the success and failure callbacks that are passed to downloadToFile are fired after the file is saved,
   -- not when the file is retrieved.

   local wrapperCallObj = {}
   function wrapperCallObj:callSuccess(task)

      local imageDataBuffer = MOAIDataBuffer.new()
      imageDataBuffer:setString(task:getString())

      local workingDir = MOAIFileSystem.getWorkingDirectory ()

      MOAIFileSystem.setWorkingDirectory ( MOAIEnvironment.documentDirectory )

      local success = imageDataBuffer:save(destination);

      MOAIFileSystem.setWorkingDirectory ( workingDir )

      EventMgr:postEvent( {state="finished", url=url, startTime=startStamp, saved=success} , "file-download");

      if success then
         if downloadSuccessFunc then
            downloadSuccessFunc(callHandler);
         end
      else
         MOAIFlurry:event( "Error-File-Save",
                           "Type",              "Failed-From-Download",
                           "Filename",          destination ,
                           "time",              tostring(os.time()-startStamp)        );

         if downloadFailureFunc then
            downloadFailureFunc(callHandler);
         end
      end

   end

   function wrapperCallObj:callFailure(task)

      downloadFailureFunc(callHandler);

      EventMgr:postEvent( {state="failed", url=url, startTime=startStamp} , "file-download");

   end

   local requestOptions = {
      priority = priority or Request.PRIORITY_NORMAL,
      callHandler = wrapperCallObj,
      successFunc = wrapperCallObj.callSuccess,
      failureFunc = wrapperCallObj.callFailure,

      url = url,
      retryCount = 3,
   }

   local request = WebRequest.newFromBase(requestOptions);

   NetworkMgr:sendWebRequest(request);

end

----------------------------------------------------------------

-- Execute a file in protected mode
function dofile_protected(fileName)

   Debug.assertType(fileName, "string", "dofile_protected -> fileName");

   -- local loadedChunk, errorMsg = loadfile(fileName);
   local loadedChunk, errorMsg = loadfile_encrypted(fileName);

   if loadedChunk then
      -- print("Returning pcall of loaded chunk.");

      local results = { pcall(loadedChunk) };

      if results[1] then
         return unpack(results, 2);
      else
         print("Failed to run loaded chunk.");
         return nil, errorMsg;
      end
   else
      -- Debug.error("Error in dofile_protected: " .. tostring(errorMsg));

      print("Error in dofile_protected: " .. tostring(errorMsg));

      return nil, errorMsg;
   end

end
