require('engine/object/object')

-- TODO Add system to requeue connections until the maximum attempt count has been reached
-- TODO Handle redirects
-- TODO Add an optional delay time between connection attempts?

local DEFAULT_TIMEOUT_TIME = 15;
local DEFAULT_REDIRECT_LIMIT = 5;
local DEFAULT_HIGH_PRIORITY_SLOT_COUNT = 2;
local DEFAULT_SLOT_COUNT = 3;

local Base = Object;

ConnectionQueue = {}


function ConnectionQueue.new()

   local obj = Base.new(ConnectionQueue);

   obj._timeoutTime = DEFAULT_TIMEOUT_TIME;
   obj._redirectLimit = DEFAULT_REDIRECT_LIMIT;
   obj._highPriorityActiveRequestLimit = DEFAULT_HIGH_PRIORITY_SLOT_COUNT;
   obj._activeRequestLimit = DEFAULT_SLOT_COUNT;

   obj._activeRequests = {};
   obj._queuedRequests = {};

   return obj;

end


function ConnectionQueue:update(dt)

   self:_startQueuedRequests();

   self:_processFinishedRequests();

end


function ConnectionQueue:queueRequest(request)

   Debug.assertType(request, {"table", "userdata"}, "ConnectionQueue:queueRequest -> request");

   Debug.assertType(request._priority, "number", "ConnectionQueue:queueRequest -> request._priority");
   Debug.assertType(request._callHandler, {"nil", "table", "userdata"}, "ConnectionQueue:queueRequest -> request._callHandler");
   Debug.assertType(request._successFunc, {"nil", "function"},          "ConnectionQueue:queueRequest -> request._successFunc");
   Debug.assertType(request._failureFunc, {"nil", "function"},          "ConnectionQueue:queueRequest -> request._failureFunc");

   table.insert(self._queuedRequests, request);

   -- print("ConnectionQueue - queuing request - Active count:", self:_getNumSlotsInUse(), "Queued count:", #self._queuedRequests);

end


-- Start requests if we have queued requests and slots available.
function ConnectionQueue:_startQueuedRequests()

   if #self._queuedRequests > 0 then

      local i = 1;

      while i <= #self._queuedRequests do

         local request = self._queuedRequests[i];

         if self:_getNumSlotsAvailable() == 0 and self:_getNumHighPrioritySlotsAvailable() == 0 then
            break;
         end

         if self:_shouldStartRequest(request) then

            table.insert(self._activeRequests, request);
            table.remove(self._queuedRequests, i);

            request.task = self:_startRequest(request);

            -- print("ConnectionQueue - starting request - Active count:", self:_getNumSlotsInUse(), "Queued count:", #self._queuedRequests);

         else

            i = i + 1;

         end

      end

   end

end


-- Remove completed requests from the active request table if no more attempts can be made.
-- Otherwise, put the requests back in the queue.
function ConnectionQueue:_processFinishedRequests()

   for i = #self._activeRequests, 1, -1 do

      local request = self._activeRequests[i];

      if request._finished then

         if request._retry then

            -- print("ConnectionQueue - Adding an attempted request back into the front of the queue.");

            table.insert(self._queuedRequests, 1, request);

         end

         -- print("ConnectionQueue - Deactivating finished request.");
         table.remove(self._activeRequests, i);

      end

   end

end


function ConnectionQueue:_hasSlotAvailableForPriority(priority)

   if priority == Request.PRIORITY_HIGH then

      return self:_getNumSlotsAvailable() > 0 or self:_getNumHighPrioritySlotsAvailable() > 0;

   elseif priority == Request.PRIORITY_LOW
   or priority == Request.PRIORITY_NORMAL then

      return self:_getNumSlotsAvailable() > 0;

   end

   Debug.error("Unknown priority.");

   return false;

end


function ConnectionQueue:_isAnyTangoRequestActive()

   for _, request in ipairs(self._activeRequests) do

      if request.typeName == TangoRequest.typeName then

         return true;

      end

   end

   return false;

end


function ConnectionQueue:_shouldStartRequest(request)

   return self:_hasSlotAvailableForPriority(request._priority);

end


function ConnectionQueue:_startRequest(request)

   request._finished = false;
   request._retry = nil;
   request._success = nil;
   request._timeFinished = nil;
   request._timeStarted = os.time();
   request._attempts = (request._attempts or 0) + 1;

   if request.typeName == WebRequest.typeName then

      return self:_startWebRequest(request);

   elseif request.typeName == PushWebRequest.typeName then
       print("\nConnection Q -> Starting Pushwebrequest.\n");
      return self:_startPushWebRequest(request);

   elseif request.typeName == TangoRequest.typeName then

      return self:_startTangoRequest(request);

   end

end


function ConnectionQueue:_startWebRequest(request)

   request._redirectHits = request._redirectHits or 0;

   -- TODO If we switch to platform-specific HttpTask classes, make those changes here.

   print("ConnectionQueue: Starting cURL request.");

   local task = MOAIHttpTaskCurl.new();

   task:setVerb(MOAIHttpTask.HTTP_GET);
   task:setUrl(request._redirectURL or request.url);
   task:setUserAgent(MOAIEnvironment.userAgent or "");
   task:setTimeout(request.timeoutTime or self._timeoutTime);
   task:setCallback(function(task, code) self:_onRequestFinished(task, code, request, request._callHandler, request._successFunc, request._failureFunc) end);

   task:performAsync();

   return task;

end


function ConnectionQueue:_startPushWebRequest(request)

   request._redirectHits = request._redirectHits or 0;

       print("\nConnectionQueue:_startPushWebRequest -> creating curl task.\n");


   local task = MOAIHttpTaskCurl.new();

   task:setVerb(MOAIHttpTask.HTTP_POST);
   task:setUrl(request._redirectURL or request.url);
   task:setUserAgent(MOAIEnvironment.userAgent or "");
   task:setTimeout(request.timeoutTime or self._timeoutTime);
   task:setCallback(function(task, code) self:_onRequestFinished(task, code, request, request._callHandler, request._successFunc, request._failureFunc) end);

   print("\nConnectionQueue:_startPushWebRequest -> about to set curl task header.\n");


   if request.data and request.data.key then
    task:setHeader("X-Hash", request.data.key);
   end
   task:setVerbose(false);

    if request.data then
        local encodedJSON = request.data.payload;
        print("ConnectionQueue: Starting cURL request for a push web request for url: " .. request.url .. " with data: " .. tostring(encodedJSON));
        task:setBody(tostring(encodedJSON));
    else
        task:setBody("");
        print("ConnectionQueue: Starting cURL request for a push web request for url: " .. request.url .. " with data: EMPTY" );
    end



   task:performAsync();

   return task;

end



function ConnectionQueue:_startTangoRequest(request)

   local tangoRequest = request.requestClass:new();

   tangoRequest:setListener(MOAITangoRequest.EVENT_SUCCESS, function(...) self:_onRequestFinished(tangoRequest, 1, request, request._callHandler, request._successFunc, request._failureFunc, {...}) end);
   tangoRequest:setListener(MOAITangoRequest.EVENT_FAILURE, function(...) self:_onRequestFinished(tangoRequest, 0, request, request._callHandler, request._successFunc, request._failureFunc, {...}) end);

   tangoRequest:send(unpack(request.requestData));

   return tangoRequest;

end


-- Optionally pass in a priority to get a count for requests with matching priority.
-- Pass nil to get a count of all requests.
function ConnectionQueue:_getNumSlotsInUse(priority)

   local total = 0;

   for _, request in ipairs(self._activeRequests) do

      if priority == nil or request._priority == priority then

         total = total + 1;

      end

   end

   return total;

end


-- Returns the count of normal- and low-priority slots available.
function ConnectionQueue:_getNumSlotsAvailable()

   return self._activeRequestLimit - self:_getNumSlotsInUse(Request.PRIORITY_LOW) - self:_getNumSlotsInUse(Request.PRIORITY_NORMAL);

end


-- Returns the count of high-priority slots available.
function ConnectionQueue:_getNumHighPrioritySlotsAvailable()

   return self._highPriorityActiveRequestLimit - self:_getNumSlotsInUse(Request.PRIORITY_HIGH);

end


function ConnectionQueue:_onRequestFinished(task, code, request, callHandler, successFunc, failureFunc, callbackReturnValues)

   print("\nConnectionQueue:_onRequestFinished -> about to set curl task header.\n");

   Debug.assertType(task, "userdata", "ConnectionQueue:_onRequestFinished -> task");
   Debug.assertType(code, "number", "ConnectionQueue:_onRequestFinished -> code");
   Debug.assertType(request, "table", "ConnectionQueue:_onRequestFinished -> request");
   Debug.assertType(callHandler, {"nil", "userdata", "table"}, "ConnectionQueue:_onRequestFinished -> callHandler");
   Debug.assertType(successFunc, {"nil", "function"}, "ConnectionQueue:_onRequestFinished -> successFunc");
   Debug.assertType(failureFunc, {"nil", "function"}, "ConnectionQueue:_onRequestFinished -> failureFunc");

   request._finished = true;
   request._timeFinished = os.time();

   if request.typeName == WebRequest.typeName then

      self:_onWebRequestFinished(task, code, request, callHandler, successFunc, failureFunc);

   elseif request.typeName == PushWebRequest.typeName then

      self:_onPushWebRequestFinished(task, code, request, callHandler, successFunc, failureFunc);

   elseif request.typeName == TangoRequest.typeName then

      self:_onTangoRequestFinished(task, code, request, callHandler, successFunc, failureFunc, callbackReturnValues);

   else

      Debug.error("ConnectionQueue:_onRequestFinished with unknown request.typeName:", request.typeName);

   end

end


function ConnectionQueue:_onWebRequestFinished(task, code, request, callHandler, successFunc, failureFunc)

   local redirectURL = task:getResponseHeader("Refresh")
                    or task:getResponseHeader("refresh")
                    or task:getResponseHeader("Location")
                    or task:getResponseHeader("location");

   -- print("ConnectionQueue - code:", code, "- redirectURL:", redirectURL);

   if request.ignoreResponse or (task:getProgress() == 1.0 and task:getSize() > 0 and task:getString()) then

      --print("ConnectionQueue - request success");
      print("ConnectionQueue - success stats - Progress:", task:getProgress(), "Redirects:", request._redirectHits, "Attempts:", request._attempts);

      request._success = true;
      request._retry = false;

      if successFunc then
         if callHandler then
            successFunc(callHandler, task);
         else
            successFunc(task);
         end
      end

   elseif redirectURL and request._redirectHits <= self._redirectLimit then

      print("ConnectionQueue - request hit redirect with URL:", redirectURL);

      request._success = false;
      request._retry = true;
      request._redirectURL = redirectURL;

      -- Redirects shouldn't count toward the attempt count
      request._attempts = request._attempts - 1;
      request._redirectHits = (request._redirectHits or 0) + 1;

      Debug.assert(request._attempts >= 0);

   elseif request._attempts >= request.retryCount or request._redirectHits > self._redirectLimit then

      -- print("ConnectionQueue - request failure");
      print("ConnectionQueue - failure stats - Progress:", task:getProgress(), "Redirects:", request._redirectHits, "Attempts:", request._attempts);
      -- print("Limits... retryCount:", request.retryCount, "redirectLimit:", self._redirectLimit);

      request._success = false;
      request._retry = false;

      if failureFunc then
         if callHandler then
            failureFunc(callHandler, task);
         else
            failureFunc(task);
         end
      end

   else

      request._retry = true;

   end

end

function ConnectionQueue:_onPushWebRequestFinished(task, code, request, callHandler, successFunc, failureFunc)

   print("\nConnectionQueue:_onRequestFinished -> about to set curl task header.\n");


   local redirectURL = task:getResponseHeader("Refresh")
                    or task:getResponseHeader("refresh")
                    or task:getResponseHeader("Location")
                    or task:getResponseHeader("location");

   print("ConnectionQueue_onPushWebRequestFinished - code:", code, "- redirectURL:", redirectURL,"- task:", task:getProgress() or "nil", " -task.size: ", task:getSize() or "nil");

   if request.ignoreResponse or (task:getProgress() == 1.0 and task:getSize() > 0 and task:getString()) then

      --print("ConnectionQueue - request success");
      print("ConnectionQueue - success stats - Progress:", task:getProgress(), "Redirects:", request._redirectHits, "Attempts:", request._attempts);

      request._success = true;
      request._retry = false;

      if successFunc then
         if callHandler then
            successFunc(callHandler, task);
         else
            successFunc(task);
         end
      end

   elseif redirectURL and request._redirectHits <= self._redirectLimit then

      print("ConnectionQueue - request hit redirect with URL:", redirectURL);

      request._success = false;
      request._retry = true;
      request._redirectURL = redirectURL;

      -- Redirects shouldn't count toward the attempt count
      request._attempts = request._attempts - 1;
      request._redirectHits = (request._redirectHits or 0) + 1;

      Debug.assert(request._attempts >= 0);

   elseif request._attempts >= request.retryCount or request._redirectHits > self._redirectLimit then

      -- print("ConnectionQueue - request failure");
      print("ConnectionQueue - failure stats - Progress:", task:getProgress(), "Redirects:", request._redirectHits, "Attempts:", request._attempts);
      -- print("Limits... retryCount:", request.retryCount, "redirectLimit:", self._redirectLimit);

      request._success = false;
      request._retry = false;

      if failureFunc then
         if callHandler then
            failureFunc(callHandler, task);
         else
            failureFunc(task);
         end
      end

   else

      request._retry = true;

   end

end



function ConnectionQueue:_onTangoRequestFinished(task, code, request, callHandler, successFunc, failureFunc, callbackReturnValues)

   request._retry = false;

   if code == 1 then

      request._success = true;

      Debug.assert(request._successFunc == successFunc);

      successFunc(unpack(callbackReturnValues));

   elseif code == 0 then

      request._success = false;

      Debug.assert(request._failureFunc == failureFunc);

      failureFunc(unpack(callbackReturnValues));

   else

      Debug.error("Tango request finished with unknown code:" .. tostring(code));

   end

end
