require('engine/texture/texture-manager')

local AnimPlayer = require "flash/anim-player"

local ResourceBundle = {}

local function newAnimPlayer ( self, symName, parentTransform )

  if parentTransform == nil then
    parentTransform = MOAITransform2D.new ()
  end

  local sid, sym

  if type(symName)=="number" then
    sid = symName
  else
    sid = self.symName2IdMap [ symName ];
  end

  return AnimPlayer.new ( self, sid, parentTransform )

end

local function roundTime (t)
  return math.floor ( t * 1000 + 0.5 ) / 1000
end

local function mergeTables ( target, source)

  for k, v in pairs ( source ) do
    if type(v) == "table" then
      target [ k ] = mergeTables ( target [ k ], v )
    else
      target [ k ] = v
    end
  end

  return target
end

local function loadBrushesQuadDeck ( brushData, texture )
   local tw
  local th

  local textureObj = TextureMgr:getTexture(texture, true);

  tw, th = textureObj:getSize ()

  local quadDeck = MOAIGfxQuadDeck2D.new ()
  quadDeck:setTexture ( textureObj )
  quadDeck:reserve ( #brushData )
  local pivots = {}
  local brushName2IdMap = {}
  for i, brush in pairs ( brushData ) do

    quadDeck:setUVRect ( i, brush.u0, brush.v0, brush.u1, brush.v1 )

    local w = brush.w * 0.5;
    local h = brush.h * 0.5;

    if ( brush.r ) then
      quadDeck:setQuad ( i, h, w, h, -w, -h, -w, -h, w )
    else
      quadDeck:setRect ( i, -w, -h, w, h )
    end

    local px, py

    if brush.r then
      px = brush.px - brush.h * 0.5
      py = - (brush.py - brush.w * 0.5)
    else
      px = brush.px - brush.w * 0.5
      py = - (brush.py - brush.h * 0.5)
    end

    pivots [ i ] = { px=px, py=py }
    brushName2IdMap [ brush.name ] = i
  end

  local flipTransform = MOAITransform2D.new()
  flipTransform:setScl ( 1, -1 )
  quadDeck:transform ( flipTransform )

  local brushDeck = { pivots = pivots, quadDeck = quadDeck, brushName2IdMap = brushName2IdMap }

  return brushDeck
end

local function dumpToTraceFile( resBundle, filename, t0, t1, step )
  local f = io.open( filename, "w")

  for sid, sym in pairs ( resBundle.syms ) do
    for lid, layer in pairs ( sym.layers ) do
      for t = t0, t1, step do
        f:write ( string.format ( '%d \t%s \t%d\t%f\t', sid, sym.name, lid, t) )

        if layer.tid then
          f:write ( 'tid=' .. layer.tid .. '\t' )
        end

        for curname, cur in pairs ( layer.curves ) do
          f:write( string.format ( '%s=%f\t', curname, cur:getValueAtTime( t ) ) )
        end
        f:write ( '\n' )
      end
    end
  end

  f:close()

end

local function createCurves ( nKeys, curveNames )
  local curves = {}
  for _, name in pairs ( curveNames ) do
    local curve = MOAIAnimCurve.new ()
    curve:reserveKeys ( nKeys )
    curves [ name ] = curve
  end

  return curves
end

local function loadSymbol ( symData, brushData, fps )
    local step = 1 / fps

    local embSymOp = nil
    local span = 0
    local layers = {}

    for lid, layerData in ipairs ( symData.layers ) do

      local nKeys = #layerData.frames

      local mergedFrame = {
        ease=0, aid=nil, start=0, length=0,
        transform={x=0, y=0, sx=1, sy=0, a=100, r=0, px=0, py=0}
      }

      local keys = { "id", "vis", "px", "py" }

      -- first pass to determine what curves to make
      propertyIsDefault = { a=true, sx=true, sy=true, px=true, py=true, x=true, y=true, vis=true, r=true }

      for j, frameData in ipairs ( layerData.frames ) do

        mergedFrame = mergeTables( mergedFrame, frameData )

        if mergedFrame.transform.a ~= 100 then
          propertyIsDefault["a"] = false
        end

        if mergedFrame.transform.sx ~= 1 then
          propertyIsDefault["sx"] = false
        end

        if mergedFrame.transform.sy ~= 1 then
          propertyIsDefault["sy"] = false
        end

        if mergedFrame.transform.x ~= 0 then
          propertyIsDefault["x"] = false
        end

        if mergedFrame.transform.y ~= 0 then
          propertyIsDefault["y"] = false
        end

        if mergedFrame.transform.r ~= 0 then
          propertyIsDefault["r"] = false
        end

        if mergedFrame.px ~= 0 then
          propertyIsDefault["px"] = false
        end

        if mergedFrame.py ~= 0 then
          propertyIsDefault["py"] = false
        end

      end

      -- Create curves
      for k,v in pairs(propertyIsDefault) do
        if not v then
          table.insert(keys, k)
        end
      end

      local cs = createCurves ( nKeys, keys )
      cs.vis:reserveKeys ( nKeys + 1 )

      local isEmbeddedSymbol
      local layer = {}
      local unwrappedR = nil

      -- Second pass to store data in curves
      for j, frameData in ipairs ( layerData.frames ) do

        local t = frameData.start * step
        local px = nil
        local py = nil

        span = math.max( span, (frameData.start + frameData.length) * step )
        if frameData.aid ~= nil or frameData.tid ~= nil then
          mergedFrame.aid = nil
          mergedFrame.tid = nil
        end
        mergedFrame.loop = 0

        mergedFrame = mergeTables( mergedFrame, frameData )

        cs.vis:setKey ( j, t, mergedFrame.vis, MOAIEaseType.FLAT )

        if ( mergedFrame.vis == 1 ) then
          if ( mergedFrame.aid ~= nil ) then

            cs.id:setKey ( j, t, mergedFrame.aid, MOAIEaseType.FLAT )
            local brush = brushData [ mergedFrame.aid ]
            if brush.r then
              px = brush.px - brush.h * 0.5
              py = - (brush.py - brush.w * 0.5)
            else
              px = brush.px - brush.w * 0.5
              py = - (brush.py - brush.h * 0.5)
            end
            mergedFrame.px = px
            mergedFrame.py = py
            isEmbeddedSymbol = false

          elseif ( mergedFrame.tid ~= nil ) then

            if embSymOp == nil then
              embSymOp = {}
            end
            if embSymOp [ roundTime(t) ] == nil then
              embSymOp [ roundTime(t) ] = {}
            end

            table.insert(embSymOp [ roundTime(t) ],
              { layerId = lid,
                vis     = mergedFrame.vis,
                symId   = mergedFrame.tid,
                loop    = mergedFrame.loop })

            isEmbeddedSymbol = true

          else
            error ( 'Either aid or tid field is expected per frame if vis=1' )
          end
        else
          cs.id:setKey ( j, t, mergedFrame.aid, MOAIEaseType.FLAT )
        end

        local transform = mergedFrame.transform

        local mode = MOAIEaseType.LINEAR
        local weight = mergedFrame.ease / 100

        if weight > 0 then
          mode = MOAIEaseType.EASE_IN
        elseif weight < 0 then
          mode = MOAIEaseType.EASE_OUT
          weight = -weight
        end

        -- Classic rotation in Auto mode (default for Flash Pro)
        -- the rotation is done to have less motion. Unwrapping
        -- angle acheives this
        if unwrappedR == nil then
          unwrappedR = transform.r
        else
          -- May be d1 and d2 is enough?
          local d1 = unwrappedR % 360 - transform.r
          local d2 = unwrappedR % 360 + 360 - transform.r
          local d3 = unwrappedR % 360 - 360 - transform.r
          local ad1 = math.abs ( d1 )
          local ad2 = math.abs ( d2 )
          local ad3 = math.abs ( d3 )

		  local delta = 0;

          if ( ad1 < ad2 and ad1 < ad3 ) then
            delta = d1
          elseif ( ad2 < ad1 and ad2 < ad3 ) then
            delta = d2
          else
            delta = d3
          end

          unwrappedR = unwrappedR - delta
        end

        if cs.x then
          cs.x:setKey ( j, t, transform.x, mode, weight )
        end

        if cs.y then
          cs.y:setKey ( j, t, -transform.y, mode, weight )
        end

        if cs.r then
          cs.r:setKey ( j, t, -unwrappedR, mode, weight )
        end

        if cs.sx then
          cs.sx:setKey( j, t, transform.sx, mode, weight )
        end

        if cs.sy then
          cs.sy:setKey( j, t, transform.sy, mode, weight )
        end

        if cs.a then
          cs.a:setKey ( j, t, transform.a/100, mode, weight )
        end

        cs.px:setKey ( j, t, mergedFrame.px, MOAIEaseType.FLAT )
        cs.py:setKey ( j, t, mergedFrame.py, MOAIEaseType.FLAT )

        if isEmbeddedSymbol then
          cs.id = nil
        end

      end

      local lastFrame = layerData.frames [ #layerData.frames ]
      local tAfterLastFrame = ( lastFrame.start + lastFrame.length ) * step
      cs.vis:setKey ( #layerData.frames + 1, tAfterLastFrame, 0, MOAIEaseType.FLAT )

      layer.curves = cs
      layer.name = layerData.name
      layer.isEmbeddedSymbol = isEmbeddedSymbol

      layers [ lid ] = layer
    end

    local embSymsCurve = nil
    if embSymOp then
      local nEmbSymKeys = 0
      for _, op in pairs ( embSymOp ) do
        nEmbSymKeys = nEmbSymKeys + #op
      end
      embSymsCurve = MOAIAnimCurve.new ()

      embSymsCurve:reserveKeys ( nEmbSymKeys )

      local i = 1
      for t, _ in pairs ( embSymOp ) do
        embSymsCurve:setKey (i, t, 1, MOAIEaseType.FLAT)
        i = i + 1
      end

    end

    local sym = {}
    sym.layers = layers
    sym.embSymsCurve = embSymsCurve
    sym.embSymOp = embSymOp
    sym.span = span
    sym.frame = step
    sym.name = symData.name

    return sym
end

local function newProp ( self, graphicsName )
  local prop = MOAIProp2D.new()

  prop:setDeck ( self.brushDeck.quadDeck )
  local idx = self.brushDeck.brushName2IdMap [ graphicsName ]
  prop:setIndex ( idx )

  local pivot = self.brushDeck.pivots [ idx ]
  prop:setPiv ( pivot.px, pivot.py )

  return prop
end

ResourceBundle.new = function ( flipbookFile, spritesheetFile, fps )

  local resBundle = {}

  local flashData = dofile ( flipbookFile )

  if not flashData then
    Debug.error("Failed to load flash data from file: " .. tostring(flipbookFile));
  end

  resBundle.brushDeck = loadBrushesQuadDeck ( flashData.brushDeck, spritesheetFile )

  resBundle.syms = {}
  resBundle.symName2IdMap = {}

  for sid, symData in ipairs ( flashData.syms ) do

    local sym = loadSymbol ( symData, flashData.brushDeck, fps )

     resBundle.symName2IdMap [ sym.name ] = sid
    resBundle.syms [ sid ] = sym

  end

  resBundle.newProp = newProp
  resBundle.newAnimPlayer = newAnimPlayer
  resBundle.dumpToTraceFile = dumpToTraceFile
  resBundle.roundTime = roundTime

  return resBundle
end

return ResourceBundle
