require('app/template/template-manager');
require('game/entity/avatar/avatar');
require('game/entity/barricade/barricade');
require('game/entity/flyout/flyout');
require('game/entity/flyout/flyoutText');
require('game/entity/powerup/powerup');
require('game/entity/pickup/pickup');
require('game/entity/projectile/projectile');

----------------------------------------------------------------------

EntityManager = {};

----------------------------------------------------------------------

function EntityManager:_new( c )

   c            =  c or  {};
   setmetatable(c, self);
   self.__index = self;
   
   return c;

end

----------------------------------------------------------------------

function EntityManager:new( game, layers )

   local em          = self:_new();

   em.__marker       =  "entity-manager";

   em.game           = game;
   em.layers         = layers;
   em.entityPool     = {};
   em.avatar         = nil;
   em.projectiles    = {};
   em.vehicles       = {};
   em.explosions     = {};
   em.pickups        = {};
   em.powerups       = {};
   em.hazards        = {};
   em.barricades     = {};
   em.flyouts        = {};

   return em;

end

----------------------------------------------------------------------

function EntityManager:onActivate()

end

----------------------------------------------------------------------

function EntityManager:onDeactivate()

   self.game   = nil;
   self.layers = nil

   EventMgr:rmvListener( self );

   for k,v in pairs(self.entityPool) do
      self:_deleteListEntities(v);
   end

   self:_deleteListEntities(self.projectiles);
   self:_deleteListEntities(self.vehicles);
   self:_deleteListEntities(self.explosions);
   self:_deleteListEntities(self.pickups);
   self:_deleteListEntities(self.powerups);
   self:_deleteListEntities(self.hazards);
   self:_deleteListEntities(self.barricades);
   self:_deleteListEntities(self.flyouts);

end

----------------------------------------------------------------------

function EntityManager:cleanup( list, callback )

   for i=#list,1,-1 do

      local e = list[i];
      
      if e.deleteMe then
      
         if callback then 
         
            callback(self, e);

         end
         
         self:deleteEntity( e );
         table.remove( list, i );

      end
   end
end

----------------------------------------------------------------------

function EntityManager:_deleteListEntities(list)

   for i, v in ipairs(list) do
      if v then
         v:delete();
      end
   end

   list = {};
end

----------------------------------------------------------------------

function EntityManager:update( dt )

   self:cleanup( self.projectiles );
   self:cleanup( self.vehicles,   self.killVehicle );
   self:cleanup( self.explosions  );
   self:cleanup( self.pickups     );
   self:cleanup( self.powerups    );
   self:cleanup( self.hazards     );
   self:cleanup( self.barricades  );
   self:cleanup( self.flyouts     );

end

----------------------------------------------------------------------

function EntityManager:newEntity( factory, layer, template, x, y, a, ... )

   local e = nil;

   self.entityPool[template] = self.entityPool[template] or {};

   if #self.entityPool[template] > 0 then

      e = self.entityPool[template][1];

      e:reset(x, y, a, ...);

      table.remove( self.entityPool[template], 1 );

   else

      e = factory:new( self.game, layer, template, x, y, a, ... );

   end

   return e;

end

----------------------------------------------------------------

function EntityManager:deleteEntity( e  )

   self.entityPool[e.template] = self.entityPool[e.template] or {};

   e:deactivate();
   e:hide();

   table.insert( self.entityPool[e.template], e );

end

----------------------------------------------------------------------

function EntityManager:spawnAvatar(  x, y )

   local currentVehicle = App.data:getCurrentPlayerVehicle();

   if not App.data:isPlayerVehicleUnlocked(currentVehicle) then
      App.data:setCurrentPlayerVehicle(1);

      currentVehicle = 1;
   end

   if not self.avatar then

      self.avatar = self:newEntity( Avatar
                                  , self.layers
                                  , App.data:getCurrentPlayerVehicleTemplate()
                                  , x, y, 0
                                  , self.game.boundary
                                  , App.data:getCurrentPlayerVehiclePaintVariantTemplate()
                                  , App.data:getCurrentPlayerMainWeaponTemplate()
                                  , App.data:getCurrentArmorTemplate()
                                  , App.data:getCurrentPlayerSideWeaponTemplate()
                                  , App.data:getCurrentPlayerTurretWeaponTemplate()               );

   else

      print("ERROR: Multiple Avatar");

   end

   return self.avatar;

end

----------------------------------------------------------------------

function EntityManager:spawnHazard( tpl, x, y, a, target  )

   local e = self:newEntity( tpl.factory, self.layers, tpl,  x, y, a, target );

   table.insert(self.hazards, e);

   return a;

end


----------------------------------------------------------------------

function EntityManager:spawnBarricade( tpl, x, y, lane )

   --We use to use a temporary template (tpl2) which would be either tpl[self.game.backgroundType] or tpl
   -- the problem is the template address will always be different == no reuse
   local b    = nil;
   if self.game and self.game.backgroundType then
      b = self:newEntity( Barricade, self.layers[2], tpl[self.game.backgroundType] or tpl, x, y, 0, lane );
   else
      b = self:newEntity( Barricade, self.layers[2], tpl, x, y, 0, lane );
   end


   table.insert(self.barricades, b);

   return b;

end
----------------------------------------------------------------------

function EntityManager:spawnProjectile( tpl, x, y, a, owner, target, parent, pOffset  )

   local p = self:newEntity( tpl.factory or Projectile, self.layers, tpl,  x, y, a, owner, target, parent, pOffset );
   
   table.insert(self.projectiles, p);

   return p;

end

----------------------------------------------------------------

function EntityManager:spawnVehicle( tpl, x, y, a, type, lane, ... )

   local m = self:newEntity(  tpl.factory or Vehicle, self.layers, tpl,  x, y, a, type, ...  );
   
   --TODO: Clean!!
   m.lane = lane;

   table.insert(self.vehicles, m);

   --TODO: Clean!!
   if type and type > 0 then
      self.game.vehicleSpawnedType[type] = self.game.vehicleSpawnedType[type] + 1;
   end

   if lane and lane >= 1 and lane <= 5 then
      self.game.lanes[lane].delay   = 0;
      table.insert( self.game.lanes[lane].vehicle, m );
   end

   if m.type then
      self.game.vehicleSpawnedTime[m.type] = 0;
   end
   
   return m;
   
end

----------------------------------------------------------------

function EntityManager:spawnChopper( tpl, x, y, a, ... )

   local m = self:newEntity(  tpl.factory or VehicleChopper, self.layers, tpl,  x, y, a, -1, ...  );
   local cSounds = {loop={"HeliLoop",},};
   EventMgr:postEvent( { sounds=cSounds, name="spawn chopper",}, "add-sfx" );
   
   table.insert(self.vehicles, m);

   return m;
   
end

----------------------------------------------------------------

function EntityManager:killVehicle( m  )

   if m.lane and m.lane >=1 and m.lane <= 5 then

      for i=#self.game.lanes[m.lane].vehicle,1,-1 do
         if self.game.lanes[m.lane].vehicle[i] == m then
            
            table.remove(self.game.lanes[m.lane].vehicle,i);

      
         end
      end
   end

   if m.type and m.type > 0 then
      self.game.vehicleSpawnedType[m.type] = self.game.vehicleSpawnedType[m.type] - 1;
   end


end
----------------------------------------------------------------

function EntityManager:spawnExplosion( tpl, x, y, a, parent, flipHoriz )

   local e = self:newEntity(  Explosion, self.layers[tpl.layer or 5], tpl,  x, y, a, flipHoriz );

	if parent then
		e.prop:setAttrLink ( MOAIProp2D.INHERIT_TRANSFORM, parent.prop, MOAIProp2D.TRANSFORM_TRAIT );
	end

   table.insert( self.explosions, e);
   
   EventMgr:postEvent( {sounds=e.sounds, name="spawnExplosion",}, "add-sfx" );

   return e;
end

----------------------------------------------------------------

function EntityManager:spawnPickup( tpl, x, y, index, count, speedK  )

   if tpl then

      --We can "combo" the score multiplier
      if tpl.type == "powerup_scoreK" then
         if self.game and self.game.avatar then
            tpl = tpl[math.clamp(self.game.avatar:getCurrentPickupK(),1,#tpl)];
         end
      end

      local p   = self:newEntity(  tpl.factory or Pickup, self.layers[5],  tpl,  x, y, 0, self.game.boundary, index, count, speedK   );

      self.game.pickupSpawnTime[tpl.type] = self.game.duration;

      table.insert( self.pickups, p);

      return p;

   end

   return nil;

end

----------------------------------------------------------------

function EntityManager:spawnPowerup( factory, tpl, x, y, a, owner  )

   local p   = self:newEntity(   factory, self.layers[tpl.layer],  tpl,  x, y, a, owner  );

   table.insert( self.powerups, p);
   
   return p;
   
end

----------------------------------------------------------------

function EntityManager:spawnFlyout( tpl, x, y, keepInsde  )

   local e = self:newEntity(   Flyout, self.layers[6],  tpl,  x, y, 0  );

   if keepInsde then

      local width = e:getWidth();

      if x + width/2 > self.game.boundary.max.x then

         e:setLocX( self.game.boundary.max.x - width/2 );

      elseif x - width/2 < self.game.boundary.min.x then

         e:setLocX( self.game.boundary.min.x + width/2 );

      end
   end

   table.insert( self.flyouts, e );

   return e;

end

----------------------------------------------------------------

function EntityManager:spawnFlyoutText( tpl, x, y, text  )

   local e   = self:newEntity( FlyoutText, self.layers[6],  tpl,  x, y, 0, text  );

   table.insert( self.flyouts, e);

   return e;

end

----------------------------------------------------------------------
