require('game/entity/drone/drone');

local MOUNT_POINT_LEFT			= "left";
local MOUNT_POINT_RIGHT			= "right";

local DRONE_DISABLED				= "disabled";
local DRONE_RESPAWN_COOLDOWN	= "respawn-cooldown";
local DRONE_RESPAWN				= "respawn";
local DRONE_READY					= "ready";
local DRONE_CHARGED           = "charged"
local DRONE_ATTACK				= "attack";

local BEAM_PULSE_FREQ			= 1.0;
local BEAM_PULSE_WIDTH			= 0.7;
local BEAM_PULSE_ALPHA			= 0.6;
local BEAM_FADE_TIME          = 10/30;
local BEAM_LENGTH_K           = 1.2;

local BEAM_STATE_OFF          = "off"
local BEAM_STATE_EXTEND       = "extend";
local BEAM_STATE_FADE         = "fade";

local PULSE_TIME					= 0.3;

----------------------------------------------------------------------

local Base        = Drone;
DroneBarricade   = Base:new();

----------------------------------------------------------------------


function DroneBarricade:new( game, layers, template,  x, y, angle, parent, offset, mountPoint, garage_index, visData, stats )

	-- modify base template to fit player vehicle
	if template and garage_index and visData then
		local rarity;
		local level;

		if mountPoint == MOUNT_POINT_LEFT then
			rarity	= math.min( DroneMgr:getDroneRarityLeft(garage_index), DroneMgr:getDroneMaxRarity() );
			level		= math.min( DroneMgr:getDroneLevelLeft(garage_index), DroneMgr:getDroneMaxLevel());
		else
			rarity	= math.min( DroneMgr:getDroneRarityRight(garage_index), DroneMgr:getDroneMaxRarity() );
			level		= math.min( DroneMgr:getDroneLevelRight(garage_index), DroneMgr:getDroneMaxLevel());
		end

		template.chargeThreshold               = visData.chargeThreshold[level];

      template.projectile.width              = visData.projectile[rarity].width;
      template.projectile.height             = visData.projectile[rarity].height;
      template.projectile.collisionWidth     = visData.projectile[rarity].collisionWidth;
      template.projectile.collisionHeight    = visData.projectile[rarity].collisionHeight;
      template.projectile.texture            = visData.projectile[rarity].texture;
      template.projectile.hitMarker.texture  = visData.projectile[rarity].hitMarker.texture;
      template.projectile.maxLifeTime        = visData.maxLifeTime[rarity];

      template.beam.texture                  = visData.beam[rarity].texture;
      template.beam.offset                   = { template.beam.offset[1] * visData.beam[rarity].offset[1], template.beam.offset[2] * visData.beam[rarity].offset[2] };
      template.beam.maxLengthK               = BEAM_LENGTH_K * ( ( (2000/30) * (template.projectile.maxLifeTime*30) ) / (template.beam.height/2) );
      template.beam.width                    = template.projectile.width;
      template.beam.extendTime               = (2000*template.projectile.maxLifeTime) / (template.beam.height/2);

      template.muzzleGlow.animId             = visData.muzzleGlow[rarity].animId;
      template.muzzleGlow.offset             = { template.muzzleGlow.offset[1] * visData.muzzleGlow[rarity].offset[1], template.muzzleGlow.offset[2] * visData.muzzleGlow[rarity].offset[2] };

		template.hitpoints                     = parent.hitPoints;

		local count = (template.bulletCount or 1) / ( (template.bulletCount or 1) * (template.bulletDelay or 0)  + (template.shootDelay or 0) );

		template.projectile.damage		= template.projectile.damage * ( App.data:getCurrentPlayerMainWeaponMaxDPS() / ( stats.rarity[rarity] - stats.level[level] ) ) / count;
	end

	local e = Base.new( self, game, layers, template, x, y, angle, parent, offset, mountPoint, garage_index );

   if template and mountPoint then

		e._pipOn				= {};
		e._pipOff			= {};

		for i=1,#template.pip_on.offset do
			e._pipOn[i]    = e:_createVisProp( layers[template.pip_on.layer or template.layer or 5], template.pip_on.priority, template.pip_on.offset[i][1], template.pip_on.offset[i][2], 0, template.pip_on.width, template.pip_on.height, template.pip_on.texture );
			e._pipOff[i]	= e:_createVisProp( layers[template.pip_off.layer or template.layer or 5], template.pip_off.priority, template.pip_off.offset[i][1], template.pip_off.offset[i][2], 0, template.pip_off.width, template.pip_off.height, template.pip_off.texture );
			e._pipOn[i]:setAttrLink( MOAIProp2D.INHERIT_LOC, e.prop, MOAIProp2D.TRANSFORM_TRAIT );
			e._pipOff[i]:setAttrLink( MOAIProp2D.INHERIT_LOC, e.prop, MOAIProp2D.TRANSFORM_TRAIT );
		end

      e.readyVFX = Base:newFlashAnim(layers[template.muzzleGlow.layer or template.layer or 5], template.muzzleGlow, e.prop);
      e.readyVFX:setLoop(true);
      --e.readyVFX:setSpan(1/20, 19/20);
      e.readyVFX:setVisible(false);

		e.beam = e:_createVisProp( layers[template.beam.layer or 5], template.beam.priority, template.beam.offset[1], template.beam.offset[2], 0, template.beam.width, template.beam.height, template.beam.texture, template.beam.scale,  template.beam.color, template.beam.blend );
		e.beam:setAttrLink( MOAIProp2D.INHERIT_LOC, e.prop, MOAIProp2D.TRANSFORM_TRAIT );
		e.beam:setVisible(false);

      --call here because avatar doesnt get called yet
      e:reset( x, y, angle, parent, offset  );
   end

   return e;

end

----------------------------------------------------------------------

function DroneBarricade:reset( x, y, angle, parent, offset )

   Base.reset(self, x, y, angle, parent, offset);

	self._chargeCounter	= 0;
	self._shootTimer		= 0;
	self._bulletTimer		= 0;
	self._bulletCounter	= self.template.bulletCount or 1;
	self._bulletOffset	= self.template.bulletDelta or { { x=0, y=0} };
	self._offsetIndex		= 1;
   self._activeTimer    = self.template.maxAbilityTime or 3;
   self._graceTimer     = self.template.maxGraceTime or 1;
	self._beamWTimer		= BEAM_PULSE_FREQ;
	self._beamLTimer		= 0;
   self._beamState      = BEAM_STATE_OFF;
	self._pipTimer			= 0;
	self._pipTracker		= {false, false, false};

end

----------------------------------------------------------------------

function DroneBarricade:update( dt, avatarSpeedY, burstOn, canShoot )

	Base.update(self, dt, avatarSpeedY);

	self:_updateChargeCounter(dt, canShoot);
	self:_vfxUpdate(dt);
	self:_updatePipPulse(dt);
   self:_checkBarricades(dt);
   self:_updateBeam(dt);
   self:_updateShoot(dt, burstOn);

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneBarricade:_updateChargeCounter( dt, canShoot )

   Debug.assert(self, "DroneBoost:_chargeUpdate -> self nil");
   Debug.assert(self.game, "DroneBoost:_chargeUpdate -> self.game nil");
   Debug.assert(self.game.avatar, "DroneBoost:_chargeUpdate -> self.game.avatar nil");

   if    not self
      or not self.game
      or not self.game.avatar
      or     self.game.avatar:isDead() then

      if self.readyVFX then
         self.readyVFX:hide();
      end

      return;
   end

	if self.state == DRONE_READY then

		self._chargeCounter = math.min(self._chargeCounter + dt, self.template.chargeThreshold);

		self:_updatePips();
      EventMgr:postEvent( {sounds = {loop={"BarricadeDroneChargedLoop",},}, name="barricade drone charged"}, "loop-stop" );

      if self:isCharged() then
         EventMgr:postEvent(	{	text="Barricade Drone Charged"	}, "hud-alert-text"	);
         EventMgr:postEvent( {sounds = {loop={"BarricadeDroneChargedLoop",},}, name="barricade drone charged"}, "add-sfx" );
         EventMgr:postEvent( {sounds = {start={"BarricadeDroneChargedStart",},}, name="barricade drone init charge"}, "add-sfx" );

         if self.readyVFX then
            self.readyVFX:show();
            self.readyVFX:start();
         end

         self.state = DRONE_CHARGED;
      else

         if self.readyVFX then
            self.readyVFX:hide();
         end
      end
   elseif self.state == DRONE_CHARGED then

   elseif self.state == DRONE_ATTACK then

      if self.readyVFX then
         self.readyVFX:hide();
      end
	else

      if self.readyVFX then
         self.readyVFX:hide();
      end
   end

end

----------------------------------------------------------------------

function DroneBarricade:_checkBarricades( dt )

   if       self.state == DRONE_CHARGED
      and   self.game.entityMgr.barricades
      and   #self.game.entityMgr.barricades > 0
      and   not self.game.avatar:isDead() then

      local x,y = self.game.avatar:getLoc();

      for _,v in ipairs(self.game.entityMgr.barricades) do

         if v:isObstacleOnScreen(0.38) then
            local bx,by = v:getLoc();

            if math.abs(x-bx) <= self.template.maxActivationX then

               self.state        = DRONE_ATTACK;
               self._beamState   = BEAM_STATE_EXTEND;

               EventMgr:postEvent( {sounds = {loop={"BarricadeDroneChargedLoop",},}, name="barricade drone charged"}, "loop-stop" );
               EventMgr:postEvent( {sounds = {start={"BarricadeDroneFire",},}, name="barricade drone fire"}, "add-sfx" );

               return;
            end
         end
      end
   elseif         self.state == DRONE_ATTACK
            and   not self.game.avatar:isDead() then

      Debug.assert(self.game.entityMgr.barricades, "DroneBarricade:_checkBarricades -> nil barricades");

      self._activeTimer = math.max(self._activeTimer - dt, 0);

      local bActive = false;

      for _,v in ipairs(self.game.entityMgr.barricades) do

         if v:isObstacleOnScreen(0.38) then
            bActive           = true;
         end
      end

      if not bActive then
         self._graceTimer  = math.max(self._graceTimer - dt, 0);
      end

      if    (  not bActive and self._graceTimer <= 0 )
         or ( self._activeTimer <= 0 ) then

         self:resetCharge();
      end
   end

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneBarricade:_vfxUpdate( dt )

	if self.state == DRONE_ATTACK and not self.game.avatar:isDead() then

	elseif self.game.avatar:isDead() then

		self:resetCharge();

		self._pipTracker	= {	false, false, false	};

		for i=1,#self._pipOn do
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(false);
		end

      self._beamState = BEAM_STATE_OFF;

      EventMgr:postEvent( {sounds = {loop={"BarricadeDroneChargedLoop",},}, name="barricade drone charged"}, "loop-stop" );

	elseif self.state == DRONE_RESPAWN or self.state == DRONE_RESPAWN_COOLDOWN then

		self:resetCharge();

		self._pipTracker	= {	false, false, false	};

		for i=1,#self._pipOn do
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(true);
		end

      EventMgr:postEvent( {sounds = {loop={"BarricadeDroneChargedLoop",},}, name="barricade drone charged"}, "loop-stop" );

      self._beamState = BEAM_STATE_OFF;
	end

end

----------------------------------------------------------------------

function DroneBarricade:_updatePipPulse( dt )

	if self._pipTracker then

		if self._pipTimer > PULSE_TIME then
			self._pipTimer	= 0;
		end

		local k;

		if self._pipTimer < PULSE_TIME/2 then
			k	= math.mapClamped(0, 1, PULSE_TIME/2, 1.5, self._pipTimer);
		else
			k	= math.mapClamped(PULSE_TIME/2, 1.5, PULSE_TIME, 1, self._pipTimer);
		end

		for i=1,#self._pipOn do
			if self._pipTracker[i] then
				self._pipOn[i]:setScl(k, k);
			else
				self._pipOn[i]:setScl(1, 1);
			end
		end

		self._pipTimer	= self._pipTimer + dt;
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneBarricade:_updateBeam( dt )

	self:_updateBeamLength(dt);
	self:_updateBeamPulse(dt);

end

----------------------------------------------------------------------

function DroneBarricade:_updateBeamLength( dt )

	if self._beamState == BEAM_STATE_OFF then

      self.beam:setVisible(false);

   elseif self._beamState == BEAM_STATE_FADE then

		self._beamLTimer	= math.min(self._beamLTimer + dt, BEAM_FADE_TIME);

		local w	= math.mapClamped(0, 1, BEAM_FADE_TIME, 0, self._beamLTimer);
		local a	= math.mapClamped(0, 0.8, BEAM_FADE_TIME, 0, self._beamLTimer);

      if a > 0 then
         self.beam:setVisible(true);
         self.beam:setScl(w, self.template.beam.maxLengthK);
         self.beam:setColor(1,1,1, a);
      else
         self.beam:setVisible(false);

         self._beamState = BEAM_STATE_OFF;
      end
   else
		self._beamLTimer	= math.min(self._beamLTimer + dt, self.template.beam.extendTime);

		local w,l = self.beam:getScl();

      l = self._beamLTimer/self.template.beam.extendTime;

		self.beam:setVisible(true);
		self.beam:setScl(w, l * self.template.beam.maxLengthK );
   end

end

----------------------------------------------------------------------

function DroneBarricade:_updateBeamPulse( dt, beamState )

	local hfreq	= BEAM_PULSE_FREQ/2;

   if self._beamState == BEAM_STATE_EXTEND then

		self._beamWTimer = self._beamWTimer + dt;

		if self._beamWTimer > BEAM_PULSE_FREQ then
			self._beamWTimer = 0;
		end

		local w,l	= self.beam:getScl();
		local a		= 1;

		if self._beamWTimer < hfreq then
			w	= math.mapClamped(0, 0.8, hfreq, BEAM_PULSE_WIDTH, self._beamWTimer);
			a	= math.mapClamped(0, 0.8, hfreq, BEAM_PULSE_ALPHA, self._beamWTimer);
		else
			w	= math.mapClamped(hfreq, BEAM_PULSE_WIDTH, BEAM_PULSE_FREQ, 0.8, self._beamWTimer);
			a	= math.mapClamped(hfreq, BEAM_PULSE_ALPHA, BEAM_PULSE_FREQ, 0.8, self._beamWTimer);
		end

		self.beam:setVisible(true);
		self.beam:setScl(w,l);
		self.beam:setColor(1,1,1,a);
   end

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneBarricade:_updatePips( percentage )

	local c				= percentage or math.min( self._chargeCounter/self.template.chargeThreshold, 1 );
	self._pipTracker	= {	false, false, false	};

   if self.state == DRONE_ATTACK then

      if c <= 1/3 then
         self._pipTracker	= { true, false, false};
      elseif c <= 2/3 then
         self._pipTracker	= { true, true, false};
      else
         self._pipTracker	= { true, true, true};
      end
   else

      if c < 1/3 then
         self._pipTracker	= { false, false, false};
      elseif c < 2/3 then
         self._pipTracker	= { true, false, false};
      elseif c < 1 then
         self._pipTracker	= { true, true, false};
      else
         self._pipTracker	= { true, true, true};
      end
   end

	for i=1,#self._pipOn do
		if self._pipTracker[i] then
			self._pipOn[i]:setVisible(true);
			self._pipOff[i]:setVisible(false);
		else
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(true);
		end
	end

end

----------------------------------------------------------------------

function DroneBarricade:resetCharge()

	self._chargeCounter	= 0;
	self._shootTimer		= 0;
	self._bulletTimer		= 0;
	self._bulletCounter	= self.template.bulletCount or 1;
	self._bulletOffset	= self.template.bulletDelta or { { x=0, y=0} };
	self._offsetIndex		= 1;
   self._activeTimer    = self.template.maxAbilityTime or 3;
   self._graceTimer     = self.template.maxGraceTime or 1;
	self._beamWTimer		= BEAM_PULSE_FREQ;
	self._beamLTimer		= 0;
	self._pipTracker		= { false, false, false };
   self.state           = DRONE_READY;
   self._beamState      = BEAM_STATE_FADE;

   if self.readyVFX then
      self.readyVFX:hide();
   end

	for i=1,#self._pipOn do
		if self._pipTracker[i] then
			self._pipOn[i]:setVisible(true);
			self._pipOff[i]:setVisible(false);
		else
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(true);
		end
	end

   EventMgr:postEvent( {sounds = {loop={"BarricadeDroneChargedLoop",},}, name="barricade drone charged"}, "loop-stop" );

end

----------------------------------------------------------------------

function DroneBarricade:isCharged()
   return self._chargeCounter >= self.template.chargeThreshold;
end

----------------------------------------------------------------------

function DroneBarricade:isChargeRequired()
	return true;
end

----------------------------------------------------------------------

function DroneBarricade:setCharge( k )

	self._chargeCounter = math.clamp( k*self.template.chargeThreshold, 0, self.template.chargeThreshold );

	self:_updateChargeCounter(0);

end

----------------------------------------------------------------------

function DroneBarricade:forceAbilityActivation()

   self.state        = DRONE_ATTACK;
   self._beamState   = BEAM_STATE_EXTEND;

   return true;

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneBarricade:_updateShoot( dt, burstOn )

	if self.state == DRONE_ATTACK then

		self._shootTimer = math.max(self._shootTimer - dt, 0 );

		if self._shootTimer <= 0 then

			self._bulletTimer	= self._bulletTimer - dt;

			if self._bulletTimer <= 0 then
				self:_shoot(burstOn);
			end
		end
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneBarricade:_getBulletOffset()

   local x,y	= self:getLoc(true);

	local delta = self._bulletOffset[self._offsetIndex];
	local dX,dY	= (delta.x or 0), (delta.y or 0);

	self._offsetIndex = self._offsetIndex + 1;

	if self._offsetIndex > #self._bulletOffset then
		self._offsetIndex		= 1;
	end

   return x+dX, y+dY, delta.a;

end

----------------------------------------------------------------------

function DroneBarricade:_shoot( burstOn )

	self._bulletCounter		= self._bulletCounter - 1;
	self._bulletTimer			= self.template.bulletDelay;

	if self._bulletCounter < 1 then
		self._shootTimer		= self.template.shootDelay;
		self._bulletTimer		= 0;
		self._bulletCounter	= self.template.bulletCount or 1;
	end

	local x,y,a	= self:_getBulletOffset();
	local p;

   p = self.game.entityMgr:spawnProjectile( self.template.projectile, self.template.muzzleGlow.offset[1], self.template.muzzleGlow.offset[2], 0, self.parent, nil, self );

	p.speedBurst = burstOn;

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneBarricade:delete()

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:delete();
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:delete();
		end
	end

	if self.beam then
		self.beam:delete();
	end

   if self.readyVFX then
      self.readyVFX:delete();
   end

	Base.delete(self);

end

----------------------------------------------------------------------

function DroneBarricade:hide()

	Base.hide(self);

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:setVisible(false);
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:setVisible(false);
		end
	end

	if self.beam then
		self.beam:setVisible(false);
	end

   if self.readyVFX then
      self.readyVFX:hide();
   end

end

----------------------------------------------------------------------

function DroneBarricade:show()

	Base.show(self);

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:setVisible(false);
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:setVisible(true);
		end
	end

	if self.beam then
		self.beam:setVisible(false);
	end

   if self.readyVFX then
      self.readyVFX:hide();
   end

end


----------------------------------------------------------------------
----------------------------------------------------------------------