require('game/entity/drone/drone');

local MOUNT_POINT_LEFT			= "left";
local MOUNT_POINT_RIGHT			= "right";

local DRONE_DISABLED				= "disabled";
local DRONE_RESPAWN_COOLDOWN	= "respawn-cooldown";
local DRONE_RESPAWN				= "respawn";
local DRONE_READY					= "ready";
local DRONE_CHARGED_INIT		= "charged-init";
local DRONE_REPAIRING			= "repairing";

local PULSE_TIME					= 0.3;

----------------------------------------------------------------------

local Base	= Drone;
DroneRepair	= Base:new();

----------------------------------------------------------------------


function DroneRepair:new( game, layers, template,  x, y, angle, parent, offset, mountPoint, garage_index, visData, stats )

	-- modify base template to fit player vehicle
	if template and garage_index and visData then
		local rarity;
		local level;

		if mountPoint == MOUNT_POINT_LEFT then
			rarity	= DroneMgr:getDroneRarityLeft(garage_index);
			level		= DroneMgr:getDroneLevelLeft(garage_index);
		else
			rarity	= DroneMgr:getDroneRarityRight(garage_index);
			level		= DroneMgr:getDroneLevelRight(garage_index);
		end

		template.bulletDelay					= stats.level.delay[level];
		template.healvfx.duration			= template.bulletDelay;
		template.healvfx.fadeStart			= 0.75 * template.healvfx.duration;
		template.hudvfx.duration			= template.healvfx.duration;
		template.hudvfx.fadeStart			= template.healvfx.fadeStart;
		template.heal							= stats.level.heal[level];
		template.chargeThreshold			= stats.rarity[rarity];
		template.beam.texture				= visData.beam[rarity];
	end

	local e = Base.new( self, game, layers, template, x, y, angle, parent, offset, mountPoint, garage_index );

   if template and mountPoint then

		if mountPoint == MOUNT_POINT_LEFT then
			e.beam	= e:_createVisProp( layers[template.beam.layer or 5], template.beam.priority, template.beam.offset[1], template.beam.offset[2], 0, template.beam.width, template.beam.height, template.beam.texture, template.beam.scale,  template.beam.color, template.beam.blend );
		else
			e.beam	= e:_createVisProp( layers[template.beam.layer or 5], template.beam.priority, -template.beam.offset[1], template.beam.offset[2], 0, template.beam.width, template.beam.height, template.beam.texture, template.beam.scale,  template.beam.color, template.beam.blend );
		end

		e.beam:setAttrLink( MOAIProp2D.INHERIT_TRANSFORM, e.prop, MOAIProp2D.TRANSFORM_TRAIT );
		e.beam:setVisible(false);

		e._speedAngle		= template.speedRot or {1, 360};
		e._baseOffset		= {	template.offset[1],		template.offset[2]	};

		e._readyVFX	= e:_createVisProp( layers[template.readyvfx.layer or template.layer or 5], template.readyvfx.priority, template.readyvfx.offset[1], template.readyvfx.offset[2], 0, template.readyvfx.width, template.readyvfx.height, template.readyvfx.texture, template.readyvfx.scale, template.readyvfx.color, template.readyvfx.blend );
		e._readyVFX:setAttrLink( MOAIProp2D.INHERIT_LOC, e.prop, MOAIProp2D.TRANSFORM_TRAIT );
		e._readyVFX:setVisible(false);

		e._pipOn				= {};
		e._pipOff			= {};

		for i=1,#template.pip_on.offset do
			e._pipOn[i]		= e:_createVisProp( layers[template.pip_on.layer or template.layer or 5], template.pip_on.priority, template.pip_on.offset[i][1], template.pip_on.offset[i][2], 0, template.pip_on.width, template.pip_on.height, template.pip_on.texture );
			e._pipOff[i]	= e:_createVisProp( layers[template.pip_off.layer or template.layer or 5], template.pip_off.priority, template.pip_off.offset[i][1], template.pip_off.offset[i][2], 0, template.pip_off.width, template.pip_off.height, template.pip_off.texture );
		end

      --call here because avatar doesnt get called yet
      e:reset( x, y, angle, parent, offset  );
   end

   return e;

end

----------------------------------------------------------------------

function DroneRepair:reset( x, y, angle, parent, offset )

   Base.reset(self, x, y, angle, parent, offset);

	self._chargeCounter	= 0;
	self._maxTurnRate		= self.template.speedRot or {90, 720};
	self._shootTimer		= 0;
	self._bulletTimer		= self.template.bulletDelay;
	self._bulletCounter	= self.template.bulletCount or 1;
	self._initTimer		= 0;
	self._vfxTimer			= 0;
	self._vfxBlinkTimer	= 0;
	self._pipTracker		= {false, false, false};
	self._pipTimer			= 0;
	self._beamLTimer		= 0;
	self._beamWTimer		= 0;
	self._beamExtendTime	= self.template.extendTime or 1/3;
	self._beamPulseFreq	= self.template.pulseFreq or 1;
	self._beamMinWidthK	= self.template.pulseMinWidthK or 0.5;
	self._beamMaxWidthK	= self.template.pulseMaxWidthK or 1.5;

	self.template.offset	= self._baseOffset or self.template.offset;

	if self._readyVFX then
		self._readyVFX:setVisible(false);
	end
   
   self.repairCooldownSFX = true;

end

----------------------------------------------------------------------

function DroneRepair:update( dt, avatarSpeedY, burstOn, canShoot )

	Base.update(self, dt, avatarSpeedY);

	self:_updateCharge(dt);
	self:_updateLoc(dt);
	self:_updateFacing(dt);
	self:_updateHeal(dt, burstOn);
	self:_updatePips();
	self:_vfxUpdate(dt);
	self:_updatePipLoc(dt);
	self:_updatePipPulse(dt);
	self:_updateBeam(dt);

end


----------------------------------------------------------------------
-- update drone angle and loc
----------------------------------------------------------------------

function DroneRepair:_updateLoc( dt )

	if self.state == DRONE_CHARGED_INIT and not self.game.avatar:isDead() then

		local x	= math.mapClamped(0, self._baseOffset[1], self.template.initTime, self._baseOffset[1]+self.template.modOffset[1], self._initTimer);
		local y	= math.mapClamped(0, self._baseOffset[2], self.template.initTime, self._baseOffset[2]+self.template.modOffset[2], self._initTimer);

		self.template.offset	= {x, y};

		if self._initTimer <= 0 then
			EventMgr:postEvent(	{	text="Repair Drone Activated"	}, "hud-alert-text"	);
		end

		if self._initTimer >= self.template.initTime then
			self.state				= DRONE_REPAIRING;
			self.template.offset	= {	self._baseOffset[1] + self.template.modOffset[1], self._baseOffset[2] + self.template.modOffset[2]	};
		end

		self._initTimer	= math.min( self._initTimer + dt, self.template.initTime );
	elseif self.state == DRONE_READY and not self.game.avatar:isDead() then

		if self._initTimer > 0 then

			self._initTimer	= math.max( self._initTimer - dt, 0 );

			local x	= math.mapClamped(0, self._baseOffset[1], self.template.initTime, self._baseOffset[1]+self.template.modOffset[1], self._initTimer);
			local y	= math.mapClamped(0, self._baseOffset[2], self.template.initTime, self._baseOffset[2]+self.template.modOffset[2], self._initTimer);

			self.template.offset = {x, y};
		else
			self.template.offset	= self._baseOffset or self.template.offset;
		end
	elseif		self.state == DRONE_DISABLED
				or	self.state == DRONE_RESPAWN_COOLDOWN
				or self.state == DRONE_RESPAWN then

		self._initTimer		= 0;
		self.template.offset	= self._baseOffset or self.template.offset;
	end

end

----------------------------------------------------------------------

function DroneRepair:_updateFacing( dt )

   if			self.game.avatar
		and	(	self.state == DRONE_READY
		or			self.state == DRONE_CHARGED_INIT
		or			self.state == DRONE_REPAIRING ) then

		local x, y		= self:getLoc(true);
		local tx,ty		= self.game.avatar:getLoc();

		if self.state == DRONE_READY then
			tx = x;
			ty	= y + 100;
		end

		local dx, dy	= tx-x, ty-y;
		self.angle		= self.angle or self:getRot();

		--Calculate around Y
		local rad		= -math.atan2(dx,dy);
		local a			= rad*180/math.pi;
		local da			= math.clampAngle180( a-self.angle );
		local rate		= math.mapClamped(0, self._maxTurnRate[1], self.template.initTime, self._maxTurnRate[2], self._initTimer);

		if self.state == DRONE_READY then
			rate			= math.mapClamped(0, self._maxTurnRate[2], self.template.initTime, self._maxTurnRate[1], self._initTimer);
		end

		if rot ~= a then

			local da2   = math.clamp(da, -rate*dt, rate*dt);

			self.angle  = self.angle+da2;
			self.da     = da2;

			self:setRot( self.angle );
		end
   else
		self:setRot(0);
	end

end

----------------------------------------------------------------------
-- update vfx
----------------------------------------------------------------------

function DroneRepair:_vfxUpdate( dt )

	if self.state == DRONE_CHARGED_INIT and not self.game.avatar:isDead() then
   
      self.repairCooldownSFX = false;
   
      if not self.repairWarmupSFX then
         EventMgr:postEvent( {sounds={ start={"RepairDroneWarmup",},}, name="repair warmup",}, "add-sfx" );
         self.repairWarmupSFX = true;
      end
   
      if self.repairSFX then
         local lSounds = { loop={"RepairDroneLoop",},};
         EventMgr:postEvent( {sounds=lSounds, name="repairloop",}, "loop-stop" );
         self.repairSFX = false;
      end

		self._pipTracker	= {	true, true, true	};

		self._vfxBlinkTimer	= self._vfxBlinkTimer + dt;

		if self._vfxBlinkTimer > PULSE_TIME then
			self._vfxBlinkTimer = 0;
		end

		if self._vfxBlinkTimer >= PULSE_TIME/2 then
			self._pipTracker	= {	false, false, false	};
		end

		for i=1,#self._pipOn do
			if self._pipTracker[i] then
				self._pipOn[i]:setVisible(true);
				self._pipOff[i]:setVisible(false);
			else
				self._pipOn[i]:setVisible(false);
				self._pipOff[i]:setVisible(true);
			end
		end

		local k	= math.max((self._vfxBlinkTimer/2) / (PULSE_TIME/2), 0);

		if k > 1 then
			k = k-1;
		end

		local a	= math.max(1-k, 0);

		self._readyVFX:setVisible(true);
		self._readyVFX:setScl(k,k);
		self._readyVFX:setColor(0, 1, 0, a);
	elseif self.state == DRONE_REPAIRING and not self.game.avatar:isDead() then
   
      if not self.repairSFX then
         local lSounds = { loop={"RepairDroneLoop",},};
         EventMgr:postEvent( {sounds=lSounds, name="repairloop",}, "add-sfx" );
         self.repairSFX = true;
      end

		self._vfxBlinkTimer	= self._vfxBlinkTimer + dt;

		if self._vfxBlinkTimer > PULSE_TIME then
			self._vfxBlinkTimer = 0;
		end

		local k	= math.max((self._vfxBlinkTimer/2) / (PULSE_TIME/2), 0);

		if k > 1 then
			k = k-1;
		end

		local a	= math.max(1-k, 0);

		self._readyVFX:setVisible(true);
		self._readyVFX:setScl(k,k);
		self._readyVFX:setColor(0, 1, 0, a);

		local p	= self._bulletCounter/self.template.bulletCount;

		self:_updatePips(p);
	elseif			self.state == DRONE_DISABLED
				or		self.state == DRONE_RESPAWN_COOLDOWN
				or		self.state == DRONE_RESPAWN
				or		self.game.avatar:isDead() then

      if self.repairSFX then
         local lSounds = { loop={"RepairDroneLoop",},};
         EventMgr:postEvent( {sounds=lSounds, name="repairloop",}, "loop-stop" );
         self.repairSFX = false;
      end

		local dead	= self.game.avatar:isDead();

		self._beamLTimer	= 0;
		self._beamWTimer	= 0;

		self._readyVFX:setVisible(false);
		self:resetCharge();

		self._pipTracker	= {false, false, false};

		for i=1,#self._pipOn do
			self._pipOn[i]:setVisible(false);

			if dead then
				self._pipOff[i]:setVisible(false);
			else
				self._pipOff[i]:setVisible(true);
			end
		end
	else
   
      if not self.repairCooldownSFX then
         EventMgr:postEvent( {sounds={ start={"RepairDroneCooldown",},}, name="repair cooldown",}, "add-sfx" );
         self.repairCooldownSFX = true;
      end

      if self.repairSFX then
         local lSounds = { loop={"RepairDroneLoop",},};
         EventMgr:postEvent( {sounds=lSounds, name="repairloop",}, "loop-stop" );
         self.repairSFX = false;
      end
      
		self._readyVFX:setVisible(false);
	end

end

----------------------------------------------------------------------

function DroneRepair:_updateBeam( dt, burstOn, canShoot )

	local beamVFX = false;

	if	self.state == DRONE_REPAIRING then
		beamVFX = true;
	end

	self:_updateBeamLength(dt, beamVFX);
	self:_updateBeamPulse(dt, beamVFX);

end

----------------------------------------------------------------------

function DroneRepair:_updateBeamLength( dt, beamOn )

	local w,l = self.beam:getScl();

	if not beamOn then
		self._beamLTimer	= math.max(self._beamLTimer - 2*dt, 0);

		l	= self._beamLTimer/self._beamExtendTime;

		if l == 0 then
			self.beam:setVisible(false);
		else
			self.beam:setScl(w, l);
		end
	else
		self._beamLTimer = math.min(self._beamLTimer + dt, self._beamExtendTime);

		l	= self._beamLTimer/self._beamExtendTime;

		self.beam:setVisible(true);
		self.beam:setScl(w, l);
	end

end

----------------------------------------------------------------------

function DroneRepair:_updateBeamPulse( dt, beamOn )

	local w,l	= self.beam:getScl();
	local hfreq	= self._beamPulseFreq/2;

	if self._beamLTimer > 0 then

		local sx	= 1;

		if self._beamWTimer >= hfreq then
			sx	= math.mapClamped(hfreq, self._beamMaxWidthK, self._beamPulseFreq, self._beamMinWidthK, self._beamWTimer);
		else
			sx	= math.mapClamped(0, self._beamMinWidthK, hfreq, self._beamMaxWidthK, self._beamWTimer);
		end

		self.beam:setScl(sx, l);

		self._beamWTimer = math.min(self._beamWTimer + dt, self._beamPulseFreq);

		if self._beamWTimer == self._beamPulseFreq then
			self._beamWTimer = 0;
		end
	end

end


----------------------------------------------------------------------
-- update pips
----------------------------------------------------------------------


function DroneRepair:_updatePips( percentage )

	local c		= percentage or math.min( self._chargeCounter/self.template.chargeThreshold, 1 );

	self._pipTracker	= {	false, false, false	};

	if c < 1/3 then
		self._pipTracker	= { false, false, false};
	elseif c < 2/3 then
		self._pipTracker	= { true, false, false};
	elseif c < 1 then
		self._pipTracker	= { true, true, false};
	else
		self._pipTracker	= { true, true, true};
	end

	for i=1,#self._pipOn do
		if self._pipTracker[i] then
			self._pipOn[i]:setVisible(true);
			self._pipOff[i]:setVisible(false);
		else
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(true);
		end
	end

	if			self.state == DRONE_READY
		and	not self.game.avatar:isDead()
		and	self:isRepairCharged()
		and	self.game.avatar:getDamageReceived() > 0 then

		self.state				= DRONE_CHARGED_INIT;
		self._initTimer		= 0;
		self._vfxTimer			= 0;
		self._vfxBlinkTimer	= 0;
		self._beamLTimer		= 0;
		self._beamWTimer		= 0;

		EventMgr:postEvent(	{	text="Repair Drone Activated"	}, "hud-alert-text"	);
	end

end

----------------------------------------------------------------------

function DroneRepair:_updatePipLoc( dt )

	if self._pipOn and self._pipOff then

		local x,y	= self:getLoc();

		for k,v in ipairs(self._pipOn) do
			local x2	= x + self.template.pip_on.offset[k][1];
			local y2	= y +	self.template.pip_on.offset[k][2];

			v:setLoc(x2, y2);
		end

		for k,v in ipairs(self._pipOff) do
			local x2	= x + self.template.pip_off.offset[k][1];
			local y2	= y +	self.template.pip_off.offset[k][2];

			v:setLoc(x2, y2);
		end
	end

end

----------------------------------------------------------------------

function DroneRepair:_updatePipPulse( dt )

	if self._pipTracker then

		if self._pipTimer > PULSE_TIME then
			self._pipTimer	= 0;
		end

		local k;

		if self._pipTimer < PULSE_TIME/2 then
			k	= math.mapClamped(0, 1, PULSE_TIME/2, 1.5, self._pipTimer);
		else
			k	= math.mapClamped(PULSE_TIME/2, 1.5, PULSE_TIME, 1, self._pipTimer);
		end

		for i=1,#self._pipOn do
			if self._pipTracker[i] then
				self._pipOn[i]:setScl(k, k);
			else
				self._pipOn[i]:setScl(1, 1);
			end
		end

		self._pipTimer	= self._pipTimer + dt;
	end

end


----------------------------------------------------------------------
-- update charging and healing mechanic
----------------------------------------------------------------------


function DroneRepair:isRepairCharged()
   return self._chargeCounter >= self.template.chargeThreshold;
end

----------------------------------------------------------------------

function DroneRepair:_updateCharge( dt )

	if			self.state == DRONE_READY
		and	not self:isRepairCharged()
		and	self:hasGameStarted() then

		self._chargeCounter = math.min(self._chargeCounter + dt, self.template.chargeThreshold);

		if self:isRepairCharged() and self.game.avatar:getDamageReceived() <= 0 then
			EventMgr:postEvent(	{	text="Repair Drone Charged"	}, "hud-alert-text"	);
		end

		self:_updatePips();
	end

end

----------------------------------------------------------------------

function DroneRepair:resetCharge()

	self._chargeCounter	= 0;
	self._shootTimer		= 0;
	self._bulletTimer		= self.template.bulletDelay;
	self._bulletCounter	= self.template.bulletCount or 1;
	self._vfxTimer			= 0;
	self._vfxBlinkTimer	= 0;
	self._pipTracker		= {false, false, false};
	self._pipTimer			= 0;

	for i=1,#self._pipOn do
		self._pipOn[i]:setVisible(false);
		self._pipOff[i]:setVisible(true);
	end

	self._readyVFX:setVisible(false);

end

----------------------------------------------------------------------

function DroneRepair:_updateHeal( dt )

	if self.state == DRONE_REPAIRING then

		self._shootTimer = math.max(self._shootTimer - dt, 0);

		if self._shootTimer <= 0 then
			self._bulletTimer = math.max(self._bulletTimer - dt, 0);

			if self._bulletTimer <= 0 then
				self:_heal(burstOn);
			end
		end
	end

end

----------------------------------------------------------------------

function DroneRepair:_heal( burstOn )

	local reset	= false;

	self:setCharge( self._bulletCounter/self.template.bulletCount );

	self._bulletCounter		= self._bulletCounter - 1;
	self._bulletTimer			= self.template.bulletDelay;

	if self._bulletCounter < 1 then
		self._shootTimer		= self.template.shootDelay;
		self._bulletCounter	= self.template.bulletCount or 1;
		reset						= true;
	end

	-- post to avatar to heal and heal vfx on avatar and HUD
	if			self.template.heal
		and	self.template.healvfx
		and	self.template.hudvfx
		and	self.game.avatar then

		EventMgr:postEvent(	{	heal = self.template.heal, avatarVFX = self.template.healvfx, hudVFX = self.template.hudvfx	}, "player-avatar-heal"	);
	end

	if reset then
		self.state	= DRONE_READY;
      
      self.repairWarmupSFX = false;

		self:resetCharge();
	end

end

----------------------------------------------------------------------

function DroneRepair:isChargeRequired()
	return true;
end

----------------------------------------------------------------------

function DroneRepair:setCharge( k )

	self._chargeCounter = math.clamp( k*self.template.chargeThreshold, 0, self.template.chargeThreshold );

   self:_updatePips();

end


----------------------------------------------------------------------
-- show, hide, and delete
----------------------------------------------------------------------


function DroneRepair:delete()

	if self._readyVFX then
		self._readyVFX:delete();
	end

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:delete();
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:delete();
		end
	end

	if self.beam then
		self.beam:delete();
	end

	Base.delete(self);

end

----------------------------------------------------------------------

function DroneRepair:hide()

	Base.hide(self);

	if self._readyVFX then
		self._readyVFX:setVisible(false);
	end

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:setVisible(false);
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:setVisible(false);
		end
	end

	if self.beam then
		self.beam:setVisible(false);
	end

end

----------------------------------------------------------------------

function DroneRepair:show()

	Base.show(self);

	if self._readyVFX then
		self._readyVFX:setVisible(false);
	end

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:setVisible(false);
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:setVisible(true);
		end
	end

	if self.beam then
		self.beam:setVisible(false);
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------
