require('game/entity/drone/drone');

local MOUNT_POINT_LEFT			= "left";
local MOUNT_POINT_RIGHT			= "right";

local DRONE_DISABLED				= "disabled";
local DRONE_RESPAWN_COOLDOWN	= "respawn-cooldown";
local DRONE_RESPAWN				= "respawn";
local DRONE_READY					= "ready";

----------------------------------------------------------------------

local Base	= Drone;
DroneRocket	= Base:new();

----------------------------------------------------------------------


function DroneRocket:new( game, layers, template,  x, y, angle, parent, offset, mountPoint, garage_index, visData, stats )

	-- modify base template to fit player vehicle
	if template and garage_index and visData then

		local rarity;
		local level;

		if mountPoint == MOUNT_POINT_LEFT then
			rarity	= math.min( DroneMgr:getDroneRarityLeft(garage_index), #visData.bulletCount );
			level		= math.min( DroneMgr:getDroneLevelLeft(garage_index), #visData.projectile );
		else
			rarity	= math.min( DroneMgr:getDroneRarityRight(garage_index), #visData.bulletCount );
			level		= math.min( DroneMgr:getDroneLevelRight(garage_index), #visData.projectile );
		end


		template.bulletCount				= visData.bulletCount[rarity];
		template.bulletDelta				= visData.bulletDelta[rarity];
		template.projectile.texture	= visData.projectile[level];
		template.hitpoints				= parent.hitPoints;

		local count = (template.bulletCount or 1) / ( (template.bulletCount or 1) * (template.bulletDelay or 0)  + (template.shootDelay or 0) );

		template.projectile.damage		= template.projectile.damage * ( App.data:getCurrentPlayerMainWeaponMaxDPS() / ( stats.rarity[rarity] - stats.level[level] ) ) / count;
	end

	local e = Base.new( self, game, layers, template, x, y, angle, parent, offset, mountPoint, garage_index );

   if template and mountPoint then

      --call here because avatar doesnt get called yet
      e:reset( x, y, angle, parent, offset  );
   end

   return e;

end

----------------------------------------------------------------------

function DroneRocket:reset( x, y, angle, parent, offset )

   Base.reset(self, x, y, angle, parent, offset);

	self._shootTimer		= random.mapClamped(0, self.template.shootDelay);
	self._bulletTimer		= 0;
	self._bulletCounter	= self.template.bulletCount or 1;
	self._bulletOffset	= self.template.bulletDelta or { { x=0, y=0} };
	self._offsetIndex		= 1;

end

----------------------------------------------------------------------

function DroneRocket:update( dt, avatarSpeedY, burstOn, canShoot )

	Base.update(self, dt, avatarSpeedY);

	self:_updateShoot(dt, burstOn, canShoot);

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneRocket:_updateShoot( dt, burstOn, canShoot )

	if self.state == DRONE_READY and (canShoot or self._bulletCounter < self.template.bulletCount) then

		self._shootTimer = self._shootTimer - dt;

		if self._shootTimer < 0 then
      
         EventMgr:postEvent( {sounds = {start={	"DroneMissile",},}, name="drone shoot"}, "add-sfx" );

			if self._bulletCounter >= self.template.bulletCount then
				self.game.entityMgr:spawnExplosion(self.template.muzzleblastvfx,		self.template.muzzleblastvfx.offset[1],		self.template.muzzleblastvfx.offset[2],	0,	self);
				self.game.entityMgr:spawnExplosion(self.template.muzzleflashvfx,		self.template.muzzleflashvfx.offset[1],		self.template.muzzleflashvfx.offset[2],	0,	self);
				self.game.entityMgr:spawnExplosion(self.template.backblastvfx,			self.template.backblastvfx.offset[1],			self.template.backblastvfx.offset[2],		0,	self);

				delay(	self.template.bulletDelay or 3/30,
							function()
								if			self
									and	self.game
									and	self.game.avatar
									and	not self.game.avatar:isDead() then

									self.game.entityMgr:spawnExplosion(self.template.muzzleblastvfx,		-1*self.template.muzzleblastvfx.offset[1],	self.template.muzzleblastvfx.offset[2],	0,	self,	true);
									self.game.entityMgr:spawnExplosion(self.template.muzzleflashvfx,		-1*self.template.muzzleflashvfx.offset[1],	self.template.muzzleflashvfx.offset[2],	0,	self,	true);
									self.game.entityMgr:spawnExplosion(self.template.backblastvfx,			-1*self.template.backblastvfx.offset[1],		self.template.backblastvfx.offset[2],		0,	self,	true);
								end
							end );
			end

			self._bulletTimer	= self._bulletTimer - dt;

			if self._bulletTimer <= 0 then
				self:_shoot(burstOn);
			end
		end
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneRocket:_getBulletOffset()

   local x,y	= self:getLoc(true);

	local delta = self._bulletOffset[self._offsetIndex];
	local dX,dY	= (delta.x or 0), (delta.y or 0);

	self._offsetIndex = self._offsetIndex + 1;

	if self._offsetIndex > #self._bulletOffset then
		self._offsetIndex		= 1;
	end

   return x+dX, y+dY, delta.a;

end

----------------------------------------------------------------------

function DroneRocket:_shoot( burstOn )

	self._bulletCounter		= self._bulletCounter - 1;
	self._bulletTimer			= self.template.bulletDelay;

	if self._bulletCounter < 1 then
		self._shootTimer		= self.template.shootDelay;
		self._bulletTimer		= 0;
		self._bulletCounter	= self.template.bulletCount or 1;
	end

	local x,y,a	= self:_getBulletOffset();
	local p;

	p = self.game.entityMgr:spawnProjectile( self.template.projectile, x, y, a, self.parent );

	p.speedBurst = burstOn;

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneRocket:delete()

		if self.glow then
			for k,v in pairs(self.glow) do
				v:delete();
			end
		end

		Base.delete(self);

end

----------------------------------------------------------------------

function DroneRocket:hide()

	Base.hide(self);

	if self.glow then
		for k,v in pairs(self.glow) do
			v:setVisible(false);
		end
	end

end

----------------------------------------------------------------------

function DroneRocket:show()

	Base.show(self);

	if self.glow then
		for k,v in pairs(self.glow) do
			v:setVisible(false);
		end
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------