require('game/entity/drone/drone');

local MOUNT_POINT_LEFT			= "left";
local MOUNT_POINT_RIGHT			= "right";

local DRONE_DISABLED				= "disabled";
local DRONE_RESPAWN_COOLDOWN	= "respawn-cooldown";
local DRONE_RESPAWN				= "respawn";
local DRONE_READY					= "ready";
local DRONE_CHARGED				= "charged";

local SHIELD_SCALE_TIME			= 10/30;
local PULSE_TIME					= 0.3;

----------------------------------------------------------------------

local Base	= Drone;
DroneShield	= Base:new();

----------------------------------------------------------------------


function DroneShield:new( game, layers, template,  x, y, angle, parent, offset, mountPoint, garage_index, visData, stats )

	-- modify base template to fit player vehicle
	if template and garage_index and visData then
		local rarity;
		local level;

		if mountPoint == MOUNT_POINT_LEFT then
			rarity	= math.min( DroneMgr:getDroneRarityLeft(garage_index), DroneMgr:getDroneMaxRarity() );
			level		= math.min( DroneMgr:getDroneLevelLeft(garage_index), DroneMgr:getDroneMaxLevel());
		else
			rarity	= math.min( DroneMgr:getDroneRarityRight(garage_index), DroneMgr:getDroneMaxRarity() );
			level		= math.min( DroneMgr:getDroneLevelRight(garage_index), DroneMgr:getDroneMaxLevel());
		end

		template.hitpoints			= parent.hitPoints;
		template.shieldHP				= math.floor( App.data:getCurrentPlayerVehicleHPMin() * stats.level[level]);
		template.chargeThreshold	= stats.rarity[rarity];
		template.readyvfx.texture	= visData.readyvfx[rarity];
		template.hitMarker.texture	= visData.impact[rarity];
		template._rarity				= rarity;
	end

	local e = Base.new( self, game, layers, template, x, y, angle, parent, offset, mountPoint, garage_index );

   if template and mountPoint then

		e._baseOffset		= {	template.offset[1],		template.offset[2]	};

		e._readyVFX	= e:_createVisProp( layers[template.readyvfx.layer or template.layer or 5], template.readyvfx.priority, template.readyvfx.offset[1], template.readyvfx.offset[2], 0, template.readyvfx.width, template.readyvfx.height, template.readyvfx.texture, template.readyvfx.scale, template.readyvfx.color, template.readyvfx.blend );
		e._readyVFX:setAttrLink( MOAIProp2D.INHERIT_LOC, e.prop, MOAIProp2D.TRANSFORM_TRAIT );
		e._readyVFX:setVisible(false);

		e._pipOn				= {};
		e._pipOff			= {};

		for i=1,#template.pip_on.offset do
			e._pipOn[i]		= e:_createVisProp( layers[template.pip_on.layer or template.layer or 5], template.pip_on.priority, template.pip_on.offset[i][1], template.pip_on.offset[i][2], 0, template.pip_on.width, template.pip_on.height, template.pip_on.texture );
			e._pipOff[i]	= e:_createVisProp( layers[template.pip_off.layer or template.layer or 5], template.pip_off.priority, template.pip_off.offset[i][1], template.pip_off.offset[i][2], 0, template.pip_off.width, template.pip_off.height, template.pip_off.texture );
			e._pipOn[i]:setAttrLink( MOAIProp2D.INHERIT_LOC, e.prop, MOAIProp2D.TRANSFORM_TRAIT );
			e._pipOff[i]:setAttrLink( MOAIProp2D.INHERIT_LOC, e.prop, MOAIProp2D.TRANSFORM_TRAIT );
		end

      --call here because avatar doesnt get called yet
      e:reset( x, y, angle, parent, offset  );
   end

   return e;

end

----------------------------------------------------------------------

function DroneShield:reset( x, y, angle, parent, offset )

   Base.reset(self, x, y, angle, parent, offset);

	self._shieldDamage	= self.template.shieldHP or 0;
	self._chargeCounter	= 0;
	self._shieldTimer		= 0;
	self._vfxBlinkTimer	= 0;
	self._pipTimer			= 0;
	self._initTimer		= 0;
	self._pipTracker		= {false, false, false};

	if self.template and self.template.hitMarker and self.template._rarity then
		parent:addShieldDrone( self, self.template._rarity or 1, self.template.hitMarker );
	end

	if self._readyVFX then
		self._readyVFX:setVisible(false);
	end

end

----------------------------------------------------------------------

function DroneShield:update( dt, avatarSpeedY, burstOn, canShoot )

	Base.update(self, dt, avatarSpeedY);

	self:_updateChargeCounter(dt);
	self:_updateLoc(dt);
	self:_vfxUpdate(dt);
	self:_updatePipPulse(dt);

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneShield:_updateChargeCounter( dt )

	if self.state == DRONE_READY and not self:isShieldCharged() then

		self._chargeCounter = math.min(self._chargeCounter + dt, self.template.chargeThreshold);

		self:_updatePips();

		if self:isShieldCharged() then
			self:_initShield();
		end
	elseif self.state == DRONE_READY and self:isShieldCharged() then
		self:_initShield();
	end

end

----------------------------------------------------------------------

function DroneShield:_initShield()

	self.state				= DRONE_CHARGED;
	self._shieldDamage	= 0;

   EventMgr:postEvent( {sounds = {start={"ShieldDroneChargedStart",},}, name="shield activate"}, "add-sfx" );

	delay(   0.75,
            function()
               if self:isShieldCharged() then
                  EventMgr:postEvent(	{	text="Shield Drone Activated"	}, "hud-alert-text"	);
                  EventMgr:postEvent( {sounds = {loop={"ShieldDroneChargedLoop",},}, name="shield charged loop"}, "add-sfx" );
               end
            end   );

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneShield:_updateLoc( dt )

	if self.state == DRONE_CHARGED and not self.game.avatar:isDead() then

		local x	= math.mapClamped(0, self._baseOffset[1], self.template.initTime, self._baseOffset[1]+self.template.modOffset[1], self._initTimer);
		local y	= math.mapClamped(0, self._baseOffset[2], self.template.initTime, self._baseOffset[2]+self.template.modOffset[2], self._initTimer);

		self.template.offset	= {x, y};

		self._initTimer	= math.min( self._initTimer + dt, self.template.initTime );
	elseif self.state == DRONE_READY and not self.game.avatar:isDead() then

      EventMgr:postEvent( {sounds = {loop={"ShieldDroneChargedLoop",},}, name="shield charged loop"}, "loop-stop" );

		if self._initTimer > 0 then

			self._initTimer	= math.max( self._initTimer - dt, 0 );

			local x	= math.mapClamped(0, self._baseOffset[1], self.template.initTime, self._baseOffset[1]+self.template.modOffset[1], self._initTimer);
			local y	= math.mapClamped(0, self._baseOffset[2], self.template.initTime, self._baseOffset[2]+self.template.modOffset[2], self._initTimer);

			self.template.offset = {x, y};
		else
			self.template.offset	= {self._baseOffset[1], self._baseOffset[2]} or self.template.offset;
		end
	elseif		self.state == DRONE_DISABLED
				or	self.state == DRONE_RESPAWN_COOLDOWN
				or self.state == DRONE_RESPAWN then

		self._initTimer		= 0;
		self.template.offset	= {self._baseOffset[1], self._baseOffset[2]} or self.template.offset;

      EventMgr:postEvent( {sounds = {loop={"ShieldDroneChargedLoop",},}, name="shield charged loop"}, "loop-stop" );
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneShield:_vfxUpdate( dt )

	if self.state == DRONE_CHARGED and not self.game.avatar:isDead() then

		local k	= math.max((self._vfxBlinkTimer/2) / (PULSE_TIME/2), 0);

		if k > 1 then
			k = k-1;
		end

		local a	= math.max(1-k, 0);

		self._readyVFX:setVisible(true);
		self._readyVFX:setScl(k,k);
		self._readyVFX:setColor(1, 1, 1, a);

      self:_updatePips( (self.template.shieldHP - self._shieldDamage) / self.template.shieldHP );
	elseif self.game.avatar:isDead() then

		self:resetCharge();

		self._pipTracker	= {	false, false, false	};

		for i=1,#self._pipOn do
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(false);
		end

		self._readyVFX:setVisible(false);
      EventMgr:postEvent( {sounds = {loop={"ShieldDroneChargedLoop",},}, name="shield charged loop"}, "loop-stop" );
	elseif self.state == DRONE_RESPAWN or self.state == DRONE_RESPAWN_COOLDOWN then

		self:resetCharge();

		self._pipTracker	= {	false, false, false	};

		for i=1,#self._pipOn do
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(true);
		end

		self._readyVFX:setVisible(false);
	else
		self._readyVFX:setVisible(false);
	end

end

----------------------------------------------------------------------

function DroneShield:_updatePipPulse( dt )

	if self._pipTracker then

		if self._pipTimer > PULSE_TIME then
			self._pipTimer	= 0;
		end

		local k;

		if self._pipTimer < PULSE_TIME/2 then
			k	= math.mapClamped(0, 1, PULSE_TIME/2, 1.5, self._pipTimer);
		else
			k	= math.mapClamped(PULSE_TIME/2, 1.5, PULSE_TIME, 1, self._pipTimer);
		end

		for i=1,#self._pipOn do
			if self._pipTracker[i] then
				self._pipOn[i]:setScl(k, k);
			else
				self._pipOn[i]:setScl(1, 1);
			end
		end

		self._pipTimer	= self._pipTimer + dt;
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneShield:_updatePips( percentage )

	local c				= percentage or math.min( self._chargeCounter/self.template.chargeThreshold, 1 );
	self._pipTracker	= {	false, false, false	};

   if self.state == DRONE_CHARGED then

      if c <= 1/3 then
         self._pipTracker	= { true, false, false};
      elseif c <= 2/3 then
         self._pipTracker	= { true, true, false};
      else
         self._pipTracker	= { true, true, true};
      end
   else

      if c < 1/3 then
         self._pipTracker	= { false, false, false};
      elseif c < 2/3 then
         self._pipTracker	= { true, false, false};
      elseif c < 1 then
         self._pipTracker	= { true, true, false};
      else
         self._pipTracker	= { true, true, true};
      end
   end

	for i=1,#self._pipOn do
		if self._pipTracker[i] then
			self._pipOn[i]:setVisible(true);
			self._pipOff[i]:setVisible(false);
		else
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(true);
		end
	end

end

----------------------------------------------------------------------

function DroneShield:getShieldHPMax()
	return self.template.shieldHP;
end

----------------------------------------------------------------------

function DroneShield:getShieldDamage()

	if self.state == DRONE_CHARGED then
		return self._shieldDamage or 0;
	end

	return self.template.shieldHP;

end

----------------------------------------------------------------------

function DroneShield:onDamage( damage )

	if self.state == DRONE_CHARGED and damage and damage > 0 then
		self._shieldDamage = math.min(self._shieldDamage + damage, self.template.shieldHP);

		if	self._shieldDamage >= self.template.shieldHP then
         EventMgr:postEvent( {sounds = {start={"ShieldDroneChargedEnd",},}, name="shield end"}, "add-sfx" );
			self:resetCharge();
		end
	end

end

----------------------------------------------------------------------

function DroneShield:isShieldCharged()
   return self._chargeCounter >= self.template.chargeThreshold;
end

----------------------------------------------------------------------

function DroneShield:resetCharge()

	self._shieldDamage	= self.template.shieldHP or 0;
	self._chargeCounter	= 0;
	self._shieldTimer		= 0;
	self._vfxBlinkTimer	= 0;
	self._pipTracker		= {	false, false, false	};
	self.state				= DRONE_READY;

	for i=1,#self._pipOn do
		if self._pipTracker[i] then
			self._pipOn[i]:setVisible(true);
			self._pipOff[i]:setVisible(false);
		else
			self._pipOn[i]:setVisible(false);
			self._pipOff[i]:setVisible(true);
		end
	end

	self._readyVFX:setVisible(false);
   EventMgr:postEvent( {sounds = {loop={"ShieldDroneChargedLoop",},}, name="shield charged loop"}, "loop-stop" );

end

----------------------------------------------------------------------

function DroneShield:isChargeRequired()
	return true;
end

----------------------------------------------------------------------

function DroneShield:setCharge( k )

	self._chargeCounter = math.clamp( k*self.template.chargeThreshold, 0, self.template.chargeThreshold );

	self:_updateChargeCounter(0);

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function DroneShield:delete()

	if self._readyVFX then
		self._readyVFX:delete();
	end

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:delete();
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:delete();
		end
	end

	Base.delete(self);

end

----------------------------------------------------------------------

function DroneShield:hide()

	Base.hide(self);

	if self._readyVFX then
		self._readyVFX:setVisible(false);
	end

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:setVisible(false);
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:setVisible(false);
		end
	end

end

----------------------------------------------------------------------

function DroneShield:show()

	Base.show(self);

	if self._readyVFX then
		self._readyVFX:setVisible(false);
	end

	if self._pipOn then
		for k,v in pairs(self._pipOn) do
			v:setVisible(false);
		end
	end

	if self._pipOff then
		for k,v in pairs(self._pipOff) do
			v:setVisible(true);
		end
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------