require('engine/anim/anim');
require('game/entity/entity');

local MOUNT_POINT_LEFT			= "left";
local MOUNT_POINT_RIGHT			= "right";

local DRONE_DISABLED				= "disabled";
local DRONE_RESPAWN_COOLDOWN	= "respawn-cooldown";
local DRONE_RESPAWN				= "respawn";
local DRONE_READY					= "ready";

local SCREEN_WIDTH_CUSHION		= 64;
local SCREEN_HEIGHT_CUSHION	= 32;

local MAX_DELTA_X					= 256;
local MAX_DELTA_Y					= 256;


----------------------------------------------------------------------

local Base	= Entity;
Drone			= Entity:new();

----------------------------------------------------------------------


function Drone:new( game, layers, template,  x, y, angle, parent, offset, mountPoint, garage_index )

	if template and mountPoint then
		local e = Entity.new( self, game, layers[template.layer or 5], template, x, y, angle, parent, offset );

		e._respawnHeight		= -MAX_DELTA_Y;
		e._screenWidthBound	= { -App.viewportWidth/2+SCREEN_WIDTH_CUSHION, App.viewportWidth/2-SCREEN_WIDTH_CUSHION };
		e._screenHeightBound	= { -App.viewportHeight/2+(template.height/2), App.viewportHeight/2+template.height };
		e._mountPosition		= mountPoint;
		e._speedX				= { template.speedMin[1], template.speedMax[1] };
		e._speedY				= { template.speedMin[2], template.speedMax[2] };
		e._respawnTRng			= template.respawnCDRange or {0, 0};

		if template.priority then
			e.prop:setPriority(template.priority);
		end

		if template.shadow then
			e.shadow = e:_createVisProp( layers[template.shadow.layer or 1], template.shadow.priority, template.shadow.offset[1], template.shadow.offset[2], 0, template.shadow.width, template.shadow.height, template.shadow.texture, template.shadow.scale, template.shadow.color );
		end

		if mountPoint == MOUNT_POINT_LEFT then
			e._rarity	= DroneMgr:getDroneRarityLeft(garage_index);
			e._level		= DroneMgr:getDroneLevelLeft(garage_index);
		else
			e._rarity	= DroneMgr:getDroneRarityRight(garage_index);
			e._level		= DroneMgr:getDroneLevelRight(garage_index);
		end

      --call here because avatar doesnt get called yet
      e:reset( x, y, angle, parent, offset  );

		return e;
   end

	return Entity.new( self, game );

end

----------------------------------------------------------------------

function Drone:reset( x, y, angle, parent, offset )

   Entity.reset( self, x, y, angle );

   self.state         = DRONE_READY;
   self.angle         = angle;
   self.parent        = parent;
   self.parentOffset  = offset;
	self._respawnTRng	 = self.template.respawnCDRange or {0, 0};
	self.repsawnTimer	 = random.mapClamped(self._respawnTRng[1], self._respawnTRng[2]);

   if self.template.pivot then
      self.prop:setPiv(self.template.pivot[1],self.template.pivot[2]);
   end

	self:setLoc(x+self.template.offset[1], y+self.template.offset[2]);

	if			self.template.collisionWidth
		and	self.template.collisionWidth > 0
		and	self.template.collisionHeight
		and	self.template.collisionHeight > 0 then

		self._testCollision	= true;
	end

	EventMgr:rmvListener( self );
	EventMgr:addListener( self,	self.onRespawnBegin,	"revive-used"		);
	EventMgr:addListener( self,	self.onRespawnEnd,	"revive-stop"		);
	EventMgr:addListener( self,	self.onEnemyKilled,	"avatar-kill"		);
	EventMgr:addListener( self,	self.onPickup,			"avatar-pickup"	);
	EventMgr:addListener( self,	self.onAvatarDmg,		"avatar-damaged"	);
	EventMgr:addListener( self,	self.onInputSwipeUp,	"input-swipeUp"	);

end

----------------------------------------------------------------------

function Drone:update( dt, avatarSpeedY )

	self:_updateDroneLoc(dt);
	self:_updateDroneRespawnCooldown(dt);
	self:_updateCollision(dt);
	self:_updateShadow(dt);

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function Drone:_createVisProp( layer, priority, x, y, angle, w, h, textureName, scale, color, blend )

	if layer and w and h and textureName then

		local gfxQuad	= MOAIGfxQuad2D.new();
		local texture	= self.game:getTexture(textureName, true);
		local scl		= scale or {1,1};
		local clr		= color or {1,1,1,1};
		local width		= w * scl[1];
		local height	= h * scl[2];

		gfxQuad:setTexture( texture );
		gfxQuad:setRect( -width/2, -height/2, width/2, height/2 );

		if texture and texture._uv then
			--setUVRect inverse the image for us... go figure!?
			gfxQuad:setUVRect(      texture._uv.xMin or 0
										,  texture._uv.yMax or 1
										,  texture._uv.xMax or 1
										,  texture._uv.yMin or 0 );
		end

		local prop = MOAIProp2D.new();
		
		prop:setDeck( gfxQuad );
		prop:setPriority( priority or 0 );
		prop:setLoc( x, y );
		prop:setColor( clr[1], clr[2], clr[3], clr[4] );

		if blend == "add" then
			prop:setBlendMode( MOAIProp.GL_SRC_ALPHA, MOAIProp.GL_ONE );
		elseif blend == "mult" then
			prop:setBlendMode( MOAIProp.GL_DST_COLOR, MOAIProp.GL_ONE_MINUS_SRC_ALPHA );
		else
			prop:setBlendMode( MOAIProp.GL_SRC_ALPHA, MOAIProp.GL_ONE_MINUS_SRC_ALPHA );
		end

		layer:insertProp( prop );

		prop.delete	=	function()
								if layer and prop then
									layer:removeProp(prop);
									prop = nil;
								end
							end;

		prop._width		= w;
		prop._height	= h;

		return prop;
	end

end

----------------------------------------------------------------------

function Drone:_updateDroneLoc( dt )

	if			self.state ~= DRONE_DISABLED
		and	self.state ~= DRONE_RESPAWN_COOLDOWN then

		local x,y	= self:getLoc();
		local tx,ty = self.parent:getLoc();
		local dy		= SCREEN_HEIGHT_CUSHION;

		tx = tx + self.template.offset[1];
		ty = ty + self.template.offset[2];

		if self.state == DRONE_RESPAWN then
			if y >= ty*1.05 then
				self.state = DRONE_READY;
			else
				dy = self._respawnHeight;
			end
		end

		local nx = math.clamp(self:_calcNewLoc(x, tx, MAX_DELTA_X, self._speedX, dt), self._screenWidthBound[1], self._screenWidthBound[2]);
		local ny = math.clamp(self:_calcNewLoc(y, ty, MAX_DELTA_Y, self._speedY, dt), self._screenHeightBound[1]+dy, self._screenHeightBound[2]);

		self:setLoc(nx, ny);
	end

end

----------------------------------------------------------------------

function Drone:_calcNewLoc( position, target, maxDelta, speed, dt )

	if position == target then
		return target;
	end

	local dist			= math.abs(target-position);
	local spd			= math.mapClamped( 0, speed[1], maxDelta, speed[2], dist );
	local delta			= spd*dt;

	if position <= target then
		return math.clamp(position+delta, target-maxDelta, target);
	else
		return math.clamp(position-delta, target, target+maxDelta);
	end

end

----------------------------------------------------------------------

function Drone:_updateDroneRespawnCooldown( dt )

	if self.state == DRONE_RESPAWN_COOLDOWN then

		self.repsawnTimer = self.repsawnTimer - dt;

		if self.repsawnTimer <= 0 then
			self.state			= DRONE_RESPAWN;
			self.repsawnTimer	= random.mapClamped(self._respawnTRng[1], self._respawnTRng[2]);
		end
	end

end

----------------------------------------------------------------------

function Drone:_updateCollision( dt )

	if self._testCollision then

		local x,y			= self:getLoc();
		local cx,cy			= self:getCollisionSize();
		local vehicles		= self.game.entityMgr.vehicles;
		local hazards		= self.game.entityMgr.hazards;
		local barricades	= self.game.entityMgr.barricades;
		local proj			= self.game.entityMgr.projectiles;

		if vehicles and #vehicles > 0 then

			for i=#vehicles,1,-1 do

				local v = vehicles[i];

				if v:isColliding( x, y, cx, cy ) then

					self:onCollideVehicle();

					break;
				end
			end
		end

		if hazards and #hazards > 0 then

			for i=#hazards,1,-1 do

				local h = hazards[i];

				if h:isColliding( x, y, cx, cy ) then

					self:onCollideHazard();

					break;
				end
			end
		end

		if barricades and #barricades > 0 then

			for i=#barricades,1,-1 do

				local b = barricades[i];

				if b:isColliding( x, y, cx, cy ) then

					self:onCollideBarricade();

					break;
				end
			end
		end

		if proj and #proj > 0 then

			for i=#proj,1,-1 do

				local p = proj[i];

				if not p.deleteMe and p.damageAvatar and p:isColliding( x, y, cx, cy ) then

					self:onCollideProjectile();

					break;
				end
			end
		end
	end

end

----------------------------------------------------------------------

function Drone:_updateShadow( dt )

	if self.shadow then

		local x,y	= self:getLoc();
		local a		= self:getRot();

		x	= x + self.template.shadow.offset[1];
		y	= y + self.template.shadow.offset[2];

		self.shadow:setLoc(x, y);
		self.shadow:setRot(a);
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function Drone:onDie()

	if self.template.explosion then
		self.deathExplosion	= self.game.entityMgr:spawnExplosion( self.template.explosion, self:getLoc() );
		self.state				= DRONE_RESPAWN;
	end

end

----------------------------------------------------------------------

function Drone:onAvatarDmg( event )
	-- event.amount is the amount of damage
end

----------------------------------------------------------------------

function Drone:onCollideVehicle()

end

----------------------------------------------------------------------

function Drone:onCollideHazard()

end

----------------------------------------------------------------------

function Drone:onCollideBarricade()

end

----------------------------------------------------------------------

function Drone:onCollideProjectile()

end

----------------------------------------------------------------------

function Drone:onEnemyKilled( event, type )

end

----------------------------------------------------------------------

function Drone:onInputSwipeUp()

end

----------------------------------------------------------------------

function Drone:onRespawnBegin()

	local x,y	= self:getLoc();

	self.state	= DRONE_DISABLED;

	if self._screenHeightBound and self._screenHeightBound[1] then
		self:setLoc(x, self._screenHeightBound[1]+self._respawnHeight);
	end

	self:show();

end

----------------------------------------------------------------------

function Drone:onRespawnEnd()

	self.state	= DRONE_RESPAWN_COOLDOWN;

end

----------------------------------------------------------------------

function Drone:onPickup( event )

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function Drone:forceAbilityActivation()
   return true;
end

----------------------------------------------------------------------

function Drone:getMagnetRadius()
   return 0;
end

----------------------------------------------------------------------

function Drone:getMountPosition()
   return self._mountPosition;
end

----------------------------------------------------------------------

function Drone:getLevel()
	return self._level;
end

----------------------------------------------------------------------

function Drone:getRarity()
	return self._rarity;
end

----------------------------------------------------------------------

function Drone:getScoreK()
	return 0;
end

----------------------------------------------------------------------

function Drone:getTarget()
	return self._target;
end

----------------------------------------------------------------------

function Drone:isBoostCharged()
	return nil;
end

----------------------------------------------------------------------

function Drone:isRepairCharged()
	return nil;
end

----------------------------------------------------------------------

function Drone:isActivationManual()
	return false;
end

----------------------------------------------------------------------

function Drone:isChargeRequired()
	return false;
end

----------------------------------------------------------------------

function Drone:setCharge( k )
	return;
end

----------------------------------------------------------------------

function Drone:setChargeFull()
	return self:setCharge( 1 );
end

----------------------------------------------------------------------

function Drone:hasGameStarted()

	local curState = StateMgr:getCurState();

	if curState and curState.hasGameStarted then
		return curState:hasGameStarted();
	end

	return false;

end


----------------------------------------------------------------------
----------------------------------------------------------------------


function Drone:delete()

	if self.shadow then
		self.shadow:delete();
	end

	EventMgr:rmvListener( self );

	Base.delete(self);

end

----------------------------------------------------------------------

function Drone:hide()

	Base.hide(self);

	if self.shadow then
		self.shadow:setVisible(false);
	end

end

----------------------------------------------------------------------

function Drone:show()

	Base.show(self);

	if self.shadow then
		self.shadow:setVisible(true);
	end

end


----------------------------------------------------------------------
----------------------------------------------------------------------