require('engine/utility/math');
require('game/entity/hazard/hazard-airStrikeBomber');
require('game/entity/hazard/hazard-airStrikeTarget');
require('game/entity/entity');
require('game/entity/warning');

----------------------------------------------------------------------

local WARNING_POS_DELTA_Y        = -75;
local LAST_TARGET_POS_DELTA_Y    =  40;
local BOMBER_DELTA_Y             =  100;

----------------------------------------------------------------------

AirStrike = Entity:new();

----------------------------------------------------------------------

function AirStrike:new( game, layer, template, x, y, a, warningLayer, target )

   local e              = Entity.new( self, game, layer, template, x, y, a, target );

   e.trackTargetK       = template.trackTargetK    or 1.5;
   e.trackTargetTime    = template.trackTargetTime or 2.5;

   e.warning            = Warning:new(         game, warningLayer, template.warning, x, game.boundary.max.y+WARNING_POS_DELTA_Y, 0, e );
   e.bomber             = AirStrikeBomber:new( game, warningLayer, template.bomber,  x, y, 0, e );
   e.targets            = {};

   for i=1, #template.targets do

      e.targets[i] = AirStrikeTarget:new( game, warningLayer, template.targets[i],  x, y, 0, e );
  
   end

   --Entity.new will call reset but at that time, warning, bomber, targets wont exist, so call again

   e:reset( x, y, a, target );

   return e;
   
end

----------------------------------------------------------------------

function AirStrike:reset( x, y, angle, warningLayer, target )

   Entity.reset( self, x, y, angle );

   self.age             = 0;
   self.visible         = true;
   self.target          = target;
   self.done            = false;
   self.delayBomber     = self.template.delayBomber or 0;
   self.delayTarget     = self.template.delayTarget or 0;

   if self.warning then
      self.warning:reset(x, self.game.boundary.max.y+WARNING_POS_DELTA_Y, 0, self);
   end

   if self.bomber then
      self.bomber:reset(x, self.game.boundary.max.y+BOMBER_DELTA_Y, 0, self);
   end

   if self.targets then

      local tgtMaxY  = self.game.boundary.max.y    + WARNING_POS_DELTA_Y;
      local tgtMinY  = self.game.avatar:getLocY()  + LAST_TARGET_POS_DELTA_Y;
      local tgtY     = tgtMaxY;
      local incY     = (tgtMaxY-tgtMinY)/(#self.targets-1);

      for i=1, #self.targets do

         self.targets[i]:reset(x, tgtY, 0, self);

         tgtY = tgtY - incY;

      end
   end
end

----------------------------------------------------------------------

function AirStrike:delete()

   self.target = nil;

   if self.warning then
      self.warning:delete();
      self.warning = nil;
   end

   if self.bomber then
      self.bomber:delete();
      self.bomber = nil;
   end

   for i=1, #self.targets do
      self.targets[i]:delete();
   end
   self.targets = nil;


   Entity.delete( self );

end

----------------------------------------------------------------------

function AirStrike:update( dt )

   local speedK = math.lerp(1, self.game.gameSpeedK, 0.5);

   self.age = self.age + dt;

   self:trackTarget(dt);


   self.done         = self.warning:update(speedK*dt);

   if self.delayBomber > 0 then
      self.delayBomber = self.delayBomber - dt;
      self.done        = false;
   else
      self.done        = self.bomber:update(speedK*dt) and self.done;
   end

   if self.delayTarget > 0 then
      self.delayTarget = self.delayTarget-dt;
      self.done        = false;
   else

      for i=1, #self.targets do
         self.done     = self.targets[i]:update(speedK*dt) and self.done;
      end
   end

   return self.done;

end

----------------------------------------------------------------------

function AirStrike:trackTarget( dt )

   if self.target then

      local myX    = self:getLocX();
      local tgtX   = self.target:getLocX();

      local trackK =(dt*self.trackTargetK)
                  * (1-self.age/self.trackTargetTime);

      self:setLocX( math.lerp( myX,tgtX, math.clamp(trackK,0,1) ) );

   end
end

----------------------------------------------------------------------

function AirStrike:isDone()

   return self.done;

end

----------------------------------------------------------------------