require('game/entity/entity');
require('engine/anim/anim');
require('game/entity/hazard/hazard-missile-flame');

----------------------------------------------------------------------

local AIR_MISSILE_OBJECT_STATE_WAIT     = 1;
local AIR_MISSILE_OBJECT_STATE_GOING    = 2;
local AIR_MISSILE_OBJECT_STATE_DONE     = 3;

local MISSILE_OUT_OF_SCREEN_DELTA       = 200;

----------------------------------------------------------------------

local Base          = Entity;
HazardMissileObject = Base:new();

----------------------------------------------------------------------

function HazardMissileObject:new( game, layers, template, x, y, a, owner )

   local e = Base.new( self, game, layers[template.layer], template, x, y, a, owner );

   e.delay = template.delay;
   e.sounds = e.template.sounds;

   e:debugCollision( layers[template.layer] );

   if template.fire then
      e.fire = HazardMissileFlame:new( game, layers, template.fire, x, y, a, e )
   end

   if template.shadow then
      e.shadow = Entity:new( game, layers[template.shadow.layer], template.shadow, x, y, a  );
   end

   --make sure we reset to have all of our child being reset
   e:reset();
   e:hide();

   return e;

end


----------------------------------------------------------------------

function HazardMissileObject:delete(  )

   if self.fire then
      self.fire:delete();
      self.fire = nil;
   end

   if self.shadow then
      self.shadow:delete();
      self.shadow = nil;
   end

   Base.delete( self );

end

----------------------------------------------------------------------

function HazardMissileObject:reset( x, y, angle, owner )

   Base.reset( self, x, y, angle , owner );

   self.state           = AIR_MISSILE_OBJECT_STATE_WAIT;
   self.owner           = owner;
   self:setSpeedY( 0 );
   self.soundPlayed = false;
   self.random = math.random() * 0.25;

   if self.fire then
      self.fire:reset(x, y, angle, self);
   end
   
   if self.shadow then
      self.shadow:reset(1000,1000, 0  );
   end



end


----------------------------------------------------------------------

function HazardMissileObject:update ( dt, avatarSpeedY )

   Base.update( self, dt, avatarSpeedY );
   
   if self.fire then
      self.fire:update(dt, avatarSpeedY);
   end
   
   if self.state == AIR_MISSILE_OBJECT_STATE_WAIT then

      self:setLocX(self.owner:getLocX());
         
      if self:getAge() > self.delay then
         self:show();
         self:setSpeedY( self.template.speed[2] );
         self.state    = AIR_MISSILE_OBJECT_STATE_GOING;
         --EventMgr:postEvent( {sounds=self.sounds, name="HazardMissileObject",}, "add-sfx" );
         
      end

   elseif self.state == AIR_MISSILE_OBJECT_STATE_GOING then

      if self:getTop() + MISSILE_OUT_OF_SCREEN_DELTA < self.game.boundary.min.y then
      
         self:hide();
         self.state = AIR_MISSILE_OBJECT_STATE_DONE;
         
         EventMgr:postEvent( {}, "missile-avoided" );
         
		else

         local avatar = self.game.avatar;
      
			if (avatar) and (not avatar.noCollide) then

        		local collision = Collision:checkCollision( self, avatar )

        		if collision then

					avatar:onHitHazardMissile();

					self:onDestroyed( collision:getImpactLoc() );

               --Don't spawn if avatar is dead (or should be)
               if avatar:isInvicible() and (self.template.specialDropList) then

                  self.game:spawnPickupFromList( self.template.specialDropList, {self:getLocX(), self:getLocY()}, self.template.specialDropSpeedK  );

               end

				end
			end
		end
   end
   
   --[[if AIR_MISSILE_OBJECT_STATE_DONE then
      EventMgr:postEvent( {sounds=self.sounds, name="HazardMissileObject",}, "loop-stop" );
   end--]]

   if self.shadow then
      local x,y = self:getLoc();
      self.shadow:setLoc( x+self.template.shadow.offset[1], y+self.template.shadow.offset[2] );
   end
   
   return self.state == AIR_MISSILE_OBJECT_STATE_DONE;

end


----------------------------------------------------------------------

function HazardMissileObject:onDestroyed( impactLocX, impactLocY )

	self:hide();
	self.state = AIR_MISSILE_OBJECT_STATE_DONE;

	if not impactLocX or not impactLocY then
		self.game.entityMgr:spawnExplosion( self.template.explosion, self:getLoc() );
	else
		self.game.entityMgr:spawnExplosion( self.template.explosion, impactLocX, impactLocY );
	end
   
   local boosting = self.game:isSpeedBurstOn();
   EventMgr:postEvent( {invincible=boosting}, "missile-destroyed" );

end
----------------------------------------------------------------------

function HazardMissileObject:hide()

   Base.hide( self );

   if self.fire then
      self.fire:hide( );
   end

   if self.shadow then
      self.shadow:hide( );
   end

end

----------------------------------------------------------------------

function HazardMissileObject:show()

   Base.show( self );

   if self.fire then
      self.fire:show( );
   end

   if self.shadow then
      self.shadow:show( );
   end

end

----------------------------------------------------------------------

function HazardMissileObject:isDone (  )

   return self.state == AIR_MISSILE_OBJECT_STATE_DONE;

end