require('engine/utility/math');
require('game/entity/hazard/hazard-missile-line');
require('game/entity/hazard/hazard-missile-object');
require('game/entity/entity');
require('game/entity/warning');

----------------------------------------------------------------------

local WARNING_POS_DELTA_Y        = -190;
local LAST_TARGET_POS_DELTA_Y    =  40;
local OBJECT_DELTA_Y             =  250;

----------------------------------------------------------------------

local Base    = Entity;
HazardMissile = Base:new();

----------------------------------------------------------------------

function HazardMissile:new( game, layers, template, x, y, a, target )

   local e              = Base.new( self, game, layers[template.layer], template, x, y, a, target );

   e.trackTargetK       = template.trackTargetK    or 1.5;
   e.trackTargetTime    = template.trackTargetTime or 2.5;
   
   e.random = math.random() * 0.25;
   e.soundPlayed = false;
   e.sounds ={start={"HazardMissile1",},};

   e.line               = HazardMissileLine:new(   game, layers, template.line,    x, 0, 0, e );
   e.warning            = Warning:new(             game, layers, template.warning, x, game.boundary.max.y+WARNING_POS_DELTA_Y, 0, e );

   e.object             = HazardMissileObject:new( game, layers, template.object,  x, y, 0, e );

   --Base.new will call reset but at that time, warning, object, targets wont exist, so call again

   e:reset( x, y, a, target );

   return e;
   
end

----------------------------------------------------------------------

function HazardMissile:reset( x, y, angle, target )

   Base.reset( self, x, y, angle );

   self.visible         = true;
   self.target          = target;
   self.done            = false;
   self.delayObject     = self.template.delayObject or 0;
   self.delayTarget     = self.template.delayTarget or 0;
   self.delayLine       = self.template.delayLine   or 0;
	self.delaySound		= self.template.delaySound  or 0;

   self.random = math.random() * 0.25;
   self.soundPlayed = false;

   if self.warning then
      self.warning:reset(x, self.game.boundary.max.y+WARNING_POS_DELTA_Y, 0, self, target );
   end

   if self.line then
      self.line:reset(x, 0, 0, self);
   end

   if self.object then
      self.object:reset(x, self.game.boundary.max.y+OBJECT_DELTA_Y, 0, self);
   end
end

----------------------------------------------------------------------

function HazardMissile:deactivate()


   self.target = nil;

   if self.warning then
      self.warning:deactivate();
   end

   if self.line then
      self.line:deactivate();
   end

   if self.object then
      self.object:deactivate();
   end

   Base.deactivate( self );

end


----------------------------------------------------------------------

function HazardMissile:delete()

   self.target = nil;

   if self.warning then
      self.warning:delete();
      self.warning = nil;
   end

   if self.line then
      self.line:delete();
      self.line = nil;
   end

   if self.object then
      self.object:delete();
      self.object = nil;
   end

   Base.delete( self );

end


----------------------------------------------------------------------

function HazardMissile:hide()

   if self.warning then
      self.warning:hide();
   end

   if self.line then
      self.line:hide();
   end

   if self.object then
      self.object:hide();
   end

end


----------------------------------------------------------------------

function HazardMissile:show()

   if self.warning then
      self.warning:show();
   end

   if self.line then
      self.line:show();
   end

   if self.object then
      self.object:show();
   end

end

----------------------------------------------------------------------

function HazardMissile:selfDestruct()

   if self.warning then
		self.warning:stop();
   end

   if self.line then
		self.line.stop();
   end

   if self.object then
		self.object:onDestroyed();
   end

end

----------------------------------------------------------------------

function HazardMissile:update( dt, avatarSpeedY )

   Base.update(self, dt, avatarSpeedY);
   
   local speedK = math.lerp(1, self.game.gameSpeedK, 0.25);

   self:trackTarget(dt);

	if self.delaySound > 0 then
		self.delaySound = self.delaySound - dt;
	else
		if not self.soundPlayed then
			if self:getAge() > (self.trackTargetTime+self.random) then
				EventMgr:postEvent( {sounds=self.sounds, name="HazardMissileObject",}, "add-sfx" );
				self.soundPlayed = true;
			end
		end
	end
   
   self.done         = self.warning:update(speedK*dt);
   

   if self.delayObject > 0 then
      self.delayObject = self.delayObject - dt;
      self.done        = false;
   else
      self.done        = self.object:update(dt, avatarSpeedY) and self.done;
   end

   if self.delayLine > 0 then
      self.delayLine   = self.delayLine-dt;
      self.done        = false;
   else

      self.done     = self.line:update(speedK*dt) and self.done;

   end

   return self.done;

end

----------------------------------------------------------------------

function HazardMissile:trackTarget( dt )

   if self.target then

      local myX    = self:getLocX();
      local tgtX   = self.target:getLocX();

      local trackK =(dt*self.trackTargetK)
                  * (1-self:getAge()/self.trackTargetTime);

      self:setLocX( math.lerp( myX,tgtX, math.clamp(trackK,0,1) ) );

   end
end

----------------------------------------------------------------------

function HazardMissile:isDone()

   return self.done;

end

----------------------------------------------------------------------