----------------------------------------------------------------------
require('engine/object/object')
require('app/template/template');
require('game/maps/map01/backgroundCharts');
require('game/maps/map01/hazardCharts');
require('game/maps/map01/projectileCharts');
require('game/maps/map01/pickupCharts');
require('game/maps/map01/vehicleCharts');
----------------------------------------------------------------------

local VEHICLE_SPAWN_CHART      =  MAP01_VEHICLE_SPAWN_CHART;
local VEHICLE_SPAWN_DELAY_MIN  = 0.15;
local WAVE_DURATION            = 20;
local WAVE_DURATION_K          = (2*math.pi)/WAVE_DURATION;
local VEHICLE_SPEED_TIME       = { 000,  180 };
local VEHICLE_SPEED            = { 000, -600 };

----------------------------------------------------------------------

GameSpawner =
{
   --consumable
   spawnEnabled            = false,
   hazardEnabled           = true,
   laneEnabled             = { true, true, true, true, true },
   laneDisableTimer        = { 0,    0,    0,    0,    0    },
   vehicleSpawnDelay       = 0,
   checkLane               = math.random(1,5),
   hazardTimer             = 0,
};

----------------------------------------------------------------------

function GameSpawner:new( game )

   local gs = Object.new( self );


   gs.__marker              =  "game-spawner";
   gs.age                   = 0;
   gs.vehicleChartTime      = 0;
   gs.vehicleChartTarget    = 0;
   gs.vehicleChartCatchupK  = 10;
   gs.vehicleChartSlowK     = 1;
   gs.vehicleSpeedDropStart = 0;
   gs.vehicleChartMax       = 60;
   gs.wave                  = 0;
   gs.waveDelta             = math.random(0,1000);
   gs.game                  = game;
   gs.entityMgr             = game.entityMgr;
   gs.vehicleChartCatchupK  = 1;
   gs.vehicleChartSkip      = 0;
   gs.vehcTemplate          = gs:getVehicleTemplate();
   gs.vehcSpeedMap          = {};
   gs.bountySpawnList       = {};
   gs.bountyActiveVehicles  = {};

   return gs;

end


----------------------------------------------------------------------

function GameSpawner:onActivate()

end

----------------------------------------------------------------------

function GameSpawner:onDeactivate()

   self.game            = nil;
   self.entityMgr       = nil;

   EventMgr:rmvListener( self );

end

----------------------------------------------------------------------

function GameSpawner:_adjustVehicleSpeed( vehicle )

   if self.game.avatar then

      if vehicle.template.useLevelSpeed then

         local speed = self.vehcSpeedMap[vehicle.template];

         if speed == nil then

            local avatarSpeedK = vehicle.template.avatarSpeedK or 1;
            local delta        = math.mapClamped(  VEHICLE_SPEED_TIME[1], VEHICLE_SPEED[1],
                                                   VEHICLE_SPEED_TIME[2], VEHICLE_SPEED[2], self.age);

            speed = avatarSpeedK*((self.game.avatar:getSpeedMaxY()*self.game.gameSpeedK)+delta+self.vehicleSpeedDropStart);

            self.vehcSpeedMap[vehicle.template] = speed;

         end

         vehicle:setSpeedY( speed + math.random(0,50) );

      elseif vehicle.template.useAvatarStartSpeed then

         local avatarSpeedK = vehicle.template.avatarSpeedK or 1;
         local avatar       = self.game.avatar:getSpeedMaxY()*self.game.gameSpeedK;
         local delta        = vehicle.template.useAvatarStartSpeedDelta or 0;
         local speed        = avatar*avatarSpeedK + delta;

         vehicle:setSpeedY( speed + math.random(0,50) );

      end

   end

end

----------------------------------------------------------------------

function GameSpawner:queueBounty( bounty )

   self.bountySpawnList = self.bountySpawnList or {};

   self.bountyActiveVehicles = self.bountyActiveVehicles or {};

   if bounty and bounty.tangoId and bounty.score and self:_isBountyValid(bounty) then

      table.insert(self.bountySpawnList, bounty);
      self.bountyActiveVehicles[bounty.tangoId] = true;

      return true;
   end

   return false;
end

----------------------------------------------------------------------

function GameSpawner:_isBountyValid( bounty )

   if bounty.score < 1 then
      return false;
   end

   for k,v in pairs(self.bountySpawnList) do
      if v.tangoId == bounty.tangoId then
         return false;
      end
   end

   if self.bountyActiveVehicles and self.bountyActiveVehicles[bounty.tangoId] then
      return false;
   end

   return true;

end

----------------------------------------------------------------------

function GameSpawner:update( dt )

   self.age    = self.age + dt;
   self.wave   = math.clamp((math.sin((self.age+self.waveDelta)*WAVE_DURATION_K)+1.)/2,0,1);
   self.wave   = self.wave*self.wave*self.wave;

   if self.spawnEnabled then

      self:updateVehicleChartTime( dt );

      self:updateHazards( self.game.gameSpeedK*dt );

      self:updateVehicles( self.game.gameSpeedK*dt );

      for i=1,5 do

         if self.laneDisableTimer[i] > 0 then
            self.laneDisableTimer[i] = self.laneDisableTimer[i] - dt;
         end
      end

   else

      --Keep decreasing so when I re-enable, it will spawn faster
      self.vehicleSpawnDelay = self.vehicleSpawnDelay - dt;

   end
end

----------------------------------------------------------------------


function GameSpawner:updateVehicleChartTime( dt )

   if dt > 0 then

      local timeK       = 1;

      if self.vehicleChartTime < self.vehicleChartMax then

         if self.vehicleChartTime < self.vehicleChartTarget then

            local toTargetRel = self.vehicleChartTime/self.vehicleChartTarget;

            timeK = math.lerp( self.vehicleChartCatchupK, 1, toTargetRel);

         else

            local toMaxRel = (self.vehicleChartTime-self.vehicleChartTarget)
                            /(self.vehicleChartMax -self.vehicleChartTarget);

            --We want to slow down faster...
            timeK = math.lerp( 1, self.vehicleChartSlowK, toMaxRel);

         end

         self.vehicleChartTime = math.min(self.vehicleChartTime + timeK*dt, self.vehicleChartMax);

      end
   end

end

----------------------------------------------------------------------

function GameSpawner:updateVehicles( dt )

   self.vehicleSpawnDelay = self.vehicleSpawnDelay - dt;

   --When bursting (especially consumable head start), we want a lot of stuff on the screen
   if self:isSpeedModeOn() then
      self.vehicleSpawnDelay = 0;
   end

   if self.vehicleSpawnDelay <= 0 then

      local chosen = self:shouldLaneSpawn(self.checkLane);

      if chosen and self.laneEnabled[self.checkLane] and self.laneDisableTimer[self.checkLane] <= 0  then

         local vehicle_template = self.vehcTemplate;

         local v;

         if App.data:isBountyUnlocked() and self.bountySpawnList and #self.bountySpawnList > 0 then

            local t = self:_getBountyVehicleTemplate(self.bountySpawnList[1], vehicle_template[chosen]);

            v = self.game.entityMgr:spawnVehicle( t,
                                                  self.game:getLaneLocX(self.checkLane),
                                                  self.game.spawnVehicleY + t.height/2,
                                                  0,
                                                  chosen,
                                                  self.checkLane,
                                                  self.bountySpawnList[1]
            );

            EventMgr:postEvent( {player=self.bountySpawnList[1]}, "bounty-vehicle-spawned" );

            table.remove(self.bountySpawnList, 1);

         else
            v = self.game.entityMgr:spawnVehicle( vehicle_template[chosen],
                                                  self.game:getLaneLocX(self.checkLane),
                                                  self.game.spawnVehicleY + vehicle_template[chosen].height/2,
                                                  0,
                                                  chosen,
                                                  self.checkLane
            );
         end

         if v then
            self:_adjustVehicleSpeed( v );
         end

         self.vehicleSpawnDelay = VEHICLE_SPAWN_DELAY_MIN;
         self.game.lanes[self.checkLane].delay = 0;

      end

      self.checkLane = self.checkLane + 1;

      if self.checkLane > #self.game.lanes then
         self.checkLane = 1;
      end
   end

end

----------------------------------------------------------------------

function GameSpawner:_getBountyVehicleTemplate( entry, chosen )

   local bounty         = BountyMgr:getBountyVehicleTemplate(entry);
   local t              = Template.clone(bounty.template);

   t.texture            = bounty.paint;
   t.armor              = bounty.armor;
   t.score              = chosen.score;
   t.hitPoints          = BountyMgr:getBountyVehicleHeath();
   t.damageToAvatar     = chosen.damageToAvatar;
   t.tangoId            = entry.tangoId;
   t.bountyPic          = ContactMgr:getContactPicture(entry.contact) or "resources/UI/PlayerPortrait.png";

   return t;

end

----------------------------------------------------------------

function GameSpawner:disableSpawn()

   self.spawnEnabled = false;

end

----------------------------------------------------------------

function GameSpawner:enableSpawn()

   self.spawnEnabled = true;

end

----------------------------------------------------------------

function GameSpawner:disableLane(index)

   self.laneEnabled[index] = false;

   print("GameSpawner:disableLane",index,self.laneEnabled[index]);
end

----------------------------------------------------------------

function GameSpawner:disableLaneFor(index, timer)

   self.laneDisableTimer[index] = math.max(self.laneDisableTimer[index],timer);

end

----------------------------------------------------------------

function GameSpawner:enableAllLanes()

   self.laneEnabled = { true, true, true, true, true };

end

----------------------------------------------------------------

function GameSpawner:enableLane(index)

   self.laneEnabled[index] = true;

end

----------------------------------------------------------------

function GameSpawner:disableHazard()

   self.hazardEnabled = false;

end

----------------------------------------------------------------

function GameSpawner:enableHazard()

   self.hazardEnabled = true;

end

----------------------------------------------------------------

function GameSpawner:resetHazardTimer()

   self.hazardTimer = 0;

end

----------------------------------------------------------------

function GameSpawner:getHazardTimer()

   return self.hazardTimer;

end

----------------------------------------------------------------

function GameSpawner:spawnHazard( tpl, x, y, target )

   self.entityMgr:spawnHazard(     tpl
                                 , x,  y
                                 , target                   );


end

----------------------------------------------------------------

function GameSpawner:spawnBarricade( tpl, x, y, lane )

   return self.entityMgr:spawnBarricade(  tpl, x,  y, lane  );

end

----------------------------------------------------------------

function GameSpawner:getHazardChart( duration )

   for i=1,#MAP01_HAZARD_SPAWN_CHART do

      local entry = MAP01_HAZARD_SPAWN_CHART[i];

      if duration < entry.duration then
         return entry;
      end
   end

   return nil;

end

----------------------------------------------------------------

function GameSpawner:updateHazards( dt )

   self.hazardTimer = self.hazardTimer+dt;

   if    (#self.entityMgr.hazards > 0) or (#self.entityMgr.barricades > 0) then

      self.hazardTimer = 0;

   elseif self.hazardEnabled then

      local chart = self:getHazardChart( self.game.duration );

      if chart and chart.templates then

         local timeMin    = chart.timeMin or 0;
         local timeMax    = chart.timeMax or 999999;
         self.hazardTimer = self.hazardTimer+dt;

         if       ( self.hazardTimer > timeMin )
            and   ( self.hazardTimer > timeMax or  math.random() < dt*chart.odds ) then

            local hazardTpl = chart.templates[math.random(1,#chart.templates)];

            if hazardTpl.factory == Barricade then

               local lane  = self.game:getClosestLaneTo(self.game.avatar:getLocX());
               self.entityMgr:spawnBarricade(    hazardTpl
                                            ,    self.game:getLaneLocX(lane)
                                            ,    self.game.boundary.max.y
                                            ,    lane );
            else

               self.entityMgr:spawnHazard(  hazardTpl
                                          , self.game.avatar:getLocX()
                                          , self.game.boundary.max.y + 100, 0
                                          , self.game.avatar );
               local mSounds = {start={"MissileWarning",}, };
               EventMgr:postEvent( { sounds=mSounds, name="game spawner",}, "add-sfx" );

            end


            --Make sure to flag there is hazard right away, otherwise we could spawn a scenario for example
            self.hazardTimer = 0;

         end
      end
   end
end

----------------------------------------------------------------

function GameSpawner:setVehicleChartSlowK ( k )

   self.vehicleChartSlowK = k;

end

----------------------------------------------------------------

function GameSpawner:setVehicleSpeedDropStart ( val )

   self.vehicleSpeedDropStart = val;

end


----------------------------------------------------------------

function GameSpawner:setVehicleChartCatchupK ( k )

   self.vehicleChartCatchupK = k;

end

----------------------------------------------------------------

function GameSpawner:setVehicleChartTarget ( time )

   self.vehicleChartTarget = time;

end

----------------------------------------------------------------

function GameSpawner:setVehicleChartMax ( time )

   self.vehicleChartMax = time;

end

----------------------------------------------------------------

function GameSpawner:getVehicleChart ( duration )

   for i=1,#VEHICLE_SPAWN_CHART do

      local entry = VEHICLE_SPAWN_CHART[i];

      if duration < entry.duration then
         return entry;
      end
   end

   return nil;

end

----------------------------------------------------------------

function GameSpawner:isSpeedModeOn( )
   return self.game:isSpeedBurstOn();
end

----------------------------------------------------------------

michelDebug = "";
function GameSpawner:shouldLaneSpawn ( lane )

   local chosen         = nil;
   local speedBurstMode = self:isSpeedModeOn();

   if self.laneEnabled[lane] and  self.laneDisableTimer[lane] <= 0 then

      local chartTime= self.vehicleChartTime;
      local chart    = self:getVehicleChart( chartTime );

      if #self.entityMgr.vehicles < chart.maxTotal then

         local delay1;
         local delay2;

         --When bursting (especially for consumable head start, we want to have stuff in the screen)
         if speedBurstMode then
            delay1 = chart.spawnDelay.speedBurstMode  or 0;
            delay2 = chart.spawnDelay2.speedBurstMode or 0;
         else
            delay1 = math.lerp(chart.spawnDelay[1], chart.spawnDelay[2], self.wave);
            delay2 = math.lerp(chart.spawnDelay2[1],chart.spawnDelay2[2],self.wave);
         end

         if       (self.game.lanes[lane].delay  > delay1)
            and   (self.game.lanes[lane].delay2 > delay2) then

            for j=1,#chart.odds do

               local time  = self.game.duration-self.game.vehicleSpawnedTime[j];
               local odds  = chart.odds[j];

               if time < chart.timeMin[j] then
                  odds = 0;
               elseif chart.timeMax[j] > 0 and time > chart.timeMax[j] then
                  odds = 1;
               end

               --When bursting (especially for consummable head start, we want to have stuff in the screen)
               if speedBurstMode then
                  odds = odds * 2;
               end

               -- there was some case where vehicleSpawnedType[j] was negative
               if        self.game.vehicleSpawnedType[j] < chart.max[j]
                    and  chart.max[j] > 0
                    and  odds  > 0                                     then

                  local r = math.random();

                  if r < odds then
                     chosen = j;
                  end
               end
            end
         end
      end
   end

   return chosen;

end

----------------------------------------------------------------------

function GameSpawner:getVehicleTemplate()

   local level    = App.data:getCurrentPlayerLevel();
   local dpsIndex = App.data:getPlayerWeaponLevel( level );

   Debug.assert(dpsIndex > 0, "GameSpawner:getVehicleTemplate -> vehicle weapon level invalid")

	dpsIndex	= math.max(dpsIndex, 1);
   
   return App.data:getSpawnList(level, dpsIndex );

end

----------------------------------------------------------------------
