require('engine/event/event-manager');
require('game/entity/avatar/avatar');
require('game/entity/background/background');
require('game/entity/explosion');
require('game/entity/vehicle');
require('game/entity-manager');
require('game/game-director');
require('game/game-spawner');
require('game/maps/map01/backgroundCharts');
require('game/maps/map01/projectileCharts');
require('game/maps/map01/pickupCharts');
require('game/maps/map01/vehicleCharts');
require('app/template/flyout/flyoutScore');

----------------------------------------------------------------

local game                       = {}

game.__marker                = "game";

----------------------------------------------------------------

AVATAR_POS_Y_DELTA                        = { 115, 175 };

local PICKUP_SPAWN_CHART                  =  MAP01_PICKUP_SPAWN_CHART;

local KILL_SCORE_MULT_DELTA_Y             = 20;

local BACKGROUND_TEMPLATES                =  nil;

local LANE_SPAWN_DELAY_2_VEHICLE_MAX_Y    = -50;
local HACK_DONT_SPAWN_ENEMIES             = false;
--JOE settings: 0= gun loop, 1= impact sounds, -1= neither
local HACK_GUN_OR_IMPACT                  = -1;
local IMPACT_COOLDOWN_MIN                 = 0.5;
local IMPACT_COOLDOWN_MAX                 = 0.75;

local HACK_ROAD_BACKGROUND_IN_ORDER       = false;
local GAME_PROFILE_ENABLED                = false;

local FLURRY_GAME_PLAYED_GROUP            =
{
   1,    2,    3,    4,    5,    6,    7,    8,    9,
   10,   20,   30,   40,   50,   60,   70,   80,   90,
   100
};

if not BUILD_DEBUG then
   HACK_ROAD_BACKGROUND_IN_ORDER = false;
   GAME_PROFILE_ENABLED = false;
end

----------------------------------------------------------------

local tableSizeTracker = {};
local disqualifiedTables = {};

function _analyzeTable(tbl, visited, path)

   visited = visited or {};

   if visited[tbl] then
      return;
   end

   for k, v in pairs(tbl) do

      if type(v) == "table" and not visited[v] and not disqualifiedTables[v] then

         if path == nil then

            path = tostring(k);
         else
            path = path .. "." .. tostring(k);
         end

         visited[v] = true;

         -- Entries in tableSizeTracker are tables like this: { size = 2, path = "test.thing.tbl" }

         if tableSizeTracker[v] then

            local newSize = 0;
            for _, _ in pairs(v) do
               newSize = newSize + 1;
            end

            if newSize > tableSizeTracker[v].size then

               tableSizeTracker[v].size = newSize;

            elseif newSize < tableSizeTracker[v].size then

               tableSizeTracker[v] = nil;
               disqualifiedTables[v] = true;

            end

         else

            local newSize = 0;
            for _, _ in pairs(v) do
               newSize = newSize + 1;
            end

            tableSizeTracker[v] = { size = newSize, path = path };

         end

         _analyzeTable(v, visited, path);

      end

   end

end

local runningProfiler = false;

function game:startProfile()

   print("game:startProfile");

   -- Declared as locals elsewhere
   tableSizeTracker = {};
   disqualifiedTables = {};

   runningProfiler = true;

end

function game:updateProfile(dt)

   if runningProfiler then
      _analyzeTable(game);
   end

end

function game:stopProfile()

   print("game:stopProfile");

   runningProfiler = false;

   for _, tbl in pairs(tableSizeTracker) do

      print(tbl.path, "size:", tbl.size);

   end

end

----------------------------------------------------------------

function game:onActivate(parent)

   print("game.onActivate", self);

   if GAME_PROFILE_ENABLED then
      self:startProfile();
   end

   local bndIdx = App.data:getValue("road-background-index", 0);

   if bndIdx == 0 then

      bndIdx = 1;
   elseif HACK_ROAD_BACKGROUND_IN_ORDER then

      bndIdx = bndIdx + 1;

      if bndIdx > #MAP01_BACKGROUND_TEMPLATES then
         bndIdx = 1;
      end
   else

      repeat
         bndIdx = math.random(1, #MAP01_BACKGROUND_TEMPLATES);
      until bndIdx ~= App.data:getValue("road-background-index", 0)
   end

   App.data:setValue("road-background-index", bndIdx);


   --local bndIdx            = math.random(1,#MAP01_BACKGROUND_TEMPLATES);
   BACKGROUND_TEMPLATES    =  MAP01_BACKGROUND_TEMPLATES[bndIdx];
   self.backgroundType     = BACKGROUND_TEMPLATES.type;
   self.spawnVehicleY      =  App.viewportHeight/2 + 10;
   self.stopVehicleY       =  self.spawnVehicleY          + 250;
   self.killVehicleY       = -App.viewportHeight/2;

   self.parent             = parent;
   self.gameOver           = false;
   self.camera             = parent.camera.game;

   self.vehicleSpawnDelay = 0;
   self.vehicleSpawnedTime= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
   self.vehicleSpawnedType= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

   self.pickupSpawnTime       = {};
   self.pickupEnabled         = { true, true, true, true, true, true, true, true }
   self.lastTime              = MOAISim.getElapsedTime ();
   self.duration              = 0;
   self.distance              = 0;
   self.distanceScore         = 0;
   self.distanceScoreK        = 0.1;
   --consumable
   self.distanceCounterEnabled= false;
   self.durationCounterEnabled= false;
   self.killScore             = 0;
   self.killCount             = 0;
   self.scoreK                = 1;
   self.secretScoreK          = 1;
   self.currency              = 0;

   --We do not use the real TextureMgr because we want to free texture use in this game
   --This is not the prettiest solution but a last minute fix to work on 3GS and avoid crashes
   self.textures             = {};

   self.weaponLevel        = App.data:getCurrentMainWeaponLevel();
   self.armorLevel         = App.data:getCurrentArmorLevel();
   self.turretLevel        = App.data:getCurrentPlayerTurretWeaponLevel();
   self.gamePlayed         = App.data:getGamePlayed();

   self.gameSpeedK         = 1;
   self.playerDamageK      = 1;

   self.boundary           = {   min   = {   x = -0.5*App.viewportWidth,
                                             y = -0.5*App.viewportHeight },
                                 max   = {   x =  0.5*App.viewportWidth,
                                             y =  0.5*App.viewportHeight },
                                 size  =  {  x =  App.viewportWidth,
                                             y =  App.viewportHeight      }        };

   self.entityMgr          = EntityManager:new( self, self.parent.layer.game );

   self.gameSpawner        = GameSpawner:new( self );
   self.gameSpawner:onActivate();

   self.gameDirector       = GameDirector:new( self );
   self.gameDirector:onActivate();


   local screenHK          = math.clamp((App.viewportHeight-960)/(1136-960),0,1);
   local avatarDY          = math.lerp( AVATAR_POS_Y_DELTA[1],AVATAR_POS_Y_DELTA[2], screenHK);

   self.avatarPosY = -App.viewportHeight/2 + avatarDY;

   self.avatar = self.entityMgr:spawnAvatar(   0, self.avatarPosY  );

   self.background = Background:new( self, self.parent.layer.background[1], BACKGROUND_TEMPLATES, 0, 0, 0 );

   self.input  =  {  down        = false,
                     move           = false,
                     lastDownPt  = { x=0, y=0 }
   };

   self.lanes  =
      {
         { locX  = -App.viewportWidth/2*4/5, delay = 10000, delay2 = 10000, vehicle={} },
         { locX  = -App.viewportWidth/2*2/5, delay = 10000, delay2 = 10000, vehicle={} },
         { locX  =  App.viewportWidth/2*0/5, delay = 10000, delay2 = 10000, vehicle={} },
         { locX  =  App.viewportWidth/2*2/5, delay = 10000, delay2 = 10000, vehicle={} },
         { locX  =  App.viewportWidth/2*4/5, delay = 10000, delay2 = 10000, vehicle={} }
      };
   self.tplPickups = {};
   for i=1, App.data:getPickupCount() do

      self.tplPickups[i] = App.data:getPickupTemplate(i);

   end

   MOAIAdjust:trackEvent(App.adjustMap["Gameplay"], "Count", tostring(App.data:getGamePlayed()) );

   MOAIFlurry:eventStart(  "GamePlay",
                           "ending",         "play",
                           "online",         OnlineMgr:getStatusString(),
                           "vehicle",        "veh-"..App.data:getCurrentPlayerVehicle().."-paint-"..App.data:getCurrentPlayerVehiclePaintVariant(),
                           "level",          ""..App.data:getCurrentPlayerLevel(),
                           "level-xp",       ""..App.data:getCurrentPlayerXP(),
                           "hour",           ""..os.date("%H", os.time())           );


end

----------------------------------------------------------------

function game:onDeactivate()

   self:stopProfile();

   if self.gameDirector then
      self.gameDirector:onDeactivate();
      self.gameDirector       = nil;
   end

   if self.gameSpawner then
      self.gameSpawner:onDeactivate();
      self.gameSpawner       = nil;
   end

   if self.entityMgr then
      self.entityMgr:onDeactivate();
      self.entityMgr          = nil;
   end

   if self.avatar then
      self.avatar:delete();
   end

   self.textures           = nil;
   self.avatar             = nil;
   self.background         = nil;
   self.input              = nil;
   self.lanes              = nil;
   self.tplPickups         = nil;

end

----------------------------------------------------------------

function game:getTexture(fileName)

   local texture = self.textures[fileName]

   if texture == nil then

      local finalFileName = TextureMgr:getFinalFileName( fileName );

      if MOAIFileSystem.checkFileExists( createFullFilename( MOAIFileSystem.getWorkingDirectory(), finalFileName ) ) then

         texture = MOAITexture.new();

         texture:setFilter( MOAITexture.GL_LINEAR,  MOAITexture.GL_LINEAR);
         texture:load(finalFileName);

      elseif MOAIEnvironment.obbDirectory and MOAIFileSystem.checkFileExists ( createFullFilename(MOAIEnvironment.obbDirectory, fileName) ) then

         -- print("Loading " .. tostring(createFullFilename(MOAIEnvironment.obbDirectory, fileName)) .. " in game.");

         texture = TextureMgr:_createTextureTEST( createFullFilename(MOAIEnvironment.obbDirectory, fileName), true, true );

      elseif MOAIFileSystem.checkFileExists( createFullFilename( MOAIEnvironment.documentDirectory, fileName ) ) then

         texture = TextureMgr:_createTextureFromUserImage( fileName, true );

      else
         Debug.error("Failed to load texture from any source in game.lua. fileName: " .. tostring(fileName));
      end

      self.textures[fileName] = texture;
   end

   return texture;

end

----------------------------------------------------------------

function game:onFocus()

end

----------------------------------------------------------------

function game:onInput()

   local x, y = self.parent:wndToWorld (  InputMgr:getTouchLoc() );

   if self.input.down == true then

      if InputMgr:wasTouchReleased() then

         self.input = { down = false, move = false };

      end

      self.input.move = true;

      self.input.lastDownPt = { x = x, y = y };

   else

      if InputMgr:wasTouchPressed() then

         self.input =
         {
            down        = true;
            move        = false;
            firstDownPt = { x = x, y = y };
         }

         self.input.lastDownPt = { x = x, y = y };

      end
   end

   self.input.curPt = { x = x, y = y };

end

----------------------------------------------------------------

function game:onStarted()

   self.gameSpawner:enableSpawn();
   self.distanceCounterEnabled= true;
   self.durationCounterEnabled= true;

   self.gameDirector:onStarted();


   EventMgr:postEvent( { game = self, count=App.data:getGamePlayed() }, "game-start" );

end


----------------------------------------------------------------

function game:onEnded()

   EventMgr:postEvent( { game = self }, "game-end" );

end

----------------------------------------------------------------

function game:onPause()

   local vehicles = self.entityMgr.vehicles;

   for i,v in pairs( vehicles) do
      v:pause();
   end

end

----------------------------------------------------------------

function game:onResume()

   local vehicles = self.entityMgr.vehicles;

   for i,v in pairs( vehicles) do
      v:resume();
   end

end

----------------------------------------------------------------

function game:setGameSpeedK(k)
   self.gameSpeedK = k;
end

----------------------------------------------------------------

function game:setPlayerDamageK(k)
   self.playerDamageK = k;
end

----------------------------------------------------------------

function game:setDistanceScoreK(k)
   self.distanceScoreK = k;
end

----------------------------------------------------------------

function game:getDistanceScoreK(k)
   return self.distanceScoreK;
end

----------------------------------------------------------------

function game:onUpdate(dt)

   self:updateProfile(dt);

   self.lastTime     = time;
   self.scoreK       = self.avatar:getScoreK();
   self.secretScoreK = self.avatar:getSecretScoreK();

   self.avatarSpeedY = self:updateAvatar(  dt );

   if not self.avatar:isDead()  then

      if self.durationCounterEnabled then
         self.duration = self.duration + dt;
      end

      if self.distanceCounterEnabled then

         self.distance      = self.distance + math.max( self.avatarSpeedY*dt, 0 );
         self.distanceScore = self.distanceScoreK*self.distance;
         self.bonusScore    = math.floor((self.distanceScore+self.killScore) * App.data:getScoreBonus());

         EventMgr:postEvent(  {distance=self.distanceScore, score=self.distanceScore+self.killScore, xLoc=self.avatar:getLocX(), game=self}, "avatar-score" );

      end
   end

   self.background:update( dt, self.avatarSpeedY );

   self:updateExplosion( dt, self.avatarSpeedY );

   self:updatePickup( dt );

   self:updateHazards( dt, self.avatarSpeedY );

   self:updateProjectiles( dt );

   self:updateLanes( dt );

   self:updateVehicles( dt, self.avatarSpeedY );

   self:updateBarricades( dt );

   self:updateFlyouts( dt, self.avatarSpeedY );

   self.gameDirector:update( dt );

   self.gameSpawner:update(dt);

   self.entityMgr:update( dt );

   if       ( self.avatar:isDead()  )
      and   ( self.avatar:getSpeedY(false) < 100 ) and not self.reviving then

      if not self.gameOver then

         EventMgr:postEvent( {distance=self.distanceScore, score=self.distanceScore+self.killScore}, "game-over" );

         MOAIFlurry:eventStop("GamePlay",
                              "ending",         "death",
                              "online",         OnlineMgr:getStatusString()      );


         local gamePlayed = App.data:getGamePlayed()+1;
         --Game played is not updated yet... +1
         MOAIFlurry:event(    "GameOver",
                              "played",      ""..gamePlayed,
                              "score",       ""..math.floor(self.distanceScore+self.killScore),
                              "distance",    ""..math.floor(self.distanceScore),
                              "kill",        ""..math.floor(self.killScore),
                              "killCount",   ""..self.killCount,
                              "gold",        ""..self.currency,
                              "key",         ""..App.data:getKeyCount(),
                              "online",      OnlineMgr:getStatusString()
                          );

         for i=1,#FLURRY_GAME_PLAYED_GROUP do
            local min = FLURRY_GAME_PLAYED_GROUP[i];

            --When hitting this milestone, give some extra information
            if gamePlayed == min then

               local firstLaunch = App.data:getValue("firstLaunch");
               local daysPlayed  = nil;
               if firstLaunch then
                  daysPlayed = ""..math.floor((os.time()-firstLaunch)/(24*60*60));
               end

               MOAIFlurry:event("GamePlayedGroup-"..min,
                                 "score",       ""..math.floor(self.distanceScore+self.killScore),
                                 "score-1000",  ""..math.floor((self.distanceScore+self.killScore)/1000),
                                 "distance",    ""..math.floor(self.distanceScore),
                                 "killScore",   ""..math.floor(self.killScore),
                                 "killCount",   ""..self.killCount,
                                 "gold",        ""..self.currency,
                                 "gold-10",      ""..math.floor(self.currency/10),
                                 "authenticate", App:createAuthenticateName(),
                                 "daysPlayed",   daysPlayed
                             );
            end
         end

      end

      self.gameOver = true;

   end

end

----------------------------------------------------------------

function game:getLaneLocX(index)

   if self.lanes and self.lanes[index] then
      return self.lanes[index].locX;
   end

   return 0;
end

----------------------------------------------------------------

function game:getClosestLaneTo(x)

   local bestDist = 100000;
   local index    = 1;

   for i=1,5 do

      local dist = math.abs(self.lanes[i].locX-x);

      if dist < bestDist then

         bestDist = dist;
         index    = i;

      end
   end

   return index;
end

----------------------------------------------------------------

function game:updateLanes(dt)

   for i=1,5 do

      local ok = true;

      self.lanes[i].delay = self.lanes[i].delay + self.gameSpeedK*dt;

      for j=#self.lanes[i].vehicle,1,-1 do

         local v = self.lanes[i].vehicle[j];

         if (v:getLocY()+v:getHeight()/2) > LANE_SPAWN_DELAY_2_VEHICLE_MAX_Y then

            ok = false;

         end
      end

      if ok then

         self.lanes[i].delay2 = self.lanes[i].delay2 + self.gameSpeedK*dt;

      else

         self.lanes[i].delay2 = 0;

      end
   end
end

----------------------------------------------------------------

function game:killScoreFlyout(x, y, score, multiplier, multTpl, scoreTpl)

   local multTemplate  = multTpl or MULTIPLIER_FLYOUT_TEMPLATES;
   local scoreTemplate = scoreTpl or SCORE_FLYOUT_TEMPLATES;
   local index         = math.min( math.ceil(multiplier), #multTemplate[self.backgroundType] );

   if multiplier > 1 then
      local rndMult = math.roundToDecimal(self.scoreK, 1);

      self.entityMgr:spawnFlyoutText( multTemplate[self.backgroundType][index], x,  y+12+multiplier,  rndMult.."X");
      self.entityMgr:spawnFlyoutText( scoreTemplate[self.backgroundType][index], x,  y-12-multiplier,  "" .. score);
   else
      self.entityMgr:spawnFlyoutText( scoreTemplate[self.backgroundType][index], x,  y,  "" .. score);
   end

end

----------------------------------------------------------------

function game:updateCamera(dt)

--[[
   if self.cameraShakeTime then

   print("game.updateCamera");

      local intensity = 100;
      local frequency = 3;
      local backK     = 0.005;
      local duration  = 0.5/frequency;

      self.cameraShakeTime = self.cameraShakeTime + dt;

      local timeK = 1-math.min(self.cameraShakeTime/duration,1);
      local shake = timeK*intensity*math.sin(self.cameraShakeTime*frequency*2*math.pi);

      self.parent.camera.game:setLoc( shake,0 );
      self.parent.camera.background:setLoc( backK*shake,0 );

      if self.cameraShakeTime > duration then

         self.cameraShakeTime = nil;
         self.parent.camera.game:setLoc( 0, 0 );
         self.parent.camera.background:setLoc( 0, 0 );

      else


      end
   end]]

end

----------------------------------------------------------------

function game:updateCollisionVehToVeh()

   local vehicles = self.entityMgr.vehicles;
   for i=#vehicles-1,1,-1 do

      local vehI = vehicles[i];
      local tb = nil;

      if vehI.collideWithVehicle  then

         for j=#vehicles,i+1,-1 do

            local vehJ  = vehicles[j];

            if vehJ.collideWithVehicle  then

               local collision = Collision:checkCollision( vehI, vehJ );

               if collision then

                  vehI:onCollision( collision  );
                  vehJ:onCollision( collision  );

                  if vehJ:isDead() and not vehJ.playDeathAnim then

                     tb = {"collision"};

                     self:killVehicle( vehJ, j, true, nil, tb, vehI );
                  end
               end
            end
         end
      end

      if vehI:isDead() and not vehI.playDeathAnim then

         tb = {"collision"};
         self:killVehicle( vehI, i, true, nil, tb, vehJ );
      end

   end
end

----------------------------------------------------------------

function game:updateFlyouts(dt, avatarSpeedY)

   for i=#self.entityMgr.flyouts,1,-1 do
      self.entityMgr.flyouts[i]:update(dt, avatarSpeedY);
   end

end

----------------------------------------------------------------

function game:updateVehicles(dt, avatarSpeedY)

   local a           = self.avatar;
   local vehicles    = self.entityMgr.vehicles;

   for i=#vehicles,1,-1 do

      local v = vehicles[i];

      v:update( dt, avatarSpeedY );

      if v.type then
         self.vehicleSpawnedTime[v.type] = self.duration;
      end

      if v.hitMarkerCooldown and v.hitMarkerCooldown >= 0 then
         v.hitMarkerCooldown = v.hitMarkerCooldown - dt;
      end

      if v:isDead() and not v.playDeathAnim then
         self:killVehicle( v, i, true);
      end

      if v.collideWithAvatar then

         local collision = Collision:checkCollision( a, v );

         if  collision then

            a:onCollision( collision, v );
            v:onCollision( collision );

            if v:isDead() and not v.playDeathAnim then

               local tb = {"collision"};

               self:killVehicle( v, i, true, nil, tb, a );

            end
         end
      end
   end

   self:updateCollisionVehToVeh();

end

----------------------------------------------------------------

function game:updateAvatar(dt)

   if self.avatar then

      self.avatar:update( dt, self.input );

      return self.avatar:getSpeedY(true);

   end

   return 0;
end


----------------------------------------------------------------

function game:onRevive()

   if self.avatar then
      self.avatar:onRevive();
      self.gameDirector:onAvatarRevive( );
   end

end

----------------------------------------------------------------

function game:areaDamageVehicle(x, y, rad, damageVehicle, damageAvatar, damageHazard, damageProjectile)

   if damageAvatar and damageAvatar > 0 then

      if self.avatar:isCollideRadius(x, y, rad) then

         local x,y = self.avatar:getLoc();
         self.avatar:onDamage(damageAvatar,"AreadDamageVehicle");

      end
   end

   if damageVehicle and damageVehicle > 0 then

      local vehicles = self.entityMgr.vehicles;
      for mi=#vehicles,1,-1 do

         local m = vehicles[mi];

         if m:isCollideRadius(x, y, rad) and not m.invicible then

            if m:onDamage(damageVehicle) and not m.playDeathAnim then

               local tb = {"area"};

               self:killVehicle( m, mi, true, nil, tb );

            end
         end
      end
   end

   if damageHazard and damageHazard > 0 then

      local barricades = self.entityMgr.barricades;
      for bi=#barricades,1,-1 do
         for oi=#barricades[bi].obstacles,1,-1 do
            local o = barricades[bi].obstacles[oi];

            if o:isCollideRadius(x, y, rad) then

               o:onDestroyed();
            end
         end
      end
      local missiles = self.entityMgr.hazards;
      for bi=#missiles,1,-1 do
         local o = missiles[bi].object;

         if o:isCollideRadius(x, y, rad) then
            o:onDestroyed();
         end
      end
   end

   if damageProjectile and damageProjectile > 0 then

      local projectiles = self.entityMgr.projectiles;
      for pi=#projectiles,1,-1 do
         local p = projectiles[pi];

         if p:isCollideRadius(x, y, rad) then
            p:onDeleteMe();
         end
      end
   end

end

----------------------------------------------------------------

function game:killVehicle(m, mi, legit, pickupChart, deathT, klr)

   --Did te player kill it?... did we already drop it
   if legit and (not m.deleteMe ) then

      local x,y = m:getLoc();

      if self.distanceCounterEnabled then
         self.killCount = self.killCount + 1;
         self.killScore = math.floor( self.killScore + self.secretScoreK * self.scoreK * m.score );
         self:killScoreFlyout( x, y, math.floor( self.scoreK*m.score ), self.scoreK, m.flyoutMult, m.flyoutScore );
      end

      if not deathT then
         deathT = {};
      end

      if self.scoreK > 1 then
         table.insert(deathT, "multiplier");
      end

      if self.avatar.powerups then
         if self.avatar.powerups["powerup_double"] then
            table.insert(deathT, "doubleshot");
         end
      end


      EventMgr:postEvent({ monster=m, death=deathT, time=self.duration, lastDmgTime=self.avatar.timeSinceDamage}, "avatar-kill");

      EventMgr:postEvent(
         {distance=self.distanceScore, score=self.distanceScore+self.killScore, xLoc=self.avatar:getLocX(), game=self}, "avatar-score" );

      if m.template.explosion then

         self.entityMgr:spawnExplosion(  m.template.explosion, x, y, 0  );

      end

      local drop             = m.drop or m.template.drop;
      local dropIndex        = m.dropIndex or m.template.dropIndex;
      local specialDrop      = m.specialDrop or m.template.specialDrop;
      local specialDropOdds  = m.specialDropOdds or m.template.specialDropOdds;
      local specialDropIndex = m.specialDropIndex or m.template.specialDropIndex;
      local specialAdditive  = m.specialAdditive or m.template.specialAdditive;
      local tangoId          = m.template.tangoId or nil;
      local playerPic        = nil;

      if tangoId then
         playerPic = ContactMgr:getContactPicture(RankingMgr:getRankEntry(RankingMgr:getRankIndex(tangoId)).contact) or "resources/UI/PlayerPortrait.png";
         EventMgr:postEvent({ Id=tangoId}, "bounty-vehicle-removed");
      end

      if drop then

         local special   = specialDropOdds  and ( math.random() <= specialDropOdds );
         local dropCount = math.random( drop[1] , drop[2] );

         if special then
            specialDropCount = math.random( specialDrop[1] , specialDrop[2] );
         end

         pickupChart = pickupChart or self:getPickupChart(self.duration);

         --drop offset isuse to allow to spawn different "special" or "scripted" pickup
         -- or to not always have the same order
         local dropOffset  = math.random(1,100);

         if special and specialDropCount then

            for i=1,specialDropCount do

               local chosenIdx = nil;
               local chosen    = nil;
               local sum       = 0;

               if specialDropIndex then
                  chosen = self.tplPickups[ specialDropIndex[(dropOffset+i)%#specialDropIndex+1] ];
               end

               if chosen then

                  chosen = BountyMgr:setBountyPickup( chosen, tangoId, playerPic );

                  self.entityMgr:spawnPickup( chosen,  x, y, i, dropCount, m.template.dropSpeedK );

                  if chosen.type == "mystery" then
                     EventMgr:postEvent({}, "pickup-mystery-spawned");
                  end
               end
            end

         end

         for i=1,dropCount do

            local chosenIdx = nil;
            local chosen    = nil;
            local sum       = 0;

            if special and not specialAdditive then

            elseif dropIndex  then

               chosen = self.tplPickups[ dropIndex[(dropOffset+i)%#dropIndex+1] ];

            else
               for j=1,#pickupChart.odds do

                  local tpl  = self.tplPickups[j];

                  local odds = self:calcPickupOdds(pickupChart, j, tpl.type);

                  if  odds > 0 then

                     sum = sum + odds;

                     local r = math.random(1000)/1000*sum;

                     if r <= odds then

                        chosen      = self.tplPickups[j];
                        chosenIdx   = j;

                     end
                  end
               end
            end

            if chosen then
               chosen = BountyMgr:setBountyPickup( chosen, tangoId, playerPic );
               self.entityMgr:spawnPickup( chosen,  x, y, i, dropCount, m.template.dropSpeedK );

               if chosen.type == "mystery" then
                  EventMgr:postEvent({}, "pickup-mystery-spawned");
               end
            end
         end
      end
   end

   m.deleteMe = true;

end

----------------------------------------------------------------

function game:updateExplosion(dt, avatarSpeedY)

   local explosions = self.entityMgr.explosions;
   for i=#explosions,1,-1 do

      explosions[i]:update(dt, avatarSpeedY);

   end
end


----------------------------------------------------------------

function game:isSpeedBurstOn()

   return self.avatar:isSpeedBurstOn();
end

----------------------------------------------------------------

function game:spawnPickupFromList(list, location, forceK)

   for i=1,#list do

      local tpl = self.tplPickups[list[i]];

      if tpl then
         self.entityMgr:spawnPickup( tpl, location[X], location[Y], i, #list, forceK );
      end
   end
end

----------------------------------------------------------------

function game:updatePickup(dt)

   local magnetRad      = self.avatar:getMagnetRadius();
   local pickups        = self.entityMgr.pickups;
   local ax,ay          = self.avatar:getLoc();
   local deleteY        = self.boundary.min.y;
   local droneL, droneR = 0,0;
   local xL,yL;
   local xR,yR;
   local wL,hL;
   local wR,hR;

   if self.avatar.droneLeft then
      droneL = self.avatar.droneLeft:getMagnetRadius();
      xL,yL  = self.avatar.droneLeft:getLoc();
      wL     = (self.avatar.droneLeft.template.width or 80)/2;
      hL     = (self.avatar.droneLeft.template.height or 80)/2;
   end

   if self.avatar.droneRight then
      droneR = self.avatar.droneRight:getMagnetRadius();
      xR,yR  = self.avatar.droneRight:getLoc();
      wR     = (self.avatar.droneRight.template.width or 80)/2;
      hR     = (self.avatar.droneRight.template.height or 80)/2;
   end

   for i=#pickups,1,-1 do

      local p = pickups[i];

      p:update(dt);

      if p.pickable then

         if magnetRad > 0 then
            self:_updatePickupMagLoc(p, ax, ay, magnetRad);
         end

         if droneL > 0 then
            self:_updatePickupMagLoc(p, xL, yL, droneL, "left");
         end

         if droneR > 0 then
            self:_updatePickupMagLoc(p, xR, yR, droneR, "right");
         end

         if not p.magnetized then
            p.magnetized = false;
         end
      end

      if not self.avatar:isDead() and p.pickable then

         self:_testPickupCollision(p);

         if droneL > 0 and xL and yL and wL and hL then
            self:_testPickupCollision(p, true, xL, yL, wL, hL);
         end

         if droneR > 0 and xR and yR and wR and hR then
            self:_testPickupCollision(p, true, xR, yR, wR, hR);
         end
      end

      if p:getTop() < deleteY then
         p:onDeleteMe();
         EventMgr:postEvent({ pickup=p, x=p:getLocX() }, "pickup-missed");
      end
   end

end

----------------------------------------------------------------

function game:_updatePickupMagLoc( p, x, y, radius, droneMount )

   if p and x and y and radius then
      local px,py = p:getLoc();

      local dx = (px-x);
      local dy = (py-y);

      if    ( ( p.magnetized == true ) and ( p.droneMount == droneMount ) )
      or ( ( (dx*dx + dy*dy) < (radius*radius)) and ( p.magnetized == false or p.magnetized == nil ) ) then

         p.magnetized = true;
         p.droneMount = droneMount;

         local relX, relY = math.abs(dx)/math.abs(p.magnetDelta[X])
         , math.abs(dy)/math.abs(p.magnetDelta[Y]);

         -- We are getting further
         if relX > 1 then
            p.magnetDelta[X] = math.sign(dx)*math.abs(p.magnetDelta[X]*0.70);
         else
            p.magnetDelta[X] = dx*0.70;
         end

         -- We are getting further
         if relY > 1 then
            p.magnetDelta[Y] = math.sign(dy)*math.abs(p.magnetDelta[Y]*0.70);
         else
            p.magnetDelta[Y] = dy*0.70;
         end

         local newX, newY = x+p.magnetDelta[X]
         , y+p.magnetDelta[Y];

         p:setLoc(newX,newY);

      end
   end

end

----------------------------------------------------------------

function game:_testPickupCollision( p, notAvatar, x, y, w, h )

   Debug.assert(p, "game.updatePickup -> p");
   Debug.assert(self.avatar, "game.updatePickup -> self.avatar");

   if p then
      local px,py  = p:getLoc();
      local pickup = false;

      if p:isCollideEntity( self.avatar ) then
         pickup = true;
      elseif notAvatar and x and y and w and h
      and    math.isInBoundary( px, py, x-w, x+w, y-h, y+h ) then

         pickup = true;
      end

      if pickup then
         EventMgr:postEvent({ pickup=p, avatar=self.avatar, duration=self.duration}, "avatar-pickup");

         self.avatar:onPickup(p);
         self.gameDirector:onAvatarPickedUp( p );

         p:onDeleteMe();
      end
   end

end

----------------------------------------------------------------

function game:getPickupChart(duration)

   for i=1,#PICKUP_SPAWN_CHART do

      local entry = PICKUP_SPAWN_CHART[i];

      if duration < entry.duration then

         return entry;

      end
   end

   return nil;

end

----------------------------------------------------------------

function game:enablePickup(index)

   self.pickupEnabled[ index ] = true;
end

----------------------------------------------------------------

function game:disablePickup(index)

   self.pickupEnabled[ index ] = false;
end

----------------------------------------------------------------

function game:calcPickupOdds(pickupChart, index, type)

   if not self.pickupEnabled[index] then
      return 0;
   end

   local odds = pickupChart.odds[index];
   local last = self.pickupSpawnTime[ type ] or 0;
   local time = self.duration-last;

   if time < pickupChart.timeMin[index] then
      return 0;
   end

   if pickupChart.timeMax[index] > 0 and time > pickupChart.timeMax[index] then
      return odds + time/pickupChart.timeMax[index];
   end

   return odds;
end

----------------------------------------------------------------

function game:updateHazards(dt, avatarSpeedY)

   for i=#self.entityMgr.hazards,1,-1 do

      local h = self.entityMgr.hazards[i];

      h:update(dt, avatarSpeedY);

      if h:isDone() then

         h.deleteMe = true;

      end
   end
end

----------------------------------------------------------------

function game:updateBarricades(dt)

   for i=#self.entityMgr.barricades,1,-1 do

      local b = self.entityMgr.barricades[i];

      b:update(dt,self.avatarSpeedY);

      if b:isDone() then

         b.deleteMe = true;

      end

   end
end

----------------------------------------------------------------

function game:updateProjectiles(dt)

   local pickupChart = nil;
   local killMeY     = self.boundary.max.y;
   local entMgr      = self.entityMgr;
   local projectiles = entMgr.projectiles
   local vehicles    = entMgr.vehicles;

   if not self.impactCD then
      self.impactCD = math.random(IMPACT_COOLDOWN_MIN, IMPACT_COOLDOWN_MAX);
   end
   self.impactCD = self.impactCD - dt;

   for i=#projectiles,1,-1 do

      local p = projectiles[i];

      if not p.deleteMe then

         p:update(dt);

         -- This is a hack because MOAI sucks
         -- When parenting we will always be 1 frame behind... unless we force it here
         if p.hasParent and p.prop then
            p.prop:buildTransform();
         end

         local px,py  = p:getLoc(p.hasParent);
         local pcx, pcy = p:getCollisionSize();
         local h      = p:getHeight()/2;

         if     p.clearLinkTimer
            and p.hasParent
            and p.prop then

            p.clearLinkTimer = p.clearLinkTimer - dt;

            if p.clearLinkTimer <= 0 then
               p.hasParent = false;
               p.prop:clearAttrLink(MOAIProp2D.INHERIT_TRANSFORM);

               if p.parentProp then
                  local a = p.parentProp:getRot();
                  local ownerA = p.owner:getRot() or 0;

                  p.parentProp = nil;
                  a = a + ownerA;

                  if p.pOffset then
                     a = p.pOffset.angle + a;
                  end

                  p:setMoveAngle(a);
                  p:setRot(a);
                  p:setLoc(px, py);
               end
            end
         end

         if (py-h) > killMeY  or (py+h) < -killMeY then

            p:onDeleteMe();

         else

            --Collide with enemy
            if not p.damageAvatar
               and p.damage
            and    p.damage ~= 0 then

               for j= #vehicles,1,-1 do

                  local v = vehicles[j];

                  if v ~= p.owner and not v.invicible then

                     local vx, vy  = v:getLoc();

                     if v:isColliding( px, py, pcx, pcy ) then

                        local newHit = true;

                        if p.noHitRepeat and p:hasHit(v) then
                           newHit = false;
                        end

                        if newHit then

                           local damage = self.playerDamageK*p.damage;
                           v:onDamage( damage );

                           if v.template.onHitSound then
                              if self.impactCD <= 0 then
                                 self.impactCD = math.random(IMPACT_COOLDOWN_MIN, IMPACT_COOLDOWN_MAX);
                                 EventMgr:postEvent( {sounds=v.template.onHitSound, name="bullet impact"}, "add-sfx" );
                              end
                           end

                           if v:isDead() and not v.playDeathAnim then

                              self:killVehicle( v, j, true );

                              if  p.speedBurst then
                                 EventMgr:postEvent( {enemy=v}, "boost-kill" );
                              end

                           elseif p.hitMarker then

                              if not v.hitMarkerCooldown then
                                 v.hitMarkerCooldown = 0;
                              end

                              if v.hitMarkerCooldown <= 0 then
                                 if p.hitMarker.noParent then
                                    entMgr:spawnExplosion(  p.hitMarker, px, py, 0  );
                                 else
                                    local w = v:getWidth()/2;
                                    local h = v:getHeight()/2;

                                    entMgr:spawnExplosion(  p.hitMarker, math.clamp(px-vx, -w, w)*0.75, math.clamp(py-vy, -h, h)*0.75, 0, v  );
                                 end

                                 v.hitMarkerCooldown = p.hitMarker.duration - dt;
                              end
                           end

                           if p.noHitRepeat then
                              p:addHitEntity(v);
                           end

                           local tplExp = p.template.explosion;

                           if tplExp then
                              local vx,vy = v:getLoc();
                              entMgr:spawnExplosion( tplExp, math.lerp(px,vx,0.5), math.lerp(py,vy,0.5) );
                           end

                           if not p.invicible then
                              p:onDeleteMe();
                           end
                        end
                     end
                  end
               end
            end

            --Collide with avatar
            if p.damageAvatar then

               --if p:isCollideEntity( self.avatar ) then
               if self.avatar:isColliding(px, py, pcx, pcy ) then

                  local shieldRatio = self.avatar:isShieldActive();

                  if shieldRatio and shieldRatio > 0 and p.damageAvatar > 0 and not self.avatar:isSpeedBurstOn() then
                     self.avatar:onDamageShield( p.damageAvatar );
                  else

                     self.avatar:onBulletHit( p.damageAvatar, px, py, p.owner );

                     if p.hitMarker then

                        if not self.avatar.hitMarkerCooldown then
                           self.avatar.hitMarkerCooldown = 0;
                        end

                        if self.avatar.hitMarkerCooldown <= 0 then
                           local vx, vy  = self.avatar:getLoc();

                           if p.hitMarker.noParent then
                              entMgr:spawnExplosion(  p.hitMarker, px, py, 0  );
                           else
                              local w = self.avatar:getWidth()/2;
                              local h = self.avatar:getHeight()/2;

                              entMgr:spawnExplosion(  p.hitMarker, math.clamp(px-vx, -w, w)*0.75, math.clamp(py-vy, -h, h)*0.75, 0, self.avatar );
                           end

                           self.avatar.hitMarkerCooldown = p.hitMarker.duration;
                        end
                     end
                  end

                  if not p.invicible then
                     p:onDeleteMe();
                  end
               end
            end

            --Collide with Barricade
            self:_updateProjectileBarricadeCollide(p, px, py, pcx, pcy);

            --Todo: Collide with Missiles
         end
      end
   end
end

----------------------------------------------------------------

function game:_updateProjectileBarricadeCollide( p, px, py, pcx, pcy )

   if p.damageBarricade and p.damageBarricade > 0 then

      local barricades = self.entityMgr.barricades;

      for i=#barricades,1,-1 do
         for j= #barricades[i].obstacles,1,-1 do

            local b = barricades[i].obstacles[j];

            if b ~= p.owner and not b.invicible then

               local bx, by  = b:getLoc();

               local newHit = true;

               if p.noHitRepeat and p:hasHit(b) then
                  newHit = false;
               end


               if newHit and b:isColliding( px, py, pcx, pcy ) then

                  b:onDestroyed();

                  if p.noHitRepeat then
                     p:addHitEntity(b);
                  end

                  local tplExp = p.template.explosion;

                  if tplExp then
                     self.entityMgr:spawnExplosion( tplExp, math.lerp(px,bx,0.5), math.lerp(py,by,0.5) );
                  end

                  if not p.invicible then
                     p:onDeleteMe();
                  end
               end
            end
         end
      end
   end

end

----------------------------------------------------------------

function game:enableDurationCounter()
   self.durationCounterEnabled = true;
end

----------------------------------------------------------------

function game:disableDurationCounter()
   self.durationCounterEnabled = false;
end

----------------------------------------------------------------

function game:enableDistanceCounter()
   self.distanceCounterEnabled = true;
end

----------------------------------------------------------------

function game:disableDistanceCounter()
   self.distanceCounterEnabled = false;
end

----------------------------------------------------------------

return game
