require( 'engine/file/fileMgr' )
require('engine/object/object')
require('engine/utility/math')

----------------------------------------------------------------

local POINTER_Y_OFFSET		= 40;
local TEXTBOX_X_OFFSET		= -19;
local BUBBLE_IN_SCALE		= {0, 1.3, 0.9, 1};
local BUBBLE_IN_TIME			= {0.1, 0.08, 0.05};
local BUBBLE_OUT_SCALE		= {1,0};
local BUBBLE_OUT_TIME		= {0.05}

----------------------------------------------------------------

local Base	= Object;
Guide	= Base:new();

----------------------------------------------------------------

function Guide:new( set, layer )

   local g		= Base.new(self);

	g.index			= 1;

	g.state			= nil;
	g.layer			= layer;

   if set.guideElements then
      g.elements		= self:copyTable(set.guideElements);
   end
	g.pages			= self:copyTable(set.pages);

	g.defaultFrame	= {x=0,y=0,w=100,h=100};


	g.propList		=	{};
	g.widgetList	=	{};
	g.timerList		=	{};
	g.pulseList		=	{};
	g.textBoxPieces =	{};

	g.disabledList		=	{};
	g.transitionList	=	{};

	g.buttonPresses	=	{};

	g.stopping = false;


   return g;

end

----------------------------------------------------------------
----------------------------------------------------------------

function Guide:load( override )

   if override then
      if self.elements then
         --delete existing elements?
      end
      self.elements = override;
   end
   if self.elements then
      for i,v in ipairs (self.elements) do

         if self[v.name] then
            self[v.name]:delete();
         end

         if v.type == "image" then

            self[v.name] = self:createProp( self.layer, v.frame.x, v.frame.y, v.frame.w, v.frame.h, v.texture, v.alpha, v.flipX, v.flipY );

         elseif v.type == "text" then

            self[v.name] = self:createTextProp( self.layer, v.frame.x, v.frame.y, v.frame.w, v.frame.h, v.text, v.fontName, v.fontSize, v.glyphsToLoad, v.black );

         elseif v.type == "textBox" then

            self:constructTextBox(v.element);

         end
      end
   end

end

----------------------------------------------------------------

function Guide:setLayer( layer )

   if layer ~= self.layer and layer then
		if self.layer then
			--remove all elements in list before adding to new layer
			for i,v in ipairs(self.propList) do
				self.layer:removeProp( v );
			end
		end

		self.layer = layer;
		for i,v in ipairs(self.propList) do

			self.layer:insertProp( v );
			v:setPriority(v.priority or 2000 );

		end
	end
end

----------------------------------------------------------------

function Guide:activatePage()

	local page = nil;
	if self.pages then

		page = self.pages[self.index];

	end

	if page then
		if page.init then

			for i,v in ipairs(page.init) do
				if v.activate and v.element then
					if self[v.activate] then
						local t = self:copyTable(v.element);
						self[v.activate](self, t);
					end
				end
			end
      end

		if page.actions then

			self:disabledWidgetCheck( self.state );

			for i,v in ipairs(page.actions) do
				if v.activate and v.element then
					if self[v.activate] then
						local t = self:copyTable(v.element);
						self[v.activate](self, t);
					end
				end
			end
		end
	end

	enableButtonInput();

end

----------------------------------------------------------------

function Guide:deactivatePage()

	disableButtonInput();
	local page = nil;
	if self.pages then
		page = self.pages[self.index];
	end

	if page then
		if page.init then

			for i,v in ipairs(page.init) do
				if v.deactivate and v.element then
					if self[v.deactivate] then
						self[v.deactivate](self, v.element);
					end
				end
			end

      end
		if page.actions then

			self:reEnableWidgets();

			for i,v in ipairs(page.actions) do
				if v.deactivate and v.element then
					if self[v.deactivate] then
						self[v.deactivate](self, v.element);
					end
				end
			end
		end
	end

end

----------------------------------------------------------------

function Guide:update( dt )

	local curState = StateMgr:getCurState();
	local mainLayer = curState.mainLayer or curState.layerTable[1][1];
	if self.state ~= curState and mainLayer then
		self.state = curState;
		self:setLayer( mainLayer );
	end

	if #self.transitionList > 0 then
		self:transition( dt );
	end

	if #self.pulseList > 0 then
		self:pulse( dt );
	end

	if self.bubbleTable then
		self:bubble( dt );
	end

	if #self.timerList > 0 then
		for i,v in ipairs(self.timerList) do
			v.time = v.time-dt;
			if v.time <= 0 then
				if self[v.endFunc] then
					self[v.endFunc](self, v.args);
					table.remove(self.timerList, i);
				end
			end
		end
	end

	if #self.transitionList <= 0 and #self.pulseList <= 0 and #self.timerList <= 0 then

		if self.stopping then
			self.stopping = false;
			self:stopGuide();
			return;
		end

		if #self.buttonPresses > 0 then
			for i,v in ipairs(self.buttonPresses) do
				v.func(v.ref, v.bRef, v.xRef, v.yRef);
				table.remove(self.buttonPresses, i);
			end
		end

		if self.index <= #self.pages then
			if self.pages[self.index].state and self.state then

				if self.pages[self.index].done then
					self:deactivatePage();
					self.index = self.index+1;
					if self.index > #self.pages then
						self.index = #self.pages;
						self.stopping = true;
					end
				end

				if self.pages[self.index].state == self.state.stateFilename then
					if not self.pages[self.index].active then
						self.pages[self.index].active = true;
						self:activatePage();
						if self.index == 1 then
							local sSound = {start={"GuideAppear",},};
							EventMgr:postEvent( { sounds=sSound, name="start guide",}, "add-sfx" );
						end
					end

				end
			end
		else
			self:stopGuide();
		end

	end



end

----------------------------------------------------------------

function Guide:nextPage()

	if self.pages[self.index] then
		self.pages[self.index].done = true;
	end

end

----------------------------------------------------------------

function Guide:stopGuide()

	for i,v in ipairs(self.propList) do
		--remove all props from the layer
		v:setVisible(false);
		self.layer:removeProp(v);
	end

	self.propList = {};

	for i,v in ipairs(self.widgetList) do
		--reset all widgets
		if v.element then
			self:pushWidget(v.element);
		end
	end
	self.widgetList = {};

	if self.pages then
		for j=1, #self.pages do
			self.pages[j].active = nil;
			self.pages[j].done = nil;
		end
	else
		print("no self.pages!");
	end

	GuideMgr:stop();
	enableButtonInput();

end

----------------------------------------------------------------

function Guide:copyTable( t )

  local u = { }
  for k, v in pairs(t) do u[k] = v end
  return setmetatable(u, getmetatable(t))

end

----------------------------------------------------------------
----------------------------------------------------------------

function Guide:transition( dt )

	for i,v in ipairs(self.transitionList) do

		local done = false;
		v.time = v.time+dt;
		local percentDone;
		if v.time <= 0 then
			percentDone = 0;
		elseif v.time >= v.totalTime then
			percentDone=1;
		else
			percentDone = v.time/v.totalTime;
		end

		if (v.sx and v.ex) or (v.sy and v.ey) then
			--move element
			local newX, newY;
			if v.sx then
				newX = v.sx + ((v.ex-v.sx) * percentDone);
			end
			if v.sy then
				newY = v.sy + ((v.ey-v.sy) * percentDone);
			end
			v.prop:setLoc( newX, newY);
		end

		if v.sAlpha and v.eAlpha then
			--fade element in or out
			local newAlpha = v.sAlpha + ((v.eAlpha-v.sAlpha)*percentDone);
			self:setPropAlpha( v.prop, newAlpha, v.black );
		end

		if (v.sSclX and v.eSclX) or (v.sSclY and v.eSclY) then
			--scale element
			local newSclX, newSclY;
			if v.sSclX then
				newSclX = v.sSclX + ((v.eSclX-v.sSclX)*percentDone);
			end
			if v.sSclY then
				newSclY = v.sSclY + ((v.eSclY-v.sSclY)*percentDone);
			end
			v.prop:setScl( newSclX, newSclY );
		end

		if percentDone >= 1 then
			table.remove(self.transitionList, i);
		end

	end

end

----------------------------------------------------------------

function Guide:reframe( element )

	if self[element.name] then

		if element.frame then
			local x,y = self[element.name]:getLoc();
			if element.frame.x then
				if element.frame.x ~= x then
					x = element.frame.x;
				end
			end
			if element.frame.y then
				if element.frame.y ~= y then
					y = element.frame.y;
				end
			end
			self[element.name]:setLoc(x,y);

			if element.frame.w and element.frame.h then
				--set w and h for self[element.name]
			end

			if element.frame.sclX and element.frame.sclY then
				self[element.name]:setScl(element.frame.sclX, element.frame.sclY);
			end
		end

		if element.priority then
			self[element.name]:setPriority(element.priority);
		end

      if element.lineSpace then
         self[element.name]:setLineSpacing(element.lineSpace);
      end
	end

end

----------------------------------------------------------------

function Guide:localFunc( element )

   if self.state[element.func] then

      local r = self.state[element.func]( self.state, element.args);

      if element.returnFunc and self[element.returnFunc] then
         self[element.returnFunc]( self, r );
      end

   else
      Debug.error("Guide tried to call a missing function named: \"" .. tostring(element.func) .. "\"")
   end

end

----------------------------------------------------------------
-- special case "on return" functions
----------------------------------------------------------------

function Guide:bountyReturnedList( rList )

   if rList then

      for i,v in pairs(rList) do

         v:enable();

         table.insert(self.widgetList, v);

      end

   end
end

----------------------------------------------------------------

function Guide:setBountiesToCollect( bounties )

   if bounties then
      self.bountiesToCollect = bounties;
      EventMgr:addListener( self, self.bountyCollect, "bounty-entry-collected");
   end

end

----------------------------------------------------------------

function Guide:bountyCollect( )

   if self.bountiesToCollect then

      if not self.bountiesCollected then
         self.bountiesCollected = 1;
      else
         self.bountiesCollected = self.bountiesCollected + 1;
      end
      if self.bountiesCollected >= self.bountiesToCollect then
         self.bountiesCollected = nil;
         self.bountiesToCollect = nil;
         EventMgr:rmvListener( self, "bounty-entry-collected" );
         self:nextPage();
      end
   end


end

----------------------------------------------------------------
----------------------------------------------------------------

function Guide:show( element )

	if self[element.name] then
		self:reframe(element);
		self[element.name]:setVisible(true);
	end

end

----------------------------------------------------------------

function Guide:hide( element )

	if self[element.name] then
		self[element.name]:setVisible(false);
	end

end

----------------------------------------------------------------

function Guide:fadeIn( element )

	if self[element.name] then

		for i,v in ipairs(self.transitionList) do
			if v.prop == self[element.name] then
				table.remove(self.transitionList, i);
			end
		end

		local t = {
						prop=self[element.name],
						name = element.name,
						totalTime = element.time,
						time = 0,
						sAlpha = 0,
						eAlpha = element.alpha or 1,
						black = element.black,

						};

		self:setPropAlpha( self[element.name], 0, element.black );
		self[element.name]:setVisible(true);
		table.insert( self.transitionList, t );
	end

end

----------------------------------------------------------------

function Guide:fadeOut( element )

	if self[element.name] then

		for i,v in ipairs(self.transitionList) do
			if v.prop == self[element.name] then
				table.remove(self.transitionList, i);
			end
		end
		local t = {
						prop=self[element.name],
						name = element.name,
						totalTime = element.time,
						time = 0,
						sAlpha = element.alpha or 1,
						eAlpha = 0,
						black = element.black,

						};
		self:setPropAlpha( self[element.name], t.sAlpha, t.black );
		table.insert( self.transitionList, t );
	end

end

----------------------------------------------------------------

function Guide:slideIn( element )

	if self[element.name] then

		for i,v in ipairs(self.transitionList) do
			if v.prop == self[element.name] then
				table.remove(self.transitionList, i);
			end
		end
		local t = {
						prop=self[element.name],
						name = element.name,
						totalTime = element.time,
						time = 0,
						sx = element.sx,
						sy = element.sy,
						ex = element.ex,
						ey = element.ey,
						};
		if not t.sx and not t.sy then
			local x,y = t.prop:getLoc();
			local w,h = t.prop:getDims();
			local scW = App.viewportWidth;
			local scH = App.viewportHeight;
			if element.rightSide then
				t.sx = x + (w+scW);
				t.ex = x;
				t.sy = y;
				t.ey = y;
			else
				t.sx = x - (w+scW);
				t.ex = x;
				t.sy = y;
				t.ey = y;
			end


		end

		t.prop:setLoc( t.sx, t.sy );
		t.prop:setVisible(true);
		table.insert( self.transitionList, t );
	end

end

----------------------------------------------------------------

function Guide:slideOut( element )

	if self[element.name] then

		for i,v in ipairs(self.transitionList) do
			if v.prop == self[element.name] then
				table.remove(self.transitionList, i);
			end
		end
		local t = {
						prop=self[element.name],
						name = element.name,
						totalTime = element.time,
						time = 0,
						sx = element.sx,
						sy = element.sy,
						ex = element.ex,
						ey = element.ey,
						};
		if not t.sx and not t.sy then
			local x,y = t.prop:getLoc();
			local w,h = t.prop:getDims();
			local scW = App.viewportWidth;
			local scH = App.viewportHeight;
			if element.rightSide then
				t.sx = x;
				t.ex = x + (w+scW);
				t.sy = y;
				t.ey = y;
			else
				t.sx = x;
				t.ex = x - (w+scW);
				t.sy = y;
				t.ey = y;
			end

		end
		table.insert( self.transitionList, t );
	end

end

----------------------------------------------------------------

function Guide:round(num, idp)

    local mult = 10^(idp or 0)
    if num >= 0 then return math.floor(num * mult + 0.5) / mult
    else return math.ceil(num * mult - 0.5) / mult end

end
--credit to Igor Skoric (i.skoric@student.tugraz.at)

----------------------------------------------------------------

function Guide:constructTextBox( element )

	local stackSize = math.ceil( ((element.textH - (element.chunkH/2))/element.chunkH)/2 );

   local dOffset = POINTER_Y_OFFSET;
   if element.dongleOffset then
      dOffset = element.dongleOffset;
   end

	local pointer = self:round( ((element.portraitY + dOffset) - element.textY)/element.textH, 0);
   if pointer == -0 then
      pointer = 0;
   end

	if #self.textBoxPieces > 0 then
		for i,v in ipairs(self.textBoxPieces) do
			v:setVisible(false);
			v=nil;
		end
	end

	self.textBoxPieces = {};


	if stackSize then
		for i=stackSize,-stackSize, -1 do
			local p;
			local x = element.textX + TEXTBOX_X_OFFSET;
			local y = element.textY + (i * element.chunkH);

			if i == stackSize then
				--create top slice
				p=self:createProp( self.layer, x, y, element.chunkW, element.chunkH,"resources/UI/Avatar_DialogueBubble_Cap.png", element.alpha, false, false );
			elseif i == -stackSize then
				--create bottom slice
				p=self:createProp( self.layer, x, y, element.chunkW, element.chunkH,"resources/UI/Avatar_DialogueBubble_Cap.png", element.alpha, false, true );
			elseif i == pointer and not element.noDongle then
				--create dongle
				local flipX = false;
				if element.portraitX > x then
					flipX=true;
					x = element.textX - (2*TEXTBOX_X_OFFSET) -3.5;
				end
				p=self:createProp( self.layer, x, y, element.chunkW, element.chunkH,"resources/UI/Avatar_DialogueBubble_Dongle.png", element.alpha, flipX, false );
			else
				--create middle slice
				p=self:createProp( self.layer, x, y, element.chunkW, element.chunkH,"resources/UI/Avatar_DialogueBubble_Body.png", element.alpha );
			end
			--add p to text box list for manipulation by the box functions
			p.centerY = element.textY;
			p.offsetY = i * element.chunkH;
			p:setPriority(element.priority or 2001);
			p.xPos = x;
			p.yPos = y;
			table.insert(self.textBoxPieces, p);
		end
	end


end

----------------------------------------------------------------

function Guide:changeText( element )

   if self[element.name] and element.text then

      local text = LocalizationMgr:getString( element.text );

      Debug.assertType(text, "string", "Guide:changeText -> text");

      self[element.name]:setString ( text );
   end

end

----------------------------------------------------------------

function Guide:bubble( dt )

	local bt = self.bubbleTable;
	bt.timers[bt.index] = bt.timers[bt.index] + dt;

	if bt.timers[bt.index] >= BUBBLE_IN_TIME[bt.index] then
		bt.index = bt.index + 1;

		if bt.index > #bt.timers then
			-- bubble is over.
			self:bubbleOver(bt.bIn);
			self.bubbleTable = nil;
			return;
		end
	end

	local delta;
	local pct;
	if bt.sclList[bt.index + 1] and bt.sclList[bt.index] then
		delta = bt.sclList[bt.index + 1] - bt.sclList[bt.index]
		pct = bt.timers[bt.index]/BUBBLE_IN_TIME[bt.index];
	else
		print("scale index out of bounds");
		return;
	end

	if self.textBoxPieces then
		for i,v in ipairs(self.textBoxPieces) do

			local sclAdd = delta * pct;
			local newSclX = bt.sclList[bt.index] + sclAdd;
			local newSclY = bt.sclList[bt.index] + sclAdd;
			local xS, yS = v:getScl(v);

			if xS < 0 then
				newSclX = -newSclX;
			end
			if yS < 0 then
				newSclY = -newSclY;
			end

			v:setScl(newSclX, newSclY);

			local newY = v.centerY + (v.offsetY * newSclX);
			if newSclX < 0 and newSclY > 0 then
				newY = v.centerY + (v.offsetY * newSclY);
			end

			v:setLoc(v.xPos, newY);
		end
	end

end

----------------------------------------------------------------

function Guide:bubbleOver( bIn )

	if bIn then
		for i,v in ipairs(self.textBoxPieces) do
			local sclX, sclY = v:getScl();
			if sclX < 0 then
				sclX = -1;
			else
				sclX = 1;
			end
			if sclY < 0 then
				sclY = -1
			else
				sclY = 1;
			end
			v:setScl(sclX, sclY);
			v:setLoc(v.xPos, v.centerY + v.offsetY);
		end
	else

		for i,v in ipairs(self.textBoxPieces) do
			v:setVisible(false);
		end
	end


end

----------------------------------------------------------------

function Guide:bubbleBoxIn( )

	local s = self:copyTable(BUBBLE_IN_SCALE);

	local t = {};
	if #BUBBLE_IN_TIME > 0  then
		for i=1,#BUBBLE_IN_TIME do
			table.insert(t, 0);
		end
	end

	self.bubbleTable = {index=1, sclList=s, timers=t, bIn=true };

	for j,k in ipairs(self.textBoxPieces) do
		local sx,sy = k:getScl();
		k:setScl(sx*0.01, sy*0.01);
		local x,y = k:getLoc();
		k:setVisible(true);
	end

	local sSound = {start={"ChatBubbleAppear",},};
	EventMgr:postEvent( { sounds=sSound, name="BubbleBoxIn",}, "add-sfx" );

end

----------------------------------------------------------------

function Guide:bubbleBoxOut( )

	local s = self:copyTable(BUBBLE_OUT_SCALE);

	local t = {};
	if #BUBBLE_OUT_TIME > 0  then
		for i=1,#BUBBLE_OUT_TIME do
			table.insert(t, 0);
		end
	end

	self.bubbleTable = {index=1, sclList=s, timers=t, bIn=false };
--[[
 	local sSound = {start={"ChatBubbleVanish",},};
	EventMgr:postEvent( { sounds=sSound, name="BubbleBoxOut",}, "add-sfx" );
  ]]
end

----------------------------------------------------------------

function Guide:swapAnim( element )

	if self.state[element.name] then
		self.state[element.name]:swapAnim(element.swapName);
	end
	if element.span then
		self.state[element.name]:setSpan(element.span);
	end
	if element.speed then
		self.state[element.name]:setSpeed(element.speed);
	end
	if element.loop then

		self.state[element.name]:setMode(MOAITimer.LOOP);
	end
	if element.play then
		self.state[element.name]:start();
	end

end

----------------------------------------------------------------

function Guide:setParent( element )

	if self.state[element.name] and self.state[element.parent] then
		if self.state[element.parent].anim then
         if element.transformOnly then

            if element.animPropIndex then
               self.state[element.name]:setAttrLink(  MOAIProp2D.INHERIT_TRANSFORM,
                                                      self.state[element.parent].anim.props[element.animPropIndex],
                                                      MOAIProp2D.TRANSFORM_TRAIT);
            else
               self.state[element.name]:setAttrLink(  MOAIProp2D.INHERIT_TRANSFORM,
                                                      self.state[element.parent].anim.root,
                                                      MOAIProp2D.TRANSFORM_TRAIT);
            end

         else

            if element.animPropIndex then
               self.state[element.name]:setParent(self.state[element.parent].anim.props[element.animPropIndex]);
            else
               self.state[element.name]:setParent(self.state[element.parent].anim.root);
            end

         end
		elseif self.state[element.parent].prop then
			self.state[element.name]:setParent(self.state[element.parent].prop);
		elseif self.state[element.parent].moaiTextBox then
			self.state[element.name]:setParent(self.state[element.parent].moaiTextBox);
		elseif self.state[element.parent].root then
			self.state[element.name]:setParent(self.state[element.parent].root);
		end
	end

end

----------------------------------------------------------------

function Guide:addTimer( element )

	local t = {name=element.name, time=element.time, endFunc=element.endFunc, args=element.args}

	table.insert(self.timerList, t);

end

----------------------------------------------------------------

function Guide:clearAllTimers()

	if self.timerList and #self.timerList > 0 then
		for i,v in ipairs(self.timerList) do
			v.time = v.time-dt;
			if v.time <= 0 then
				if self[v.endFunc] then
					self[v.endFunc](self, v.args);
					table.remove(self.timerList, i);
				end
			end
		end
	end

end

----------------------------------------------------------------

function Guide:pullWidget( element )

	if element.name and element.type and self.layer then

		if element.type == "button" then
			self:addButton(element, self.state, self.layer);
		elseif element.type == "text" then
			self:addText(element, self.state, self.layer);
		elseif element.type == "anim" then
			self:addAnim(element, self.state, self.layer);
		elseif element.type == "list" then
			self:addList(element, self.state, self.layer);
		elseif element.type == "image" then
			self:addImage(element, self.state, self.layer);
		end
	else
		print("Guide:pullWidget missing arg(s):", element.name, element.type, self.layer);
	end

end

----------------------------------------------------------------

function Guide:pushWidget( element )

	if element.name and element.type and self.layer then

		if element.type == "button" then
			self:resetButton(element, self.state, self.layer);
		elseif element.type == "text" then
			self:resetText(element, self.state, self.layer);
		elseif element.type == "anim" then
			self:resetAnim(element, self.state, self.layer);
		elseif element.type == "list" then
			self:resetList(element, self.state, self.layer);
		elseif element.type == "image" then
			self:resetImage(element, self.state, self.layer);
		end
	else
		print("Guide:pushWidget missing arg(s):", element.name, self.type, self.layer);
	end

end

----------------------------------------------------------------

function Guide:hideWidget( element )

	if element.type then
		if element.type == "button" and self.state[element.name] then
			self.state[element.name]:disable();
		end
	end

	if self.state[element.name] then
		self.state[element.name]:hide();
	end

end

----------------------------------------------------------------

function Guide:showWidget( element )

	if element.type then
		if element.type == "button" and self.state[element.name] then
			self.state[element.name]:enable();
		end
	end

	if self.state[element.name] then
		self.state[element.name]:show();
	end

end

----------------------------------------------------------------

function Guide:addPulse( element )

	if self.state[element.name] then

		local t =	{
							name=element.name,
							buttonAnim=self.state[element.name],
							time=0,
							totalTime=element.time or 0.5;
							startScl=element.sScl or 1,
							endScl=element.eScl or 1.3,
						};
		table.insert(self.pulseList, t);
	end

end

----------------------------------------------------------------

function Guide:stopAllPulses()

	if self.pulseList then

		for i=#self.pulseList, 1, -1 do
			self.pulseList[i].buttonAnim:setScale(self.pulseList[i].startScl or 1);
			table.remove(self.pulseList, i);
		end
	end

end

----------------------------------------------------------------

function Guide:pulse( dt )

	for i,v in ipairs(self.pulseList) do

		if v.reverse then
			v.time = v.time-dt;
		else
			v.time = v.time+dt;
		end

		if v.time > v.totalTime then
			v.reverse = true;
			v.time = v.totalTime;
		elseif v.time < 0 then
			v.reverse = false;
			v.time = 0;
		end
		local percentDone = v.time/v.totalTime;
		local newScl = v.startScl + ((v.endScl-v.startScl) * percentDone);
		v.buttonAnim:setScale(newScl, newScl);

	end



end

----------------------------------------------------------------
----------------------------------------------------------------

function Guide:addText( text, state, layer )

	state[text.name].element = text;
	state[text.name].oldPriority =  state[text.name].priority;

	if state[text.name].priority then
		state[text.name].priority = state[text.name].priority + 2002
	else
		state[text.name].priority = 2002;
	end
	state[text.name]:setPriority(state[text.name].priority or text.priority);

	table.insert(self.widgetList, state[text.name]);

end

----------------------------------------------------------------

function Guide:resetText( text, state, layer )

	if state[text.name] and state[text.name].oldPriority then

		state[text.name]:setPriority(state[text.name].oldPriority);
		state[text.name].element = nil;
		state[text.name].oldPriority = nil;

	end
end

----------------------------------------------------------------

function Guide:addAnim( anim, state, layer )

	if state[anim.name] then

		state[anim.name].element = anim;
		state[anim.name].oldPriority =  state[anim.name].priority;
		if state[anim.name].priority then
			state[anim.name].priority = state[anim.name].priority + 2000
		else
			state[anim.name].priority = 2000;
		end
		state[anim.name]:setPriority(anim.priority or state[anim.name].priority);

		table.insert(self.widgetList, state[anim.name]);

	end

end

----------------------------------------------------------------

function Guide:resetAnim( anim, state, layer )

	if state[anim.name] then
		if state[anim.name].oldPriority then
			state[anim.name]:setPriority(state[anim.name].oldPriority);
			state[anim.name].oldPriority = nil;
			state[anim.name].element = nil;
		end
	end

end

----------------------------------------------------------------

function Guide:addImage( image, state, layer )

	state[image.name].element = image;
	state[image.name].oldPriority =  state[image.name].priority;
	if state[image.name].priority then
		state[image.name].priority = state[image.name].priority + 2000
	else
		state[image.name].priority = 2000;
	end
	state[image.name]:setPriority(state[image.name].priority or image.priority);

	table.insert(self.widgetList, state[image.name]);

end

----------------------------------------------------------------

function Guide:resetImage( image, state, layer )

	state[image.name]:setPriority(state[image.name].oldPriority);
	state[image.name].oldPriority = nil;
	state[image.name].element = nil;

end

----------------------------------------------------------------

function Guide:addList( list, state, layer )

	state[list.name].element = list;

	for i,v in ipairs(state[list.name]) do

		v.oldPriority =  v.priority;
		if v.priority then
			v.priority = v.priority + 2001;
		else
			v.priority = 2001;
		end

		v:setPriority(v.priority);

		table.insert(self.widgetList, v);

	end

end

----------------------------------------------------------------

function Guide:resetList( list, state, layer )

	state[list.name].element=nil;

	for i,v in ipairs(state[list.name]) do

		v:setPriority(v.oldPriority);
		v.oldPriority = nil;

	end

end

----------------------------------------------------------------

function Guide:addButton( button, state, layer )

		self:replaceButtonFuncs( state[button.name] )

		state[button.name].element = button;
		state[button.name].oldPriority =  state[button.name].priority;

		if state[button.name].priority then
			state[button.name].priority = state[button.name].priority + 2001
		else
			state[button.name].priority = 2001;
		end
		state[button.name].releaseCalls = button.releaseCalls;
		state[button.name].pressCalls = button.pressCalls;
		state[button.name].cancelCalls = button.cancelCalls;

		state[button.name]:setPriority(button.priority or state[button.name].priority);

		state[button.name]:enable();

		state[button.name]:setVisible( true );

		--state[button.name]:setColor(1,1,1,1);

		table.insert(self.widgetList, state[button.name]);

end

----------------------------------------------------------------

function Guide:resetButton( button, state, layer )

	if state[button.name] then

		if state[button.name].oldPriority then
			state[button.name]:setPriority(state[button.name].oldPriority);
		end

		state[button.name].oldPriority = nil;
		state[button.name].element = nil;

		self:restoreButtonFuncs( state[button.name] );
	end


end

----------------------------------------------------------------

function Guide:replaceButtonFuncs( button )

	button.oldOnPress			= button.onPressedFunc
	button.oldOnRelease		= button.onReleasedFunc
	button.oldOnCancel		= button.onCanceledFunc
	button.oldOnMove			= button.onMovedFuc

	button.onReleasedFunc	= self.buttonReleased;
	button.onCanceledFunc	= self.buttonCanceled;
	button.onPressedFunc		= self.buttonPressed;
	button.onMovedFunc		= self.buttonMoved

end

----------------------------------------------------------------

function Guide:restoreButtonFuncs( button )


	button.onReleasedFunc	= button.oldOnRelease
	button.onCanceledFunc	= button.oldOnCancel
	button.onPressedFunc		= button.oldOnPress
	button.onMovedFunc		= button.oldOnMove

	if button.oldOnRelease then
		button.oldOnRelease	= nil;
	end

	if button.oldOnCancel then
		button.oldOnCancel	= nil;
	end

	if button.oldOnPress then
		button.oldOnPress		= nil;
	end

	if button.oldOnMove then
		button.oldOnMove		= nil;
	end

	if button.guideCall then
		button.guideCall		= nil;
	end

end

----------------------------------------------------------------
----------------------------------------------------------------

function Guide:buttonReleased( button, x, y  )

	if GuideMgr.guide then

		if button.releaseCalls then
			for i,v in ipairs (button.releaseCalls) do
				GuideMgr.guide[v](GuideMgr.guide);
			end
		end

		if button.oldOnRelease then

			local t=	{func=button.oldOnRelease, ref=self, bRef=button, xRef=x, yRef=y, };

			if #GuideMgr.guide.buttonPresses > 0 then
				for j,k in ipairs(GuideMgr.guide.buttonPresses) do
					if k.func == t.func then
						return;
					end
				end
			end

			table.insert(GuideMgr.guide.buttonPresses, t);

		end
	end

end

----------------------------------------------------------------

function Guide:buttonPressed( button, x, y )

	if GuideMgr.guide then

		if button.pressCalls then
			for i,v in ipairs (button.pressCalls) do
				GuideMgr.guide[v](GuideMgr.guide);
			end
		end

		if button.oldOnPress then

			local t=	{func=button.oldOnPress, ref=self, bRef=button, xRef=x, yRef=y, };

			if #GuideMgr.guide.buttonPresses > 0 then
				for j,k in ipairs(GuideMgr.guide.buttonPresses) do
					if k.func == t.func then
						return;
					end
				end
			end

			table.insert(GuideMgr.guide.buttonPresses, t);

		end
	end

end

----------------------------------------------------------------

function Guide:buttonCanceled( button, x, y )

	if GuideMgr.guide then

		if button.oldOnCancel then
			button.oldOnCancel(self, button, x, y );
		end

		if button.cancelCalls then
			for i,v in ipairs (button.cancelCalls) do
				GuideMgr.guide[v](GuideMgr.guide);
			end
		end

		if button.oldOnCancel then

			local t=	{func=button.oldOnCancel, ref=self, bRef=button, xRef=x, yRef=y, };

			if #GuideMgr.guide.buttonPresses > 0 then
				for j,k in ipairs(GuideMgr.guide.buttonPresses) do
					if k.func == t.func then
						return;
					end
				end
			end

			table.insert(GuideMgr.guide.buttonPresses, t);

		end
	end

end

----------------------------------------------------------------
----------------------------------------------------------------

function Guide:nextPhase()

	self.phaseTarget = self.phaseTarget + 1;

	if self.phaseTarget > #self.phaseStates then
		self:deactivate();
	else
		if self.phaseStates[self.phase] == self.phaseStates[self.phaseTarget] then
			-- no state change, just an update to buttons, text, etc
			self:resetWidgets(self.phaseTarget);
			self:disabledWidgetCheck();
			self.state:disableAllWidget();
			self:addWidgets(self.phaseTarget);
			self.phase = self.phaseTarget;
		else
			--state transition
			self.transition = true;
			self:resetWidgets(self.phaseTarget);
			self:reEnableWidgets();

		end
	end

end

----------------------------------------------------------------

function Guide:disabledWidgetCheck( state )

	self.disabledState = state;

	for i,v in ipairs( self.disabledState.widgets) do
		if not v:isEnabled() then
			table.insert(self.disabledList, v);
		end
	end


	self.disabledState:disableAllWidget();

end

----------------------------------------------------------------

function Guide:reEnableWidgets( ignoreState )

	if self.disabledState then

			self.disabledState:enableAllWidget();

			for i,v in ipairs(self.disabledList) do
				v:disable();
			end
			self.disabledList = {};
	end

end

----------------------------------------------------------------
----------------------------------------------------------------

function Guide:createProp( layer, x, y, w, h, texture, alpha, flipX, flipY )

   if not x then
	  x = self.defaultFrame.x;
   end

   if not y then
	  y = self.defaultFrame.y;
   end

   if not w then
	  w = self.defaultFrame.w;
   elseif w == -1 then
	  w = App.viewportWidth * 2;
   end

   if not h then
	  h = self.defaultFrame.h;
   elseif h == -1 then
	  h = App.viewportHeight * 2;
   end

   local gfxQuad  = MOAIGfxQuad2D.new ();

   gfxQuad:setTexture ( TextureMgr:getTexture(texture, true) );
   gfxQuad:setRect ( -w/2, -h/2, w/2, h/2 );

   local prop  = MOAIProp2D.new ();

   prop:setDeck ( gfxQuad );
   prop:setBlendMode( MOAIProp.GL_SRC_ALPHA, MOAIProp.GL_ONE_MINUS_SRC_ALPHA );

   if flipX then
	  prop:setScl(-1,1);
   end
   if flipY then
	  prop:setScl(1,-1);
   end

   if alpha then
	  self:setPropAlpha(prop, alpha);
   end

   prop:setLoc(x,y);

   if layer then
      layer:insertProp ( prop );
   end

   prop:setVisible(false);

   table.insert(self.propList, prop);

   return prop;

end
----------------------------------------------------------------

function Guide:createTextProp( layer, x, y, w, h, text, fontName, fontSize, glyphsToLoad, black )

   if not x then
	  x = self.defaultFrame.x;
   end

   if not y then
	  y = self.defaultFrame.y;
   end

   if not w then
	  w = self.defaultFrame.w;
   elseif w == -1 then
	  w = App.viewportWidth;
   end

   if not h then
	  h = self.defaultFrame.h;
   elseif h == -1 then
	  h = App.viewportHeight;
   end

   if not text then
		text="";
   else
      text = LocalizationMgr:getString( text );
   end

   local textProp  = MOAITextBox.new ();

   local metaFont = FontMgr:getMetaFont(fontName, fontSize, glyphsToLoad);

   textProp:setFont ( metaFont.font );
   textProp:setGlyphScale ( metaFont.scale );
   textProp:setYFlip ( true );
   textProp:setRect ( -w/2, -h/2, w/2, h/2 );
   textProp:setString ( text );
   textProp.text = text;
   textProp:setAlignment ( MOAITextBox.CENTER_JUSTIFY, MOAITextBox.CENTER_JUSTIFY );

   -- XXX parent is not defined in this scope!
   if parent then
      textProp:setAttrLink ( MOAIProp2D.INHERIT_TRANSFORM, parent, MOAIProp2D.TRANSFORM_TRAIT );
   end

   textProp:setLoc( x, y );

   if layer then
      layer:insertProp(textProp);
   end

   if black then
	  textProp:setColor(0,0,0,1);
   end

   textProp:setVisible(false);

   table.insert(self.propList, textProp);

   return textProp;

end

----------------------------------------------------------------

function Guide:setPropAlpha( prop, a, black )

   if prop then
		if not black then
			prop:setColor(1,1,1,a);
		else
			prop:setColor(0,0,0,a);
		end
   end

end

----------------------------------------------------------------

function Guide:dummy()

	print("dummy call");

end
