require('app/contact/contact-manager')
require('app/loader/load-manager' )
require('app/ranking/ranking-manager')
require( 'engine/file/fileMgr' )
require('engine/object/object')
require('state/banner/banner-manager')
require('state/overlay/overlay-busy')
require('game/maps/map01/backgroundCharts')

----------------------------------------------------------------

StateMgr =
{
   curState = nil,
   loadedStates = {},
   cachedStates = {},
   stateStack = {},
   history = {},
}

local CMD_CLEAR_STATE            = 1;
local CMD_SWAP_STATE             = 2;
local CMD_PUSH_STATE             = 3;
local CMD_POP_STATE              = 4;

local MAX_HISTORY                = 10;

ButtonAnimAction = nil;

UI_HEADER_HEIGHT                 = 68;
SCREEN_DESIGN_SIZE               = 960;
MIN_VIEWPORT_HEIGHT_HEADER       = SCREEN_DESIGN_SIZE+UI_HEADER_HEIGHT;

----------------------------------------------------------------

function StateMgr:init( stateFile, headerAvailable, ... )

   self:_initHeader( headerAvailable );

   BannerMgr:init();

--[[  Music is handled whenever transitioning from one state to another. if a transition is undefined, it will play
   the splash screen music at volume 1. transitions with an oldState defined have priorty over ones with just
   the newState. thus each state can have a default setting, and then different settings for transitioning from
   specific states. --]]

   self.syncBusyDelay         = nil;
   self.syncBusyShowTimeMin   = 1;

   self.currentSong = "";
   --To Joe. It would be better to store the music in the state itself
   self.musicStates = {
                        {newState="state/state-game.lua", song="resources/musics/MX_Spy.ogg", volume=1, fade=2,},
                        {newState="state/state-options.lua", volume=0.65, fade=0.5},
                        {newState="state/state-pause.lua", volume=0.6, fade=0.5},
                        {newState="state/state-death-save.lua", volume=0.6, fade=0.5},
                        {newState="state/state-consumable-overlay.lua", song="resources/musics/MX_Spy.ogg", volume=1, fade=0},
                        {oldState="state/state-consumable-overlay.lua", newState="state/state-auto-purchase-popup-maz.lua", song="resources/musics/MX_Spy.ogg", volume=0.6, fade=1},
                        };
   self.overlayBusy  = OverlayBusy:new();

   --this listener just plays the "purchase success" sfx
   EventMgr:addListener( self, self.onDataServerError,    "data-server-error" );
   EventMgr:addListener( self, self.onDataSaveError,      "data-save-error" );
   EventMgr:addListener( self, self.onDataLoadError,      "data-load-error" );
   EventMgr:addListener( self, self.onDataMerged,         "data-merged" );
   EventMgr:addListener( self, self.onNeedDataResolution, "need-data-resolution" );

   self:_push( stateFile, ... );

end

----------------------------------------------------------------

function StateMgr:_initHeader( headerAvailable )

   self.headerAvailable    =  headerAvailable;

   if self.headerAvailable then

      --The more space we have, the less we need to push the camera down
      local xtraSpace      = math.max((App.viewportHeight-(SCREEN_DESIGN_SIZE+UI_HEADER_HEIGHT))/2,0);
      self.cameraDelta     = math.max(UI_HEADER_HEIGHT/2 + xtraSpace/5,0);

      self._cameraHack     = MOAICamera2D.new();
      self._cameraHack:setLoc(0,self.cameraDelta);


   else

      self.headerLocY      = nil;
      self._cameraHack     = nil;
      self.cameraDelta     = 0;

   end

end

----------------------------------------------------------------

function StateMgr:_isBusy()

   local busyState = nil

   if self.curState then
      if self.curState.isBusy then
         busyState = self.curState:isBusy();
      else
         Debug.assert( type(self.curState) == "table", "Current State not a table" );
         Debug.assert( self.curState.isBusy,          (self.curState.stateFilename or "").. " <- State dont have isBusy");
      end
   end

   return busyState;

end

----------------------------------------------------------------

function StateMgr:PauseGameplay()

   if self.curState and self.curState.PauseGameplay and type(self.curState.PauseGameplay) == "function" then
      self.curState:PauseGameplay();
   end

end

----------------------------------------------------------------

function StateMgr:onEnterBackground()


   if self.curState then
      if self.curState.onEnterBackground then
         self.curState:onEnterBackground();
      else
         Debug.assert( type(self.curState) == "table", "Current State not a table" );
         Debug.assert( self.curState.isBusy,          (self.curState.stateFilename or "").. " <- State dont have onEnterBackground");
      end
   end

end

----------------------------------------------------------------

function StateMgr:onEnterForeground()

   if self.curState then
      if self.curState.onEnterForeground then
         self.curState:onEnterForeground();
      else
         Debug.assert( type(self.curState) == "table", "Current State not a table" );
         Debug.assert( self.curState.isBusy,          (self.curState.stateFilename or "").. " <- State dont have onEnterForeground");
      end
   end
end

----------------------------------------------------------------

function StateMgr:onBackButton()

   --joe brute force fix for android bug. needs better solution inside GuideMgr
   if ( not GuideMgr ) or ( not GuideMgr.active ) then

      if areButtonInputEnabled() then
         if self.curState then
            self.curState:onBackButton();
         end
      end

   end

end

----------------------------------------------------------------

--0.0073939548981007 (during game)
function StateMgr:update( dt )

   --To avoid case where the time will change in the middle of the function
   local osTime = MOAISim.getDeviceTime();

   self:_processCmdState();

   if self.syncBusy then

      self.curState:onUpdate(dt);
      self.overlayBusy:update(dt);

      self.overlayBusy:setText( self.curState:getBusyName() or "" );

      if self.curState:shouldReportBusy() then
         if (MOAISim.getDeviceTime()-(self.lastFlurry or 0) )  >= 1  then
            MOAIFlurry:event( "Busy-Tick",
                              "state", self.curState.stateFilename or "unknown",
                              "time", ""..math.floor((osTime-(self.syncBusyTimer or 0))),
                              "message", self.curState:getBusyName() or "unknown");
            self.lastFlurry = MOAISim.getDeviceTime();
         end
      end

      --To avoid visual glitch, make sure we are showing for a little while
      if      (not StateMgr:_isBusy())
         and ((osTime-(self.syncBusyTimer or 0)) > self.syncBusyShowTimeMin) then

         self.syncBusyTimer = nil;
         self.syncBusy      = nil;
         self.overlayBusy:setLayer(nil);

      end

   else

      if StateMgr:_isBusy() then
         self.syncBusyTimer = (self.syncBusyTimer or osTime);
      else
         self.syncBusyTimer = nil;
      end

      if self.syncBusyTimer and self.syncBusyDelay and ((osTime-self.syncBusyTimer) >= self.syncBusyDelay) then

         if self.curState:shouldReportBusy() then
            MOAIFlurry:event("Busy-Start",   "delay", ""..self.syncBusyDelay,
                                             "time",  ""..math.floor(osTime-self.syncBusyTimer),
                                             "store",  tostring(StoreMgr:isBusy() or false),
                                             "state",  tostring(self.curState:isBusy() or false),
                                             "loader", tostring(LoaderMgr:isBusy() or false));
         end
         self.syncBusy        = true;
         self.syncBusyTimer   = osTime;
         self.lastFlurry      = MOAISim.getDeviceTime();
         self.overlayBusy:setLayer(self:addLayer());
         self.curState:onCancelInput()
         BannerMgr:setLayer( self:addLayer() );
      else

         local handled = BannerMgr:onInput();

         if self.curState then
            self.curState:onInput ( handled )
            self.curState:onUpdate (dt)
         end

      end
   end

   BannerMgr:update(dt);

   LoaderMgr:update(dt);

   self:_processCmdState();


end

----------------------------------------------------------------

function StateMgr:setSyncBusyDelay( delay )

   self.syncBusyDelay = delay;

end

----------------------------------------------------------------

function StateMgr:addStateLayers ( state )

   assert(state.layerTable);

	if state.layerTable [ 1 ] then

		for j, layer in ipairs ( state.layerTable [ 1 ] ) do

         if state:isUsingCameraHack() then
            layer:setCamera( self._cameraHack );
         else
            layer:setCamera( nil );
         end

			MOAIRenderMgr.pushRenderPass ( layer )

		end
	end
end

----------------------------------------------------------------

function StateMgr:_printStack( )

   if BUILD_DEBUG then
      print("--------------------");
      print("StateMgr - Stack");
      for i, state in ipairs ( self.stateStack ) do
         print(state.stateFilename);
      end

      print("--------------------");
   end

end

----------------------------------------------------------------

function StateMgr:addLayer(  )


   local layer = MOAILayer2D.new ();

   layer:setViewport ( App.viewport )


   MOAIRenderMgr.pushRenderPass ( layer )

   return layer;

end

----------------------------------------------------------------

function StateMgr:rebuildRenderStack ( )

	MOAIRenderMgr.clearRenderStack ()

	for i, state in ipairs ( self.stateStack ) do

      if not state.IS_POPUP then
         MOAIRenderMgr.clearRenderStack ();
      end

      self:addStateLayers ( self.stateStack[i] );

	end

end

----------------------------------------------------------------

function StateMgr:loadState ( stateFile )

	if not self.loadedStates [ stateFile ] then

		local newState = dofile ( stateFile )
		self.loadedStates [ stateFile ] = newState
		self.loadedStates [ stateFile ].stateFilename = stateFile

      newState:onLoad();

	end

	return self.loadedStates [ stateFile ];

end

----------------------------------------------------------------

function StateMgr:getCurState ( )

	return self.curState;

end

function StateMgr:isStateInStack ( stateName )

   for i, state in ipairs ( self.stateStack ) do

      if state.stateFilename == stateName then
         return true;
      end
   end

   return false;

end

----------------------------------------------------------------

function StateMgr:makePopup ( state )

	self.state.IS_POPUP = true

end


----------------------------------------------------------------

function StateMgr:_processCmdState()

   while self.stateCmd do

      local type     = self.stateCmd.type;
      local dest     = self.stateCmd.dest;
      local params   = self.stateCmd.params;

      self.stateCmd  = nil;

      if type == CMD_CLEAR_STATE then

         self:_clear( dest, unpack(params) );

      elseif type == CMD_POP_STATE then

         self:_pop( dest, unpack(params) );

      elseif type == CMD_PUSH_STATE then

         self:_push( dest, unpack(params) );

      elseif type == CMD_SWAP_STATE then

         self:_swap( dest, unpack(params) );

      end

   end

end

----------------------------------------------------------------

function StateMgr:clear ( stateFile, ... )

   --A lot of time, swap, pop and push happens within a state,
   -- which can cause problem... so shield by only allowing one at a time and do not process right away
   -- that way, it will be outside of the state and would be more robust
   if not self.stateCmd then

      self.stateCmd =
      {
         type     = CMD_CLEAR_STATE,
         dest     = stateFile,
         params   = {...}
      }

   end

end

----------------------------------------------------------------

function StateMgr:swap ( stateFile, ... )

   --A lot of time, swap, pop and push happens within a state,
   -- which can cause problem... so shield by only allowing one at a time and do not process right away
   -- that way, it will be outside of the state and would be more robust
   if not self.stateCmd then

      self.stateCmd =
      {
         type     = CMD_SWAP_STATE,
         dest     = stateFile,
         params   = {...}
      }

   end

end

----------------------------------------------------------------

function StateMgr:push ( stateFile, ... )

   --A lot of time, swap, pop and push happens within a state,
   -- which can cause problem... so shield by only allowing one at a time and do not process right away
   -- that way, it will be outside of the state and would be more robust
   if not self.stateCmd then

      self.stateCmd =
      {
         type     = CMD_PUSH_STATE,
         dest     = stateFile,
         params   = {...}
      }

   end

end

----------------------------------------------------------------

function StateMgr:pop ( stateFile, ... )

   --A lot of time, swap, pop and push happens within a state,
   -- which can cause problem... so shield by only allowing one at a time and do not process right away
   -- that way, it will be outside of the state and would be more robust
   if not self.stateCmd then

      self.stateCmd =
      {
         type     = CMD_POP_STATE,
         dest     = stateFile,
         params   = {...}
      }

   end

end


----------------------------------------------------------------

function StateMgr:_historyToString()

   local ret = "";

   if self and type(self.history)=="table" then

      for i=1,#self.history do

         local entry = self.history[i] or {};

         ret = ret .. "  " .. tostring(entry.action) .. "->" .. tostring(entry.dest);

      end

   else
      Debug.assert(self,                                 "StateMgr:_historyToString -> self" );
      Debug.assert(self and type(self.history)=="table", "StateMgr:_historyToString -> self and self.history" );
   end

   return ret;

end

----------------------------------------------------------------

function StateMgr:_addToHistory( action, dest )

   if self and type(self.history)=="table" then

      --Only keep the last  elements to our table
      if #self.history >= MAX_HISTORY then
         table.remove(self.history,1);
      end

      table.insert( self.history, {action=action, dest=dest} );

   else

      Debug.assert(self,                                 "StateMgr:_addToHistory -> self" );
      Debug.assert(self and type(self.history)=="table", "StateMgr:_addToHistory -> self and self.history" );

   end

end

----------------------------------------------------------------

function StateMgr:_clear ( stateFile, ... )

   print("StateMgr:_clear", stateFile, ... );

   self:_addToHistory("clear", stateFile);

   if self.curState then

      self.curState:onLoseFocus ()
      self.curState  = nil;

      for i=#self.stateStack,1,-1 do
         local state = self.stateStack [i];
         if state then
            print("state", state.stateFilename);
            state._stackInstCount = state._stackInstCount-1;
            if state._stackInstCount <= 0 then
               state:onDeactivate();
               state._stackInstCount = nil;
            end
         end
      end

      self.stateStack = {}

   end

   self:_push( stateFile, ... );

end


----------------------------------------------------------------

function StateMgr:_swap ( stateFile, ... )

   print("StateMgr:swap", #self.stateStack, self.curState.stateFilename );

   local stamp    = MOAISim.getDeviceTime ();
   local newState = self:loadState ( stateFile );

   self:_addToHistory("swap", stateFile);

   --For now, only report and let it crash to avoid worst bug if this function get call over and over
   if newState == nil then
      Debug.error("StateMgr:_swap->newState is nil");
      Debug.error("StateMgr:_swap->newState is nil: history="..self:_historyToString());
   end

   self.curState:onLoseFocus ()

   self.curState._stackInstCount = self.curState._stackInstCount-1;
   if self.curState._stackInstCount <= 0 then
      self.curState:onDeactivate();
      self.curState._stackInstCount = nil;
   end

   local oldName = self.curState.stateFilename;

   -- update the current state to the new one
   self.stateStack [ #self.stateStack ] = newState;
   self.curState = self.stateStack [ #self.stateStack ]

   if not self.curState._stackInstCount then
      self.curState._stackInstCount = 1;
      self.curState:onActivate();
   else
      self.curState._stackInstCount = self.curState._stackInstCount +1;
   end

   self.curState:onFocus ( ... );

   self:rebuildRenderStack ()

   local s,v,p,f = self:musicCheck ( oldName, self.curState.stateFilename )
   self:playMusic ( s, v, p, f );

   self:_printStack();

   if self.syncBusy then
      self.overlayBusy:setLayer(self:addLayer());
   end

   BannerMgr:setLayer( self:addLayer() );

   GuideMgr:endFocus();

   EventMgr:postEvent({state=self.curState.stateFilename}, "StateMgr.state.new");

end

----------------------------------------------------------------

--Dont call this function directly
function StateMgr:_push( stateFile, ... )

   self:_addToHistory("push", stateFile);

   local newState = self:loadState ( stateFile )

   --For now, only report and let it crash to avoid worst bug if this function get call over and over
   if newState == nil then
      Debug.error("StateMgr:_push->newState is nil");
      Debug.error("StateMgr:_push->newState is nil: history="..self:_historyToString());
   end

   -- do the old current state's onLoseFocus
   local oldState = nil;
   if self.curState then

      self.curState:onLoseFocus ()

      oldState = self.curState.stateFilename;

   end

   -- update the current state to the new one
   table.insert ( self.stateStack, newState )
   self.curState = self.stateStack [ #self.stateStack ]

   if not self.curState._stackInstCount then
      self.curState._stackInstCount = 1;
      self.curState:onActivate();
   else
      self.curState._stackInstCount = self.curState._stackInstCount +1;
   end

   self.curState:onFocus ( ... );

   self:rebuildRenderStack ()

   local s,v,p,f = self:musicCheck( oldState, stateFile );
   self:playMusic ( s, v, p, f );

   self:_printStack();

   if self.syncBusy and self.overlayBusy then
      self.overlayBusy:setLayer(self:addLayer());
   end

   BannerMgr:setLayer( self:addLayer() );

   GuideMgr:endFocus();

   EventMgr:postEvent({state=self.curState.stateFilename}, "StateMgr.state.new");

end

----------------------------------------------------------------

function StateMgr:_pop ( stateFile, ... )

   self:_addToHistory("pop", stateFile);

   --For now, only report and let it crash to avoid worst bug if this function get call over and over
   if #self.stateStack <= 1 then
      Debug.error( "StateMgr:_pop->#self.stateStack <= 1" );
      Debug.error( "StateMgr:_pop->#self.stateStack <= 1: history=" .. self:_historyToString() );
   end

   print("StateMgr:pop", #self.stateStack, self.curState.stateFilename);

   -- Check for an issue where curState will be set to nil
   Debug.assert( #self.stateStack > 1, "curState is about to be nil! #self.stateStack = " .. #self.stateStack .. ", self.curState = " .. (self.curState or {}).stateFilename );

   if #self.stateStack > 1 then

      self.curState:onLoseFocus ()

      self.curState._stackInstCount = self.curState._stackInstCount-1;
      if self.curState._stackInstCount <= 0 then
         self.curState:onDeactivate();
         self.curState._stackInstCount = nil;
      end

      local oldState = self.curState.stateFilename;
      self.curState  = nil

      table.remove ( self.stateStack, #self.stateStack )
      self.curState = self.stateStack [ #self.stateStack ]

      self:rebuildRenderStack ()

      Debug.assert(self.curState, "curState was nil!");

      self.curState:onFocus ( ... )

      local s,v,p,f = self:musicCheck ( oldState, self.curState.stateFilename )
      self:playMusic ( s, v, p, f );

      print("StateMgr:pop-2", #self.stateStack, self.curState.stateFilename);
      self:_printStack();

      if stateFile then
         self:_swap( stateFile, ... );
      end

      if self.syncBusy then
         self.overlayBusy:setLayer(self:addLayer());
      end
      BannerMgr:setLayer( self:addLayer() );

      GuideMgr:endFocus();

      EventMgr:postEvent({state=self.curState.stateFilename}, "StateMgr.state.new");

   else
      self:_addToHistory("invalid_pop", stateFile);
   end

end

----------------------------------------------------------------

function StateMgr:musicCheck ( oldState, newState )

   local song = nil;
   local volume = nil;
   local new = nil;
   local old = nil;
   local pause = false;
   local fade = nil;

   for i,v in pairs(self.musicStates) do
      if v.newState then
         if v.newState == newState then
            if v.oldState then
               if v.oldState == oldState then

                  if v.volume then volume = v.volume else volume = nil end;
                  if v.pause then pause = true else pause = false end;
                  if v.song then song = v.song else song = nil end;
                  if v.fade then fade = v.fade else fade = nil end;

                  song = self:updateInGameMusic(newState, song);

                  return song, volume, pause, fade;
               end
            else

               if v.volume then volume = v.volume else volume = nil end;
               if v.pause then pause = true else pause = false end;
               if v.song then song = v.song else song = nil end;
               if v.fade then fade = v.fade else fade = nil end;

               song = self:updateInGameMusic(newState, song);

               return song, volume, pause, fade;
            end
         else
         -- there's no state specifically defined in the transition list, so call in the default
            song = "resources/musics/Spy_Splash Remix_Chillax_v2.ogg";
            volume = 0.65;
            fade = 2;
            pause = false;
         end
      end
   end

   song = self:updateInGameMusic(newState, song);

   return song, volume, pause, fade;

end


----------------------------------------------------------------

function StateMgr:playMusic ( musicFile, vol, pause, fadeIn )

   if pause then
      if self.music then
         if self.music:isPlaying() then
            self.music:pause();
            return;
         end
      end
   end

   if musicFile and musicFile ~= "" then

      if self.currentSong ~= musicFile then

         print("Play Music:", musicFile, vol, pause, fadeIn);

         if self.music then
            self.music:stop(0.5);
         end

         local layer = AudioMgr:getLayer("music");
         self.music = layer:addMusic( musicFile );
         self.music:play( fadeIn );
      end
      self.currentSong = musicFile;
   else
      musicFile = self.currentSong;
      fadeIn = 2;
   end

   if vol then

      if self.music then
         local curVol = self.music:getMyVolume( )
         if curVol ~= vol then
            self.music:setMyVolume(vol);
         end

      else

         local layer = AudioMgr:getLayer("music");
         self.music = layer:addMusic( musicFile );
         self.music:play( fadeIn );
         self.music:setMyVolume(vol);
      end
   end

   if self.music then
      if self.music:isPaused() then
         self.music:play();
      end
   end

end


----------------------------------------------------------------


function StateMgr:updateInGameMusic( newState, song )

   local bndIdx = App.data:getValue("road-background-index", 1);

   if       bndIdx
      and   ( newState == "state/state-game.lua"
      or    newState == "state/state-consumable-overlay.lua") then

      if MAP01_BACKGROUND_TEMPLATES and MAP01_BACKGROUND_TEMPLATES[bndIdx] and MAP01_BACKGROUND_TEMPLATES[bndIdx].music then
         return MAP01_BACKGROUND_TEMPLATES[bndIdx].music;
      end
   end

   return song;

end


----------------------------------------------------------------

function StateMgr:onDataLoadError()

   print("StateMgr:onDataLoadError", ( self.curState or {} ) .stateFilename );

   if self.curState then
      self.curState:onDataLoadError();
   end

end

----------------------------------------------------------------

function StateMgr:onDataSaveError()

   print("StateMgr:onDataSaveError", ( self.curState or {} ) .stateFilename );

   self.dataSaveError = ( self.dataSaveError or 0 )  + 1;

   if self.curState then
      self.curState:onDataSaveError(self.dataSaveError);
   end

end

----------------------------------------------------------------

function StateMgr:onDataServerError()

   print("StateMgr:onDataServerError");

   self.dataServerError = ( self.dataServerError or 0 )  + 1;

   if self.curState then
      self.curState:onDataServerError(self.dataServerError);
   end

end

----------------------------------------------------------------

function StateMgr:onDataMerged()

   print("StateMgr:onDataMerged");

   if self.curState then
      self.curState:onDataMerged();
   end

end

----------------------------------------------------------------

function StateMgr:onNeedDataResolution(event)

   if self.curState then
      self.curState:onNeedDataResolution(event);
   end

end

----------------------------------------------------------------

function StateMgr:getDataSaveErrorCount()
   return self.dataSaveError or 0;
end

----------------------------------------------------------------

function StateMgr:clearDataSaveError()
   self.dataSaveError  = nil;
end

----------------------------------------------------------------

function StateMgr:getDataServerErrorCount()
   return self.dataServerError or 0;
end

----------------------------------------------------------------

function StateMgr:clearDataServerError()
   self.dataServerError  = nil;
end

----------------------------------------------------------------

function StateMgr:clearBanner()
   BannerMgr:clear();
end

----------------------------------------------------------------------

function areButtonInputEnabled()

   return ButtonAnimAction == nil;

end

----------------------------------------------------------------------

function enableButtonInput()

   ButtonAnimAction = nil;

end

----------------------------------------------------------------------

function disableButtonInput()

   ButtonAnimAction = true;

end

----------------------------------------------------------------
